%**********************************************************
%   Tree View and navigation for disk files
%**********************************************************


%**********************************************************
%   Prototypes
%**********************************************************

define dtr();
define tree_branch(ddir, instr);
define tree_check_subdirs(ddir);
define tree_collapse();
define tree_copy_all();
define tree_create_menu();
define tree_descend();
define tree_display_help();
define tree_end();
define tree_expand_all();
define tree_expand_one();
define tree_find();
define tree_find_again();
define tree_help();
define tree_home();
define tree_init_keymap();
define tree_list_keys();
define tree_menubar();
define tree_move_down();
define tree_move_up();
define tree_node_name();
define tree_node_path();
define tree_node_show();
define tree_node_show_path();
define tree_parent();
define tree_parent_show();
define tree_pgdn();
define tree_pgup();
define tree_point();
define tree_quit();
define tree_refresh();
define tree_save();
define tree_show(root);
define tree_sibling_next();
define tree_sibling_prev();
define tree_test();
define tree_top_root();
define tree_to_listing();
define tree_upper_dir();
define tree_user_root();

   _debug_info = 1;
   _traceback  = 1;

%**********************************************************
%   Variables
%**********************************************************


private variable
    tree_root       = "",
    tree_cwd        = "",   %   current working directory
    tree_angle_char = '`',  %   on last child in a directory
    tree_angle_str  = "`",  %   on last child in a directory
    tree_anglem     = tree_angle_str + "-",
    tree_anglep     = tree_angle_str + "+";

variable TreeView   = "/TreeView/";


%**********************************************************
%   Functions
%**********************************************************


define tree_test()
{
}


define dtr()
%----------------------------------------------------------
%   Tree view of a directory with first level subdirectories
%   Entry point from the outside.
%----------------------------------------------------------
{
    ERROR_BLOCK _clear_error();

    %   save current dir to position tree on it

    variable dir;
    (,dir,,) = getbuf_info();
    if (dir==Null_String) tree_cwd = getcwd();
    else tree_cwd = dir;

    %   tree_cwd does have a terminating slash

    %   close file listing, if there
    sw2buf (DirListing);
    set_buffer_modified_flag (0);
    delbuf (DirListing);

    %   prepare for the tree view
    sw2buf (TreeView);

    tree_init_keymap();
    tree_create_menu();
    hide_status_indicators(1);

    %   display view of current directory

    tree_show (tree_cwd);

    !if (tree_check_subdirs (tree_cwd))
        flush ("No subdirectories, press Backspace");
}


define tree_init_keymap()
{
    reset_slim_keymap ();
    $1 = "tree_slimkmap";
    !if (keymap_p($1)) copy_keymap ($1, "slim_kmap");
    use_keymap ($1);

    undefinekey(Key_Del,$1);

    variable i;
    for (i=32; i<256; i++) undefinekey (char(i), $1);

    definekey ("tree_collapse",         Key_Left,   $1);
    definekey ("tree_copy_all",         "^A",       $1);
    definekey ("tree_end",              Key_End,    $1);
    definekey ("tree_expand_all",       "*",        $1);
    definekey ("tree_expand_one",       Key_Right,  $1);
    definekey ("tree_find",             "^F",       $1);
    definekey ("tree_find_again",       Key_F3,     $1);
    definekey ("tree_help",             Key_F1,     $1);
    definekey ("tree_home",             Key_Home,   $1);
    definekey ("tree_menubar",          Key_F10,    $1);
    definekey ("tree_move_down",        Key_Down,   $1);
    definekey ("tree_move_up",          Key_Up,     $1);
    definekey ("tree_node_show",        " ",        $1);
    definekey ("tree_pgdn",             Key_PgDn,   $1);
    definekey ("tree_pgup",             Key_PgUp,   $1);
    definekey ("tree_quit",             "^W",       $1);
    definekey ("tree_refresh",          "^R",       $1);
    definekey ("tree_save",             "^S",       $1);
    definekey ("tree_sibling_next",     Key_Tab,    $1);
    %definekey ("tree_test",             "^T",       $1);
    definekey ("tree_to_listing",       Key_Enter,  $1);
    definekey ("tree_to_listing",       Key_F6,     $1);
    definekey ("tree_upper_dir",        Key_BS,     $1);

    %   max. 22 keys to avoid scrolling in keys list (help)
}


define tree_create_menu()
{
    %   The menu bar is created only once.
    %   When the create command is issued again,
    %   there is a check for the existence of the menu bar.
    %   To avoid garbage, delete before appending.

    variable tmpbuf = whatbuf();

    sw2buf (TreeView);

    menu_create_menu_bar (TreeView);

    menu_use_menu_bar (TreeView);
    menu_append_popup (TreeView, "&File");
    menu_append_popup (TreeView, "&Display");
    menu_append_popup (TreeView, "&Help");

    $1 = TreeView+".&File";

    menu_delete_items ($1);
    menu_append_item ($1, "&Open",              "tree_to_listing");
    menu_append_separator ($1);
    menu_append_item ($1, "&Save",              "tree_save");
    menu_append_item ($1, "&Refresh",           "tree_refresh");
    menu_append_item ($1, "&Close",             "tree_quit");

    $1 = TreeView+".&Display";

    menu_delete_items ($1);
    menu_append_item ($1, "Upper &Directory",   "tree_upper_dir");
    menu_append_item ($1, "&Previous Sibling",  "tree_sibling_prev");
    menu_append_item ($1, "&Next Sibling",      "tree_sibling_next");
    menu_append_separator ($1);
    menu_append_item ($1, "&Top Directory",     "tree_top_root");
    menu_append_item ($1, "&User Directory",    "tree_user_root");
    menu_append_separator ($1);
    menu_append_item ($1, "&Expand One Level",  "tree_expand_one");
    menu_append_item ($1, "Expand &All",        "tree_expand_all");
    menu_append_item ($1, "&Collapse",          "tree_collapse");
    menu_append_separator ($1);
    menu_append_item ($1, "A&nchor Current Dir","tree_node_show");
    menu_append_item ($1, "&Show Path",         "tree_node_show_path");
    menu_append_item ($1, "Copy &All",          "tree_copy_all");

    $1 = TreeView+".&Help";

    menu_delete_items ($1);
    menu_append_item ($1, "&Quick Help", "tree_display_help");
    menu_append_item ($1, "&List Keys", "tree_list_keys");

    sw2buf (tmpbuf);
}


define tree_menubar()
{
    select_pulldown_file();
}


define tree_help()
{
    select_pulldown_help();
}


define tree_display_help()
{
    variable
        stringA="TreeView",
        stringB="press F1",
        stringC="&Tree View";

    display_hlp2 (stringA,stringB,stringC);
}


define tree_save()
{
    variable b_name = Null_String;

    b_name = read_path_from_mini ("Save to file:");

    %   if response empty or aborted, do nothing
    if (b_name == "") return;

    %   aborted without pressing enter
    if (LASTKEY[0] != 13) return;

    %   select entire listing for saving

    push_spot();
    bob();
    push_mark();
    eob();

    %   check file status before proceeding

    variable b_status = file_status (b_name);

    if (b_status == 2)
        error ("Name already in use for a directory");

    else if (b_status == -1)
        error ("File exists and is not accessible");

    else if (b_status == -2)
        error ("Invalid path");

    else if (b_status == -3)
        error ("Undetermined error");

    else if (b_status == 0)
    {
        write_region_to_file (b_name);
    }

    else if (b_status == 1)
    {
        variable ch = get_mini_response
        ("File exists: append/overwrite/cancel? (a/o/c) ");

        if (ch=='a' or ch=='A')
            append_region_to_file (b_name);

        else if (ch=='o' or ch=='O')
            write_region_to_file (b_name);
    }
    pop_spot();
    tree_refresh();
}


define tree_copy_all()
%----------------------------------------------------------
%   Copy entire tree to clipboard.
%----------------------------------------------------------
{
    push_spot();
    bob();
    push_mark();
    eob();
    call ("copy_region");
    pop_spot();
    flush ("Copied all to clipboard");
}


define tree_to_listing()
{
    variable s = tree_node_path();
    tree_quit();
    sw2buf (DirListing);
    dls (s);
}


define tree_point()
%----------------------------------------------------------
%   Go to first character of current name in tree
%----------------------------------------------------------
{
    bol();

    if(ffind("|+ "));
    else if(ffind("|- "));
    else if(ffind("`+ "));
    else if(ffind("`- "));
    else return;

    go_right(3);
}


define tree_descend()
%----------------------------------------------------------
%   If current tree node has any descendants displayed,
%   move to last displayed descendant.
%----------------------------------------------------------
{
    if (1 == what_line) return;
    %   there is no need for this function on top line
    %   it is only used for collapsing
    %   collapsing on top line takes to upper dir

    tree_point();
    if(looking_at("/")) return;
    variable col = what_column();

    go_down (1);
    goto_column (col);

    while (what_char() == '|')
    {
        go_down (1);
        goto_column (col);
    }

    %    recursive call possible
    if (looking_at ("`")) tree_descend();
    else go_up(1);
}


define tree_pgup ()
{
    variable start_line = window_line ();
    %   cursor is on start_line

    goto_top_of_window ();
    %   move to top of window, if not already there

    if (window_line () != start_line) return;
    %   really moved, current line is not start_line

    else
    {
        %   already top, now paginate
        call ("page_up");
        goto_top_of_window ();
    }
}


define tree_pgdn ()
{
    variable start_line = window_line ();
    %   cursor is on start_line

    goto_bottom_of_window ();
    %   move to bottom of window, if not already there

    if (window_line () != start_line) return;
    %   really moved, current line is not start_line

    else
    {
        % already at bottom of window, now paginate
        call ("page_down");
        goto_bottom_of_window ();
    }
}


define tree_sibling_next()
%----------------------------------------------------------
%   Shift not pressed:  to next sibling in tree
%   Shift pressed:      to previous sibling in tree
%----------------------------------------------------------
{
    if (SHIFTKEY_PRESSED)
    {
        tree_sibling_prev();
        return;
    }

    %   Go to the next sibling, not necessarily the following line
    %   Returns 1 if successful, 0 otherwise.

    if(what_line()==1) return 0;

    %   line is a node, not top root line

    %   do nothing on angle-lines
    bol();
    if (ffind ("`- ") or ffind("`+ "))
    {
        go_right(3);
        return 0;
    }

    variable col, r=1;;
    tree_point();
    go_left (2);
    col = what_column();

    do
    {
        !if (down (1))
        {
            r=0;
            break;
        }
        if (goto_column_best_try (col) != col)
        {
            r=0;
            break;
        }
    } while ( not (looking_at ("-") or looking_at ("+")) );

    %   do not leave push/pop garbage behind
    %   if going to sibling failed, return to spot
    %   if successful, stay put but still pop the spot

    if (r==0) pop_spot();
    else
    {
        push_mark();
        pop_spot();
        pop_mark (1);
    }

    %tree_point();
    return 1;
}


define tree_sibling_prev()
%----------------------------------------------------------
%   To previous sibling in tree
%----------------------------------------------------------
{
    %   Go to previous sibling, if possible
    %   Return 1 if successful, 0 otherwise.

    if(what_line()==1) return 0;

    variable col, r=1;

    tree_point();
    push_spot();

    go_left (3);
    col = what_column();
    go_up (1);
    goto_column (col);

    if (what_char() != '|') r=0;

    else
    {
        %   While that string can be looked at, we go up.
        %   We stop when the strings disappears or
        %   moving up in same column becomes impossible.

        while ( not (looking_at ("|- ") or looking_at ("|+ ")) )
        {
            !if (up (1))
            {
                r=0;
                break;
            }
            if (goto_column_best_try (col) != col)
            {
                r=0;
                break;
            }
        }
    }

    %   do not leave push/pop garbage behind
    %   if going to sibling failed, return to spot
    %   if successful, stay put but still pop the spot

    if (r==0) pop_spot();
    else
    {
        push_mark();
        pop_spot();
        pop_mark (1);
    }

    return r;
}


define tree_move_down()
{
    go_down(1);
    if(is_whiteline) go_up(1);
    tree_point();
}


define tree_move_up()
{
    go_up(1);
    tree_point();
}


define tree_node_show()
{
    EXIT_BLOCK set_readonly (1);
    set_readonly (0);

    %   Anchor tree at current node
    %   provided there is something to list

    if (what_line()==1) return;

    variable s = tree_node_path();

    if(0==tree_check_subdirs (s))
    {
        %   nothing to list
        flush ("No subdirectories.");

        tree_point();
        call ("previous_char_cmd");
        call ("previous_char_cmd");
        call ("delete_char_cmd");
        insert ("-");
        tree_point();
        return;
    }

    %   if control here, s has something to list

    tree_show (s);
}


define tree_node_show_path()
%----------------------------------------------------------
%   Will flush the absolute path name of current tree node.
%----------------------------------------------------------
{
   flush (tree_node_path());
}


define tree_parent()
%----------------------------------------------------------
%    Go to the parent of current node in tree.
%----------------------------------------------------------
{
    tree_point();
    variable cn = what_column();
    if (cn==1) return;

    forever
    {
        tree_point();
        if(what_column() < cn) break;
        go_up(1);
    }
}


define tree_parent_show()
%----------------------------------------------------------
%   Will anchor tree at parent of current root.
%   Used for tree_upper_dir and tree_collapse
%   when cursor on top line.
%----------------------------------------------------------
{
    variable root, offset, tmp_node;
    tmp_node = tree_node_path();

    if (tree_cwd != "/")
    {
        offset = string_match (tree_cwd, "[/][^/]*[/]$", 1);
        root = substr (tree_cwd, 1, offset);
        tree_show (root);

        %   highlight sub-directory we are emerging from
        tmp_node="+ "+substr(tmp_node, offset+1,-1);
        fsearch (tmp_node);
    }
}


define tree_upper_dir()
%----------------------------------------------------------
%   To upper directory without collapsing.
%   If on top line, display branch one up.
%----------------------------------------------------------
{
    while (is_whiteline) go_up(1);

    if (what_line() == 1) tree_parent_show();
    else tree_parent();
}


define tree_quit()
{
    sw2buf (TreeView);
    set_buffer_modified_flag (0);
    set_readonly (0);
    delbuf (TreeView);
}


define tree_refresh()
{
   tree_show (tree_cwd);
}


define tree_top_root()
{
    tree_show ("/");
}


define tree_user_root()
{
    variable homedir = dircat(getenv ("HOME"), Null_String);
    tree_show (homedir);
}


define tree_check_subdirs (ddir)
%----------------------------------------------------------
%   Returns count of (first level) subdirectories in ddir
%----------------------------------------------------------
{
    return tree_branch(ddir,Null_String);
}


define tree_node_name()
%----------------------------------------------------------
%    Returns the name of the current node in tree
%----------------------------------------------------------
{
    variable ll=0, any=Null_String, s=Null_String;

    if (what_line() == 1)
    {
        ll = strlen (tree_cwd);
        if (ll==1) return "/";
        else substr (tree_cwd, 1, ll-1);
        %   ll is positive
    }

    tree_point();
    push_mark();
    eol();
    any = bufsubstr();

    return any;
}


define tree_node_path()
%----------------------------------------------------------
%   Returns the complete pathname of current node.
%----------------------------------------------------------
{
    variable path = Null_String;
    path = tree_node_name();

    if(substr(path,1,1)=="/") return path;
    %   if first character a slash, that's it

    push_spot();
    while (what_line() != 1)
    {
        tree_parent();
        if (strlen(path))
            path = tree_node_name() + path;
    }
    pop_spot();

    return path;
}


define tree_show (root)
{
    EXIT_BLOCK set_readonly (1);
    set_readonly (0);

    %    Anchor the tree at 'root'
    variable tree_status_str = "  |  F1=Help  F10=Menu  ";
    set_status_line ("  "+root+tree_status_str, 0);

    % Buffer status
    erase_buffer();
    bob();

    insert (root);
    newline();

    () = tree_branch (root, "  ");

    tree_cwd = root;

    bob();

    set_buffer_modified_flag (0);

    %   when emerging from sub-directory
    %   highlight will stay on that sub-directory
    %   this is done in tree_parent_show
}


define tree_branch (ddir, instr)
{
    %   ddir    top directory of tree branch
    %           must end with a slash
    %
    %   instr   indentation string before each node.

    %   directory is listed
    %   list is cleansed of everything not a directory
    %   then sorted case-insensitive
    %   then displayed in a tree

    ERROR_BLOCK _clear_error();

    variable A = listdir(ddir);

    %   list is unsorted, i.e. physical disk order
    %   filenames only, no path

    variable L = 0, i, stf, tmp_mode;

    L = length(A);

    if(L==0) return 0;

    %   go through the list, set non-directories to empty
    for (i=0;i<L;i++)
    {
        stf = lstat_file (ddir+A[i]);

        %   reconcile check_subdirectries

        if(stf==NULL)
        {
            A[i]=Null_String;
            continue;
        }

        tmp_mode = stf.st_mode;

        %   we only want links to dirs
        if (stat_is("dir",tmp_mode));
        else if (stat_is("lnk",tmp_mode))
        {
            stf = stat_file (ddir+A[i]);

            if (stf==NULL)
            {
                A[i]=Null_String;
                continue;
            }

            if (stat_is("dir",tmp_mode))
            {
                A[i] = readlink(ddir+A[i]);
                if (substr(A[i],1,1)!="/")
                {
                    %   link to same level in tree
                    %   will produce replicas, get rid of it
                    A[i]=Null_String;
                }
            }
            else A[i]=Null_String;
        }

        else A[i]=Null_String;
    }

    %   shrink the list, retain only directories

    A = A[where(A!=Null_String)];
    L = length(A);

    %   this line only relevant for tree_check_subdirs
    if (strlen(instr)==0) return L;

    if(L==0) return 0;

    %   case-insensitive alphabetical sort
    %   based on strcmpi from site.sl
    A = A[array_sort(A,"strcmpi")];

    %   now paint the tree

    for (i=0;i<L;i++)
    {
        eol();
        insert (instr);

        if (i == L-1) insert (tree_anglep + " ");
        else insert ("|+ ");

        insert(A[i]);
        insert("/");
        if (i != L-1) newline();
    }

    %   Links are listed with target name starting with a slash
    return 1;
}


define tree_expand_one()
{
    EXIT_BLOCK set_readonly (1);

    set_readonly (0);

    variable s=Null_String, sind=Null_String;

    tree_point();
    variable col = what_column();

    if (col==1) return;
    %   this takes care of line 1 at top

    %   first handle the links: no expansion for links

    if(looking_at("/"))
    {
        %   this line is a link

        if (bfind("+ "))
        {
            call ("delete_char_cmd");
            insert ("-");
        }

        flush ("Directory links are not expanded!");
        return;
    }

    push_spot();

    %   variable 'sind' to take the string
    %   from bol to rightmost tree stroke

    bol();
    push_mark();
    goto_column (col-3);
    sind = bufsubstr();

    %   if line is a node, cursor is on rightmost tree stroke

    if (looking_at ("|+"))
    {
        sind += "|  ";
        call ("next_char_cmd");
        call ("delete_char_cmd");
        insert ("-");
    }

    else if (looking_at (tree_anglep))
    {
        sind += "   ";
        call ("next_char_cmd");
        call ("delete_char_cmd");
        insert ("-");
    }

    else
    {
        pop_spot();
        return;
    }

    tree_point();
    s = tree_node_path();

    variable tbr;

    %   prepare for insertion of branch
    eol();
    newline();
    tbr = tree_branch (s, sind);

    %   if no branch, delete empty line
    if (tbr<1)
    {
        if (is_whiteline and what_char==10) call ("delete_char_cmd");
    }

    pop_spot();
}


define tree_expand_all()
{
    EXIT_BLOCK set_readonly (1);

    set_readonly (0);

    variable m=0, s=Null_String;

    s = tree_node_path();
    if (s=="/")
    {
        flush("Too big, do it in portions.");
        return;
    }

    if (what_line() > 1)
    {
        bol();
        if (ffind("|- ") or ffind("`- "))
        {
            tree_collapse();
            %   recursive call!
            tree_expand_all();
            return;
        }
    }

    %   delimiter for expansion needed
    s = Null_String;
    if (what_line() == 1);
    else if (down (1))
    {
        s = tree_node_path();
        go_up(1);
    }

    variable lc = what_line();

    forever
    {
        tree_expand_one();

        %   buffer overflows for deep branches
        %   need to pop stuff pushed in probably by stat/lstat
        pop ();

        if (m/100 > (m-1)/100)
            flush (" Please wait, line " + string(m));

        m++;

        !if (down (1)) break;

        if(is_whiteline or eobp) break;

        %   break out using s as delimiter
        if(s!=Null_String)
        {
            if (is_substr(tree_node_path(),s)) break;
        }
    }

    update_sans_update_hook (0);

    goto_line(lc);
    tree_point();
}


define tree_collapse()
{
    EXIT_BLOCK set_readonly (1);
    set_readonly (0);

    if (what_line() == 1)
    {
        variable llstr  = line_as_string();
        variable offset = string_match (llstr, "[^/]*[/]$", 1);
        variable subdir = substr (llstr, offset, strlen(llstr));

        tree_parent_show();
        () = fsearch ("+ "+subdir);
        return;
    }

    %   line is not 1

    tree_point();

    if (looking_at("/"))
    {
        %   symlink
        tree_parent();
        return;
    }

    go_left(3);
    if (looking_at("|+ ") or looking_at ("`+ "))
    {
        tree_parent();
        return;
    }

    eol();
    push_mark();
    %   mark must be at end of line for deleting

    %   move down to bottom of branch as displayed
    tree_descend();

    eol();
    del_region();
    pop_mark (0);

    %   back to line where collapse command was issued

    bol();
    !if(re_fsearch ("[|`]+[-]+[ ]+"))
    {
        flush("Exception Error!");
        return;
    }

    go_right(1);
    call ("delete_char_cmd");
    insert ("+");

    tree_point();
}


define tree_home()
{
    bob();
}


define tree_end()
{
    eob();
    %while(is_whiteline) go_up(1);
    tree_point();
}


define tree_list_keys ()
{
    EXIT_BLOCK set_readonly (1);
    set_readonly (0);

    erase_buffer();
    bob();

    newline();
    dump_bindings ("tree_slimkmap");

    readable_keys();

    bob();
    push_mark ();
    eob();
    bskip_chars(" \t\n");

    sort2 ();

    while (is_whiteline) call("delete_char_cmd");
    %   this sort is not case sensitive

    bob();
    set_buffer_modified_flag (0);

    set_status_line("  Press any key to exit  ",0);

    variable m = "&Keybindings TreeView";
    dummy_menu(m);

    update(0);
    flush_input(getkey());

    menu_use_menu_bar (TreeView);
    tree_show (tree_cwd);
}


define tree_find()
{
    variable sstr=Null_String;
    sstr = read_mini2 ("Search for:", Null_String,Null_String);
    !if (strlen (sstr)) return;
    LAST_SEARCH=sstr;
    () = fsearch(sstr);
}


define tree_find_again ()
{
    %   since the line and not the string is highlighted
    %   the string may occur more than once in same line
    %   and it is hadr to notice

    if (right(1)) ()=fsearch(LAST_SEARCH);
}


%% MAKE_CSTRUCT_INT_FIELD(Stat_Type, st.st_dev, "st_dev", 0),
%% MAKE_CSTRUCT_INT_FIELD(Stat_Type, st.st_ino, "st_ino", 0),
%% MAKE_CSTRUCT_INT_FIELD(Stat_Type, st.st_mode, "st_mode", 0),
%% MAKE_CSTRUCT_INT_FIELD(Stat_Type, st.st_nlink, "st_nlink", 0),
%% MAKE_CSTRUCT_INT_FIELD(Stat_Type, st.st_uid, "st_uid", 0),
%% MAKE_CSTRUCT_INT_FIELD(Stat_Type, st.st_gid, "st_gid", 0),
%% MAKE_CSTRUCT_INT_FIELD(Stat_Type, st.st_rdev, "st_rdev", 0),
%% MAKE_CSTRUCT_INT_FIELD(Stat_Type, st.st_size, "st_size", 0),
%% MAKE_CSTRUCT_INT_FIELD(Stat_Type, st.st_atime, "st_atime", 0),
%% MAKE_CSTRUCT_INT_FIELD(Stat_Type, st.st_mtime, "st_mtime", 0),
%% MAKE_CSTRUCT_INT_FIELD(Stat_Type, st.st_ctime, "st_ctime", 0),
%% MAKE_CSTRUCT_INT_FIELD(Stat_Type, st_opt_attrs, "st_opt_attrs", 0),
