%*********************************************************
%
%   site.sl
%
%   This file must be present in $JED_ROOT/lib. JED loads it
%   first - even before reading command line arguments. The
%   command line arguments are then passed to a hook declared
%   in this file for further processing.
%
%   In addition to some hooks, this file defines utility
%   functions. User specific stuff should be placed in
%   the user startup file (sue.rc).
%
%   At the end of this file, basic functions are loaded
%   or initialized.
%
%*********************************************************

   _traceback = 1;
   _debug_info = 1;

%*********************************************************
%   Prototypes
%*********************************************************

define backtab();
define back_to_prev_position();
define begin_of_word_left();
define begin_of_word_right();
define buffer_filename();
define buffer_modified();
define call_function();
define check_buffer_filename(fullpfile);
define close_unconditionally();
define command_line_hook();
define del_line();
define del_to_eol();
define dircat(dir, file);
define disk_new_file();
define disk_save_as();
define disk_save_buffer();
define disk_save_permission(ffile);
define dummy_menu(ttitle);
define duplicate_ceiling_char();
define duplicate_floor_char();
define duplicate_line();
define end_of_trail();
define expand_jedlib_file(f);
define find_jedlib_file(file);
define goto_bottom_of_window();
define goto_middle_of_window();
define goto_top_of_window();
define go_down();
define go_left();
define go_right();
define go_up();
define insert_code_for_key();
define is_overwrite_mode();
define is_readonly();
define is_whiteline();
define jed_startup_hook();
define keyname(ks,modf);
define line_as_string();
define make_autosave_filename(dir, file);
define make_backup_filename(dir, file);
define menu_hlp();
define nada();
define newline();
define no_mode();
define parse_filename(fn);
static define patch_cmdline_file(file);
define readable_keys();
define read_mini1(prompt, dflt, inistr);
define read_mini2(prompt, dflt, inistr);
define read_path_from_mini(p);
define rename_buffer(name);
define require();
define require0(ffile);
define reset_screen();
define reset_slim_keymap();
define runhooks();
define scroll_down();
define scroll_up();
define search_path_for_file(path, file);
define select_word();
define set_buffer_modified_flag();
define set_buffer_no_autosave();
define set_buffer_no_backup();
define set_buffer_undo();
define set_color();
define set_color_scheme(scheme);
define set_overwrite();
define set_readonly();
define set_sue_palette();
define shape_cursor0();
define shape_cursor1();
define slcancel();
define smart_tabs();
define spaces_for_tabs();
define strcmpi(a, b);
define str_replace_all(sstr, old, new);
define str_split(sstr, n);
define sue_defaults();
define toggle_crmode();
define toggle_overwrite();
define toggle_readonly();
define toggle_undo();
define toggle_wrap();
define to_buffer(whichbf);
define trim_lines();
define vinsert();
define what_block();
define wordset();
define write_scratch(ssttrr);
define _set_buffer_flag(x);
define _set_or_unset_buffer_flag();
define _test_buffer_flag(x);
define _toggle_buffer_flag(f);
define _unset_buffer_flag(x);


%   from other modules, prototype here to be global
define display_help ()              ;
define display_hlp2 ()              ;
define dls ()                       ;


%*********************************************************
%   Global Variables
%*********************************************************


variable Default_Jedrc_Startup_File = "sue.rc";
variable _Jed_Color_Scheme = "sue_color";
variable _Jed_Default_Color_Scheme = "sue_color";
variable DirListing  = "/DirListing/";
variable slash = "/";
variable Compile_Output_Buffer = "/compiler-output/";
variable red = Assoc_Type [Integer_Type];
variable r_dir = Assoc_Type [Integer_Type];
%   used for recently edited files and accessed directories

variable Null_String        = "";
variable LASTDIR            = Null_String;
variable Jed_Home_Directory = Null_String;
variable Color_Scheme_Path  = Null_String;
variable LAST_SEARCH        = Null_String;
variable o_ListBuffer       = Null_String;
variable f_ListBuffer       = Null_String;
variable WordSet            = Null_String;
variable InsBuffer          = Null_String;


variable SEARCH_TYPE = 0;
variable TOGGLEWRAP = 0;
variable Line_Mark;
public variable mS = 0;
public variable globl_str = Null_String;
public variable globl_int = 0;
public variable AddSpace  = "AddOn";

variable c_kmap = "ctrl_kmap";
%   global keymap when <ctrl> is pressed
%   however, crtl-a to ctrl-z are not here
%   as low ascii, they are in the unshifted global map

variable s_kmap     = "shift_kmap";
variable s_c_kmap   = "shiftctrl_kmap";
variable e_kmap     = "esc_kmap";
variable a_kmap     = "alt_kmap";


%   strings same as in Linux console keymap

variable Key_BS     = _Backspace_Key;
variable Key_Enter  = "^M";
variable Key_Space  = " ";
variable Key_Tab    = "\t";

variable Key_F1     = "^[[[A";
variable Key_F2     = "^[[[B";
variable Key_F3     = "^[[[C";
variable Key_F4     = "^[[[D";
variable Key_F5     = "^[[[E";
variable Key_F6     = "^[[17~";
variable Key_F7     = "^[[18~";
variable Key_F8     = "^[[19~";
variable Key_F9     = "^[[20~";
variable Key_F10    = "^[[21~";
variable Key_F11    = "^[[23~";
variable Key_F12    = "^[[24~";

variable Key_Up     = "^[[A";
variable Key_Down   = "^[[B";
variable Key_Right  = "^[[C";
variable Key_Left   = "^[[D";

variable Key_Home   = "^[[1~";
variable Key_Ins    = "^[[2~";
variable Key_Del    = "^[[3~";
variable Key_End    = "^[[4~";
variable Key_PgUp   = "^[[5~";
variable Key_PgDn   = "^[[6~";

%   some custom key names
variable Key_Ctrl_K_Ins = strcat("^K", Key_Ins);
variable Key_Ctrl_K_Del = strcat("^K", Key_Del);
variable Key_Ctrl_K_BS  = strcat("^K", Key_BS);

variable greykeys = String_Type [27,2];
greykeys [*,0] =
    [
        Key_F1   ,
        Key_F2   ,
        Key_F3   ,
        Key_F4   ,
        Key_F5   ,
        Key_F6   ,
        Key_F7   ,
        Key_F8   ,
        Key_F9   ,
        Key_F10  ,
        Key_F11  ,
        Key_F12  ,

        Key_Up   ,
        Key_Down ,
        Key_Right,
        Key_Left ,

        Key_Home ,
        Key_Ins  ,
        Key_Del  ,
        Key_End  ,
        Key_PgUp ,
        Key_PgDn ,
        Key_BS   ,
        Key_Enter,
        Key_Space,
        Key_Tab  ,


        Key_Ctrl_K_Ins
    ];

greykeys [*,1]=
    [
        "F1",
        "F2",
        "F3",
        "F4",
        "F5",
        "F6",
        "F7",
        "F8",
        "F9",
        "F10",
        "F11",
        "F12",

        "Up",
        "Down",
        "Right",
        "Left",

        "Home",
        "Ins",
        "Del",
        "End",
        "PgUp",
        "PgDn",
        _Backspace_Key,
        "Enter",
        "Space",
        "Tab",
        "^K Insert"
    ];



%   curly, round, square brackets used when matching a fence
%   they are all in ASCII range, normally not affected by fonts

variable curlyL     = "\173";
variable curlyR     = "\175";
variable roundL     = "\050";
variable roundR     = "\051";
variable squareL    = "\133";
variable squareR    = "\135";


%*********************************************************
%   Jed Directories
%*********************************************************

%   Jed_Home_Directory specifies the user's directory
%   where personal jed-related files are to be found.
%   Normally, this corresponds to the user's home directory
%   unless the user has specified an alternate directory
%   via the JED_HOME environment variable.

Jed_Home_Directory = getenv ("JED_HOME");
if (Jed_Home_Directory == NULL)
    Jed_Home_Directory = getenv ("HOME");

variable RecentFiles = Jed_Home_Directory + "/recently_edited.txt";
variable RecentDirs = Jed_Home_Directory + "/recently_accessed.txt";
Color_Scheme_Path = strcat (JED_ROOT, "/lib/colors");

set_jed_library_path (strcat (get_jed_library_path (), ",."));

%*********************************************************
%   functions to handle directories/files
%*********************************************************


define dircat (dir, file)
%----------------------------------------------------------
%   Usage:
%   String_Type dircat (String_Type D, String_Type F)
%
%   This function provides the clean pathname of file F
%   in directory D.
%----------------------------------------------------------
{
    %   Many functions can handle dir = NULL,
    %   e.g., dircat (getenv (...));

    if (dir == NULL) dir = "";
    if (file == NULL) file = "";

    variable n = strlen (dir);

    if (n)
    {
        %variable slash = "/";
        if (dir[-1] != '/') dir += slash;
    }

    expand_filename (dir + file);
}


define buffer_filename ()
%---------------------------------------------------------
%   Usage:
%   string buffer_filename ()
%
%   Returns the name of the file associated with the current
%   buffer. If there is none associated with it, the empty
%   string is returned.
%----------------------------------------------------------
{
    variable file, dir;

    (file, dir,,) = getbuf_info ();

    !if (strlen (file)) return Null_String;
    %   it happens

    return (dir + file);
}


define check_buffer_filename (fullpfile)
%---------------------------------------------------------
%   Needs full path as string parameter.
%   Returns 1 if any buffer is associated with it,
%   or else 0.
%   Used to avoid different buffers for same disk file.
%----------------------------------------------------------
{
    variable ffile, ddir;

    (ddir, ffile) = parse_filename (fullpfile);

    !if (bufferp(ffile)) return 0;

    variable tmpbf=whatbuf (), r=0;

    sw2buf(ffile);
    if (fullpfile==buffer_filename ()) r = 1;
    sw2buf(tmpbf);

    return r;
    %   if 1, a buffer is already associated to the file
}


define search_path_for_file (path, file)
%----------------------------------------------------------
%   Usage:
%   string search_path_for_file (string P, string F)
%
%   The search_path_for_file function searches the
%   directories specified by the comma separated set of
%   directories P for the filename F. If the
%   file exists, it returns the expanded filename, otherwise
%   it returns NULL.
%----------------------------------------------------------
{
    variable dir;
    foreach (strtok (path, ","))
    {
        dir = ();
        file = dircat (dir, file);

        if (file_status (file) == 1) return file;
    }
    return NULL;
}


define expand_jedlib_file (f)
%----------------------------------------------------------
%   Search for file f in jed library directory and return
%   expanded pathname if found or the Null string otherwise.
%----------------------------------------------------------
{
   f = search_path_for_file (get_jed_library_path (), f);
   if (f == NULL) return "";
   return f;
}


define find_jedlib_file (file)
%----------------------------------------------------------
%   Find a file from JED_LIBRARY,
%   returns number of lines read, 0 if not found.
%----------------------------------------------------------
{
    file = expand_jedlib_file (file);
    !if (strlen (file)) return 0;
    () = find_file (file);
}


define parse_filename (fn)
%----------------------------------------------------------
%   Usage:
%   (dir, file) = parse_filename (fn)
%
%   This function breaks a file specification into dir and
%   filename. For example, /a/b/c returns /a/b/ (dir) and
%   c (file).
%----------------------------------------------------------
{
   return (path_dirname (fn), path_basename (fn));
}


define read_mini1 (prompt, dflt, inistr)
{
    mS = 1;
    %   set the history store

    if
    (
        _NARGS!=3
        or typeof (prompt) != String_Type
        or typeof (dflt)   != String_Type
        or typeof (inistr) != String_Type
    )
        return Null_String;

    variable rply = Null_String;

    _clear_error();
    flush_input();
    %   just in case

    rply = read_mini (prompt, dflt, inistr);

    if (LASTKEY[0] == 23)
    {
        pop ();
        %   anomaly, probably from escape key

        rply = Null_String;
    }

    %   set proper reply when key pressed was ^w, ascii 23,
    %   or escape.
    %   seems superfluous but there is an elusive bug
    %   using the minibuffer repeatedly in quick sequence
    %   this is a workaround, the bug was not found

    return rply;
}


define read_mini2 (prompt, dflt, inistr)
{
    mS = 0;
    %   set the history store

    ERROR_BLOCK
    {
        _clear_error();
        flush_input();
        %   just in case
    }

    if
    (
        _NARGS!=3
        or typeof (prompt) != String_Type
        or typeof (dflt)   != String_Type
        or typeof (inistr) != String_Type
    )
        return Null_String;

    variable rply = Null_String;

    rply = read_mini (prompt, dflt, inistr);

    if (LASTKEY[0] == 23)
    {
        pop ();
        %   anomaly, probably from escape key

        rply = Null_String;
    }

    %   set proper reply when key pressed was ^w, ascii 23,
    %   or escape.
    %   seems superfluous but there is an elusive bug
    %   using the minibuffer repeatedly in quick sequence
    %   this is a workaround, the bug was not found

    return rply;
}


define read_path_from_mini (p)
%----------------------------------------------------------
%   usage
%   string read_path_from_mini (string p)
%
%   Prompts for a file name using the parameter as a prompt.
%   Returns the filename read from the minibuffer. It can
%   also been used to return a directory.
%----------------------------------------------------------
{
    mS=1;
    %   history store

    variable thisdir = Null_String;

    (,thisdir,,) = getbuf_info ();

    return read_mini1 (p, Null_String, thisdir);
}


define rename_buffer (name)
{
    variable flags = getbuf_info ();
    pop ();
    setbuf_info (name, flags);
}


static define patch_cmdline_file (file)
{
   variable ch = file[0];
   file = strcompress (file, "/");
   if (ch == '/') file = strcat ("/", file);
   return file;
}



%*********************************************************
%   functions to set/unset/check flags
%*********************************************************


define _test_buffer_flag (x)
%----------------------------------------------------------
%   if flag x set:  returns 1
%   else            returns 0
%
%   values for hex integer x are defined in getbuf_info
%----------------------------------------------------------
{
    variable flags;

    (,,,flags) = getbuf_info ();
    return flags & x;
}


define _set_buffer_flag (x)
%----------------------------------------------------------
%   values for hex integer x are defined in getbuf_info
%----------------------------------------------------------
{
    getbuf_info ();
    () | x;
    setbuf_info (());
    %    parameter is hexadecimal, e.g. 0x10 for overwrite mode
}


define _unset_buffer_flag (x)
%----------------------------------------------------------
%   values for hex integer x are defined in getbuf_info
%----------------------------------------------------------
{
    getbuf_info ();
    () & ~x;
    setbuf_info (());
}


define _set_or_unset_buffer_flag ()
%----------------------------------------------------------
%   usage: _set_or_unset_buffer_flag (set, flag)
%   argument set is 1 for setting, 0 for unsetting
%   values for hex integer flag are defined in getbuf_info
%----------------------------------------------------------
{
    exch ();
    %  (set,flag) ==> (flag,set)

    if (()) _set_buffer_flag (());
    else _unset_buffer_flag (());
}


define _toggle_buffer_flag (f)
%----------------------------------------------------------
%   values for hex integer f are defined in getbuf_info
%----------------------------------------------------------
{
    setbuf_info (getbuf_info () xor f);
}


define set_buffer_modified_flag ()
%----------------------------------------------------------
%   If argument what is 1, consider buffer modified.
%   If argument what is 0, consider buffer unchanged.
%----------------------------------------------------------
{
    _set_or_unset_buffer_flag ((), 0x1);
}


define buffer_modified ()
%----------------------------------------------------------
%   Usage:
%   Int_Type buffer_modified ()
%
%   returns non-zero if the buffer modified flag
%   is set. It returns zero if the buffer modified flag
%   is not been set. This works on the current buffer.
%----------------------------------------------------------
{
   _test_buffer_flag (0x01);
}


define set_buffer_undo ()
%----------------------------------------------------------
% set undo mode for buffer.
% If argument is 1, undo is on.  0 turns it off
%----------------------------------------------------------
{
   _set_or_unset_buffer_flag ((), 0x20);
}


define set_readonly ()
%----------------------------------------------------------
%   Usage:
%   void set_readonly (int what)
%
%   0   turn readonly off
%   1   turn readonly on
%----------------------------------------------------------
{
   _set_or_unset_buffer_flag ((), 0x08);
}


define is_readonly ()
%----------------------------------------------------------
%   This function returns a non-zero value if the buffer
%   is read-only; otherwise it returns 0.
%----------------------------------------------------------
{
   _test_buffer_flag (0x08);
}


define is_overwrite_mode ()
%----------------------------------------------------------
%   usage
%   Int_Type is_overwrite_mode ()
%
%   This function returns a non-zero value
%   if the buffer is in overwrite-mode;
%   otherwise it returns 0.
%----------------------------------------------------------
{
    _test_buffer_flag (0x10);
}


define set_overwrite ()
%----------------------------------------------------------
%   set_overwrite (Int_Type x)
%   If the parameter is non-zero, the buffer will be
%   put in overwrite mode; otherwise it will be put in insert
%   mode.
%----------------------------------------------------------
{
   _set_or_unset_buffer_flag ((), 0x10);
}


define toggle_crmode ()
%----------------------------------------------------------
%   Toggles end-of-line mark between DOS/UNIX
%   DOS  mark is CRLF
%   UNIX mark is LF
%----------------------------------------------------------
{
    if (is_readonly) return;

   _toggle_buffer_flag (0x400);
   set_buffer_modified_flag (1);
}


define toggle_readonly ()
%----------------------------------------------------------
%   This function toggles the read-only status of the current
%   buffer.
%----------------------------------------------------------
{
   _toggle_buffer_flag (0x08);
}


define toggle_overwrite ()
%----------------------------------------------------------
%   This function toggles the overwrite mode of the current
%   buffer.
%----------------------------------------------------------
{
   _toggle_buffer_flag (0x10);
}


define toggle_undo ()
%----------------------------------------------------------
%   This function toggles the undo mode of the current buffer.
%----------------------------------------------------------
{
   _toggle_buffer_flag (0x20);
}


define toggle_wrap ()
%----------------------------------------------------------
%   This function toggles the wrap mode in current buffer
%   and sets the wrap column after prompting for it.
%----------------------------------------------------------
{
    if (is_readonly) return;
    if (check_rectangle) return;

    EXIT_BLOCK
        flush("Wrap column is " + string(WRAP));

    if(TOGGLEWRAP == 1) TOGGLEWRAP=0;
    else TOGGLEWRAP=1;
    variable str_wrap="-";
    if (TOGGLEWRAP) str_wrap = "w";
    set_mode (str_wrap,TOGGLEWRAP);
    !if(TOGGLEWRAP) return;

    variable tmpwrap=Null_String;
    tmpwrap = read_mini2 ("Wrap at column:","",string(WRAP));

    variable iWRAP = integer(tmpwrap);

    if(iWRAP > 20) WRAP=iWRAP;
    else WRAP=70;
}


define set_buffer_no_backup ()
{
   _set_buffer_flag (0x100);
}


define set_buffer_no_autosave ()
{
   _unset_buffer_flag (0x02);
}



%*********************************************************
%   positioning, inserting, deleting
%*********************************************************


define end_of_trail ()
{
    eol ();
}


define go_left ()
{
   () = left ();
}


define go_right ()
{
   () = right ();
}


define go_up ()
{
   () = up ();
}


define go_down ()
{
   () = down ();
}


define back_to_prev_position ()
%----------------------------------------------------------
%   go back to Previous Cursor Position
%   when switching buffers previous position is (1,1)
%----------------------------------------------------------
{
    pcp ();
}


define newline ()
{
    insert_char ('\n');
}


define del_to_eol ()
{
    if (markp) return;
    push_mark ();
    eol ();
    if (markp) del_region ();
}


define del_line ()
{
    if (markp) return;
    bol();
    delete_line ();

    if (eobp)
    {
        %   nothing was deleted, just give the impression
        go_up(1);
        bol();
    }
}


define close_unconditionally ()
{
    call("close_buffer");
}


define trim_lines ()
%----------------------------------------------------------
%   trailing spaces are trimmed line by line
%   empty lines are not deleted
%----------------------------------------------------------
{
    push_spot ();

    bob ();
    do
    {
        call("eol_cmd");
        %   does the trimming on its own
    }
    while (down (1));

    pop_spot ();
}


define to_buffer (whichbf)
%----------------------------------------------------------
%   switches to buffer whichbf, argument is a string
%----------------------------------------------------------
{
    if (MINIBUFFER_ACTIVE) return;
    %   just in case

    variable n, buf;
    n = buffer_list ();

    %   get number of buffers on the stack
    if (whichbf) _stk_reverse (n-1);
    %   which is not zero, i.e. previous buffer
    %   0 will take to next buffer

    loop (n)
    {
        buf = ();
        %   buf is the buffer name (string variable)

        n--;
        if (buf[0] == ' ') continue;
        sw2buf (buf);
        _pop_n (n);
        return;
    }
}


define goto_top_of_window ()
%----------------------------------------------------------
%   Positions cursor at top of window
%----------------------------------------------------------
{
   go_up (window_line () - 1);
   bol ();
}


define goto_bottom_of_window ()
%----------------------------------------------------------
%   Positions cursor at bottom of window.
%----------------------------------------------------------
{
   go_down (window_info ('r') - window_line ());
}


define goto_middle_of_window ()
%----------------------------------------------------------
%   Positions cursor in the middle of window.
%   If there are too few lines, positions on bottom one.
%----------------------------------------------------------
{
    if (check_rectangle) return;
    goto_top_of_window ();
    go_down (window_info ('r') / 2);
}


define scroll_up ()
%------------------------------------------------
%   If cursor is not in the middle of window, this
%   function will center the current line. If current
%   line is already centered, the effect is scroll up
%   in place.
%   Returns nothing.
%------------------------------------------------
{
    variable lm = window_info ('r') / 2;
    variable cc = what_column ();


    if (window_line () != lm)
    {
        %   cursor not mid-window, move current line there
        recenter (lm);
        goto_column(cc);
        return;
    }

    %   cursor is already mid-window, scroll up
    if (down (1)) recenter (lm);
    goto_column(cc);
}


define scroll_down ()
%------------------------------------------------
%   If cursor is not in the middle of window, this
%   function will center the current line. If current
%   line is already centered, the effect is scroll
%   down in place. Returns nothing.
%------------------------------------------------
{
    variable lm = window_info ('r') / 2;
    variable cc = what_column ();


    if (window_line () != lm)
    {
        %   cursor not mid-window, move current line there
        recenter (lm);
        goto_column(cc);
        return;
    }

    %   cursor is already mid-window, scroll up
    if (up (1)) recenter (lm);
    goto_column(cc);
}


define spaces_for_tabs ()
%----------------------------------------------------------
%   Will insert spaces up to the next tab stop,
%   deternined by the variable TAB_DEFAULT counting from 1.
%----------------------------------------------------------
{
    if (TAB_DEFAULT < 1) TAB_DEFAULT=4;

    variable n = what_column ();
    %   variable n is used for multiple purposes

    n = (1 + (n - 1)/TAB_DEFAULT)*TAB_DEFAULT - (n - 1);
    %   number of spaces up to next tab stops
    %   tab stops are defined by TAB_DEFAULT starting at 1

    loop(n) insert_char(' ');
}


define backtab ()
{
    variable n,m;

    n = what_column ();
    m = (n-1) mod TAB_DEFAULT;
    if (m<=0) m += TAB_DEFAULT;

    %   tab stops are defined by TAB_DEFAULT starting at 1

    if (n > TAB_DEFAULT) go_left(m);
}


define smart_tabs ()
%----------------------------------------------------------
%   will insert spaces aligning with words one line up
%   cursor will jump and align with words in line above
%   here words include everyhting but whitespace
%
%   if screen is split, function takes to the other window!
%----------------------------------------------------------
{
    variable winf = window_info ('r');
    if(winf < 20)
    {
        otherwindow();
        return;
    }

    if (eobp) return;
    if (what_block==1) return;

    variable c0, c;

    push_spot ();
    c0 = what_column ();

    !if (up(1))
    {
        pop_spot ();
        return;
    }

    c = goto_column_best_try(c0);

    if (c < c0)
    {
        pop_spot ();
        return;
    }

    skip_white ();

    %   has cursor moved?
    if (c == what_column ())
    {
        %   cursor has not moved
        %   skip non-white, then skip white
        skip_chars ("^ \t");
        skip_white ();
    }

    %   if end of line reached, return
    if (eolp)
    {
        pop_spot ();
        return;
    }

    c = what_column ();
    pop_spot ();
    loop(c-c0) insert_char(' ');
    %   loop inactive for c<=c0
}


define wordset ()
{
    variable i;
    for (i=33;i<127;i++) WordSet += char(i);
    for (i=128;i<256;i++) WordSet += char(i);
    %   this word definition includes everything
    %   except backspace (127)

    define_word (WordSet);
}


define begin_of_word_left ()
%----------------------------------------------------------
%   to begin of current word, if cursor on word (past begin)
%   or to begin of word left otherwise
%----------------------------------------------------------
{
    variable p = what_column ();

    bskip_word_chars ();

    if (p != what_column ()) return;
    %   if cursor was on word or adjacent right
    %   it has moved to begin and we are finished

    %   cursor has not moved, so it was between words
    %   backskip delimiters
    bskip_non_word_chars ();

    % stuck at begin of line?

    %   do not wrap if there is a rectangle
    if (check_rectangle) return;

    if (bolp)
    {
        bskip_chars ("\n\t ");
        bskip_non_word_chars ();
    }

    bskip_word_chars ();
}


define begin_of_word_right ()
%----------------------------------------------------------
%   to begin of next word
%----------------------------------------------------------
{
    variable p = what_column ();

    skip_non_word_chars ();

    if (p != what_column ()) return;
    %   if cursor was between words,
    %   it has moved to begin of next word and we are finished

    %   cursor was not between words,
    %   so skip tokens then delimiters
    skip_word_chars ();
    skip_non_word_chars ();

    %   do not wrap if there is a rectangle
    if (check_rectangle) return;

    % stuck at end of line?
    if (eolp)
    {
        skip_chars ("\n\t ");
        skip_non_word_chars ();
    }
}


define select_word ()
%----------------------------------------------------------
%   If cursor on a word, select the word
%----------------------------------------------------------
{
    %   is cursor on word character?
    variable c_str = char(what_char ());
    !if (is_substr(WordSet,c_str)) return;

    %   cursor is on a word

    variable c_col = what_column ();

    skip_word_chars();

    if (what_column () == c_col)
    {
        %   cursor has not moved, end of buffer
        begin_of_word_left();
        push_mark();
        eol();
        return;
    }

    %   cursor has moved

    push_mark();
    call("set_mark_cmd");
    begin_of_word_left ();
}


define duplicate_floor_char ()
%----------------------------------------------------------
%   duplicates at current position the character below it
%----------------------------------------------------------
{
    if (is_readonly) return;

    push_spot ();
    %   save cursor position

    variable p = what_column ();

    !if (down (1))
    {
        pop_spot ();
        return;
    }
    %   cursor on bottom line, nothing to duplicate

    if (goto_column_best_try (p) != p)
    {
        pop_spot ();
        return;
    }
    %   cursor moved but not to same column
    %   target line is shorter, no char to duplicate

    if (eolp ())
    {
        pop_spot ();
        return;
    }
    %   cursor at end of line, nothing to duplicate

    %   if control here, normal case

    variable ch = what_char ();
    %   get the character to duplicate

    pop_spot ();
    %   go back to original position

    insert_char (ch);
    %   duplicate character from target line
    %   will work even in overwrite mode
}


define duplicate_ceiling_char ()
%----------------------------------------------------------
%   duplicates at current position the character above it
%----------------------------------------------------------
{
    if (is_readonly) return;

    push_spot ();
    %   save cursor position

    variable p = what_column ();

    if (up(1) < 1)
    {
        pop_spot ();
        return;
    }
    %   cursor was on top line, could not move up,
    %   nothing to duplicate

    if (goto_column_best_try (p) != p)
    {
        pop_spot ();
        return;
    }
    %   cursor moved but not in same column
    %   target line is shorter, no char to duplicate

    if (eolp ())
    {
        pop_spot ();
        return;
    }
    %   cursor is at end of line, nothing to duplicate

    %   if control here, normal case

    variable ch = what_char ();
    %   get the character to duplicate

    pop_spot ();
    %   go back to original position

    insert_char (ch);
    %   duplicate character from target line
    %   will work even in overwrite mode
}


define duplicate_line ()
%----------------------------------------------------------
%   inserts a duplicate of current line below it, cursor stays put
%----------------------------------------------------------
{
    if (is_readonly) return;
    if (what_block==1) return;

    push_spot ();
    bol ();
    push_mark ();
    eol ();
    bufsubstr ();

    newline ();
    insert(());
    pop_spot ();
}



%*********************************************************
%   strings
%*********************************************************


define str_replace_all (sstr, old, new)
%----------------------------------------------------------
%   Usage:
%   string str_replace_all (string sstr, string old, string new)
%
%   Replace all occurrences of old with new in sstr.
%   Return the modified string sstr.
%
%   This function works solely within a string.
%   It does not replace instances of a pattern across a buffer.
%----------------------------------------------------------
{
    (sstr,) = strreplace (sstr, old, new, strlen (sstr));
    return sstr;
}


define vinsert ()
%----------------------------------------------------------
%   Usage:
%   void vinsert (string, fmt,...);
%
%   This function will insert string formatted like in Sprintf
%   with a variable number of arguments.
%----------------------------------------------------------
{
    _NARGS-1;
    Sprintf;
    insert;
}


define line_as_string ()
%----------------------------------------------------------
%   Usage:
%   string line_as_string ()
%
%   This function returns the current line as a string.
%   This does not include the newline character at the end
%   of the line. The editing point is left at the end of
%   the line, that is, function does not preserve the point.
%----------------------------------------------------------
{
   bol ();
   push_mark ();
   eol ();
   bufsubstr ();
}


define str_split (sstr, n)
%----------------------------------------------------------
%   splits a string into two substrings, returns both
%   string and split position given by parameters
%----------------------------------------------------------
{
    variable sub1, sub2;
    sub1 = substr (sstr, 1, n - 1);
    sub2 = substr (sstr, n, -1);
    return (sub1, sub2);
    % so call it (sub1, sub2) = str_split(sstr, n);
}



%*********************************************************
%   sundry functions
%*********************************************************


define nada ()
{
    flush ("nada");
}


define slcancel ()
{
    call("cancell");
}


define reset_slim_keymap ()
%----------------------------------------------------------
%   slim_kmap is defined in keymap.c
%   and used repeatedly for different purposes
%   each time, it must be reset
%----------------------------------------------------------
{
    variable i;
    $1 = "slim_kmap";

    undefinekey (Key_F1, $1);

    %for (i=1; i<32; i++)
    for (i=1; i<27; i++)
        undefinekey (char(i), $1);

    for (i=32; i<256; i++)
        definekey ("self_insert_cmd", char(i), $1);

    definekey ("newline_and_indent",        Key_Enter, $1);
    definekey ("next_char_cmd",             Key_Right, $1);
    definekey ("previous_char_cmd",         Key_Left,  $1);
    definekey ("previous_line_cmd",         Key_Up,    $1);
    definekey ("next_line_cmd",             Key_Down,  $1);
    definekey ("bol",                       Key_Home,  $1);
    definekey ("eol",                       Key_End,   $1);
    definekey ("delete_char_cmd",           Key_Del,   $1);
    definekey ("backward_delete_char",      Key_BS,    $1);
    definekey ("write_buffer",              "^S",      $1);
    definekey ("close_buffer",              "^W",      $1);
    definekey ("kbd_quit",                  "^\\",     $1);
}


define what_block ()
%----------------------------------------------------------
%   determine char/column/line block
%----------------------------------------------------------
{
    return block_type ();

    %   0   default
    %           character stream    if shiftkey pressed
    %           no block            otherwise
    %   1   column
    %   2   line
}


define is_whiteline ()
%----------------------------------------------------------
%   return 1    line has only whitespace
%   return 0    otherwise
%----------------------------------------------------------
{
    push_spot ();
    bol ();
    skip_chars(" \t");
    variable r = eolp ();
    pop_spot ();
    return r;
}


define no_mode ()
%----------------------------------------------------------
%   Will reset to the default no_mode
%   after some other mode was in force.
%----------------------------------------------------------
{
    WANT_SYNTAX_HIGHLIGHT = 0;
    set_mode(Null_String, 0);
    use_keymap("global");
}


define set_sue_palette ()
%----------------------------------------------------------
%   sets palette to color defaults
%----------------------------------------------------------
{
    %   reset palette
    tt_send ("\033]R");

    variable sc = dircat (JED_ROOT, "lib/colors/sue_color.sl");
    require(sc);
}


define sue_defaults ()
%----------------------------------------------------------
%   Restores sue's defaults - namely:
%
%   colors          sue
%   mode            no_mode
%   functions keys  no keymap interference with modifiers
%----------------------------------------------------------
{
    no_mode ();
    set_sue_palette ();
    revise_keys();

    %   reset interpreter
    while (_stkdepth) pop();
    _clear_error();
}





define require0 (ffile)
%----------------------------------------------------------
%   Just a better name for evalfile but void, no return.
%   Call with one string as parameter,
%   ffile must be in jed's library directory or have full path.
%----------------------------------------------------------
{
    () = evalfile (ffile);
    %   example: require("madmode");
    %   madmode.sl would have to be in the library directory
}


define require ()
%----------------------------------------------------------
%   Just a better name for evalfile but void, no return.
%   Call with strings as arguments.
%   First argument is a file in jed's library directory
%   otherwise it must have full path.
%   Second argument is the namespace. If absent, Global
%   namespace will be used.
%   example: require("madmode");
%   madmode.sl would have to be in the library directory
%----------------------------------------------------------
{
    if (_NARGS == 0) return;

    variable ffile=Null_String, nnamespace=Null_String;

    variable args = __pop_args (_NARGS);
    %   this is an array, array elements are structures,
    %   these structures have only one field called 'value',
    %   this field is a string

    if (_NARGS == 1)
    {
        ffile = args [0].value;
        %   default namespace is "Global"
        () = evalfile (ffile);
        return;
    }

    if (_NARGS == 2)
    {
        ffile = args [0].value;
        nnamespace = args [1].value;
        () = evalfile (ffile,nnamespace);
    }
}


define write_scratch (ssttrr)
{
    variable from_bf = whatbuf ();
    sw2buf ("*scratch*");
    eob();
    newline ();
    newline ();
    insert (ssttrr);
    sw2buf (from_bf);
}


define set_color ()
{
    _set_color ();
}


define set_color_scheme (scheme)
%----------------------------------------------------------
%   Allows to pick up and set a color scheme from
%   the collection available in the Color_Scheme_Path
%----------------------------------------------------------
{
    variable file;
    if (scheme == NULL) return;
    scheme = string (scheme);
    %  for back-compatability, file may be an integer

    foreach (strtok (get_jed_library_path (), ","))
    {
        $1 = ();
        Color_Scheme_Path = dircat ($1, "colors");
        if (2 == file_status (Color_Scheme_Path)) break;
    }

    file = search_path_for_file (Color_Scheme_Path, scheme + ".sl");
    if (file == NULL)
    {
        % Try .slc file
        file = search_path_for_file (Color_Scheme_Path, scheme + ".slc");
        if (file == NULL)
        {
            vmessage ("Color scheme %S is not supported", scheme);
            return;
        }
    }
    %  strip .sl[c] to get pre-parsed version
    _Jed_Color_Scheme = path_sans_extname (file);
    require(_Jed_Color_Scheme);
}


define dummy_menu (ttitle)
%----------------------------------------------------------
%   uses the menu bar like a heading
%----------------------------------------------------------
{
    variable n = slash + string(_time) + slash;
    menu_create_menu_bar (n);
    menu_use_menu_bar (n);
    menu_delete_items (n);
    menu_append_popup (n, ttitle);
}


define keyname(ks,modf)
%----------------------------------------------------------
%   converts key string ks to usual key name
%----------------------------------------------------------
{
    variable i, kstring=Null_String;
    variable L=length(greykeys[*,0]);

    if(ks[0]==127 and strlen(ks)==1) return "Backspace";

    if(ks[0]==27 and strlen(ks)==1) return "Escape";

    else if(ks[0]==27 and strlen(ks)==2)
        return char(ks[1]);

    else if(ks[0]==27)
    {
        if (modf==8) ks = strcat("^" + substr(ks,2,-1));
        else ks = strcat("^[" + substr(ks,2,-1));
    }

    %   problem with enter and tab
    else if(ks[0] < 32 and strlen(ks)==1)
    {
        if (ks[0]==9 or ks[0]==13)
            return "Ctrl "+char(64+ks[0]);
        else return char(64+ks[0]);
    }

    else if(ks[0] > 31 and strlen(ks) == 1) return ks;

    for (i=0; i < L; i++)
    {
         !if (strcmp(greykeys[i,0],ks)) break;
    }

    if (i<L) kstring=greykeys[i,1];

    return kstring;
}


define readable_keys ()
%----------------------------------------------------------
%   This functions replaces the strings from dump_bindings
%   with readable names for the functions and grey keys
%   replacements are padded with spaces to ensure same length
%----------------------------------------------------------
{
    %   do not touch the spaces in the strings!
    %   they give to the replacements the same length
    %   as the original strings from dump_bindings

    bob ();
    replace     ("ESC [ A",     "Up     ");

    bob ();
    replace     ("ESC [ B",     "Down   ");

    bob ();
    replace     ("ESC [ C",     "Right  ");

    bob ();
    replace     ("ESC [ D",     "Left   ");

    bob ();
    replace     ("ESC [ 2 ~",   "Insert   ");

    bob ();
    replace     ("ESC [ 3 ~",   "Delete   ");

    bob ();
    replace     ("ESC [ 4 ~",   "End      ");

    bob ();
    replace     ("ESC [ 1 ~",   "Home     ");

    bob ();
    replace     ("ESC [ 5 ~",   "PgUp     ");

    bob ();
    replace     ("ESC [ 6 ~",   "PgDn     ");

    bob ();
    replace     ("ESC [ [ A",   "F1       ");

    bob ();
    replace     ("ESC [ [ B",   "F2       ");

    bob ();
    replace     ("ESC [ [ C",   "F3       ");

    bob ();
    replace     ("ESC [ [ D",   "F4       ");

    bob ();
    replace     ("ESC [ [ E",   "F5       ");

    bob ();
    replace     ("ESC [ 1 7 ~", "F6         ");

    bob ();
    replace     ("ESC [ 1 8 ~", "F7         ");

    bob ();
    replace     ("ESC [ 1 9 ~", "F8         ");

    bob ();
    replace     ("ESC [ 2 0 ~", "F9         ");

    bob ();
    replace     ("ESC [ 2 1 ~", "F10        ");

    bob ();
    replace     ("ESC [ 2 3 ~", "F11        ");

    bob ();
    replace     ("ESC [ 2 4 ~", "F12        ");

    bob ();
    replace     ("SPACE", "Space");

    bob ();
    replace     ("^M   ", "Enter");

    bob ();
    replace     ("ESC   ", "Escape");
}


define reset_screen ()
%----------------------------------------------------------
%   This function clears screen garbage brought in by
%   interfering applications not under editor's control,
%   e.g. commands issued to the shell. After which,
%   it redraws the editor's screen.
%----------------------------------------------------------
{
    () = fflush(stderr);
    clearerr(stderr);
    () = fflush(stdout);
    clearerr(stdout);
    call("redraw");
    sue_defaults ();
}


define insert_code_for_key ()
%----------------------------------------------------------
%   after activating quoted insert
%   the code of a grey key will be inserted
%   e.g.  pressing F1 will insert ^[[A
%   for self-inserting characters it's business as usual
%----------------------------------------------------------
{
    if (markp)
    {
        flush("Not allowed when block highlighted.");
        return;
    }

    variable alert =
    "Next key (even a command) will be inserted as a code";

    flush (alert);

    call ("quoted_insert");

    %   after activating quoted_insert
    %   the code of a grey key will be inserted
    %   e.g. pressing F1 will insert ^[[A
    %   or pressing Tab will insert a real tab
    %   for self-inserting characters it's business as usual
    %   for alt-key combos only ^[ will insert
    %   because the combo is not one code but a sequence of codes
    %   and only the first one (escape) is inserted
    %   however, alt with a code from keypad will insert properly
}


define strcmpi (a, b)
%----------------------------------------------------------
%   compares strings like strcmp, case-insensitive
%----------------------------------------------------------
{
  return strcmp (strup(a), strup(b));
}


define disk_new_file ()
%----------------------------------------------------------
%   Open new buffer/file with random name
%----------------------------------------------------------
{
    variable
        strtime = string(_time),
        fn = Null_String,
        thisdir = Null_String;

    (,thisdir,,) = getbuf_info ();
    fn = strcat (strtime, ".new");
    fn = strcat (thisdir, fn);
    mS=1;

    forever
    {
        %   a name is generated seeded from current time
        %   but user can confirm or change it
        %   name is only accepted if no file with same name exists
        %   program will loop until unique and valid name is entered

        fn = read_mini1
        ("Name new file:", Null_String, fn);

        if (LASTKEY[0] == 23)
            return;

        if (file_status (fn) == 2) return;
        %   use to escape query, just wipe out prompt

        if (file_status (fn) == 0) break;
        %   name is ok, proceed

        message ("Already open, try a different name.");
    }

    () = find_file (fn);
    %   new buffer is opened

    sue_defaults ();
}


define disk_save_permission (ffile)
%----------------------------------------------------------
%   Will check if a file is writable by current user
%   returning 1 for yes, 0 for no
%----------------------------------------------------------
{
    if (ffile==Null_String) return 0;

    variable chk, st, perms;

    st = stat_file (ffile);

    if(NULL==st) return 1;
    %   we don't know why, just go ahead and try

    %   convert mode to octal
    perms = sprintf("%o",st.st_mode);
    %   get current user permissions
    perms = substr(perms,4,1);
    chk = integer(perms);
    return (chk==2 or chk==3 or chk==6 or chk==7);
}


define disk_save_as ()
%----------------------------------------------------------
%   current buffer is saved prompting for a file name
%----------------------------------------------------------
{
    variable f_name = Null_String;
    f_name = read_path_from_mini ("Save file as:");

    %   evasive bug in read_file_from_mini
    %   even when pressing escape,
    %   the string returned is not empty
    if (LASTKEY[0] == 23) f_name = Null_String;
    %   discard reply for ^w, ascii 23, or escape

    !if (strlen(f_name)) return;

    %   check file status before proceeding

    if (check_buffer_filename(f_name))
    {
        variable refusl =
        "File is already associated to a buffer, choose different name.";
        flush (refusl);
        return;
    }

    variable f_status = file_status (f_name);

    if (f_status == 2);

    else if (f_status == -1)
        error ("File exists and is not accessible");

    else if (f_status == -2)
        error ("Invalid path");

    else if (f_status == -3)
        error ("Undetermined error");

    else if (f_status == 0)
        write_buffer (f_name);

    else if (f_status == 1)
    {
        !if(disk_save_permission (f_name))
        {
            flush("No permission to write to file: " + f_name);
            return;
        }

        if (1 == get_y_or_n ("File exists: overwrite"))
            write_buffer (f_name);
    }


    if (f_status == 0 or f_status == 1) sue_defaults ();
}


define disk_save_buffer ()
%----------------------------------------------------------
%   current buffer is saved if modified, no questions asked
%----------------------------------------------------------
{
    !if (buffer_modified) return;

    variable tmpblk = what_block ();

    variable file = buffer_filename ();

    !if (strlen (file))
    {
        disk_save_as ();
        return;
    }

    !if(disk_save_permission (file))
    {
        flush("No permission to write to file: " + file);
        return;
    }

    trim_lines ();
    () = write_buffer (file);
    update_sans_update_hook (1);
    %   saving while line block present
    %   turns the highlight into a stream
    %   in that case, restore the lines
    if (tmpblk==2) stream_to_lines ();
}


%*********************************************************
%   hooks
%*********************************************************

define menu_hlp ()
%----------------------------------------------------------
%   DON'T TOUCH!
%   This function is called from menu.c to display
%   help for menu items. It is activated pressing F1
%   on a menu entry.
%----------------------------------------------------------
{
    %   get the goodies from stack
    variable helpkey0 = ();
    variable tmpnm0 =  ();

    variable helpkey = helpkey0 + " [#]";
    variable tmpnm =  tmpnm0 + " [#]";

    %   prepare the buffer for displaying help
    sw2buf(" /MenuHelp/");
    erase_buffer();
    EXIT_BLOCK delbuf(" /MenuHelp/");

    %   insert help file
    if (1 > insert_file(strcat (JED_ROOT,"/lib/sue.hlp")))
    {
        flush("Help file not found!");
        return;
    }
    bob();

    %   search for proxy keyword first, real search follows
    !if (bol_fsearch(tmpnm))
    {
        flush("Not found: " + tmpnm0);
        return;
    }

    %   search for the real keyword now
    !if (bol_fsearch(helpkey))
        flush ("Not found: " + helpkey0);

    else
    {
        %   search key found

        go_up(1);
        bol();
        push_mark();
        bob();
        del_region();

        %   text above keyword deleted
        %   search for delimiter downstream

        ()=bol_fsearch(helpkey);
        () = ffind ("[#]");
        del_to_eol();
        go_down(5);
        if (fsearch("[#]"))
        {
            bol();
            push_mark();
            eob();
            del_region();
        }

        %   display help text at line 2
        () = bol_bsearch(helpkey0);
        recenter(2);
        eol();

        %   we insert now a description of the keybinding
        %   for the menu entry, if any.
        %   heavy crap from menu.c, get_key4node

        goto_column(40);
        insert ("current keybinding: ");

        if(KEY4NODE==Null_String)
        {
            variable nada = "none";
            push_spot();
            if(bol_bsearch("Cancel Selection"))
                nada = "Escape";
            pop_spot();
            insert (nada);
        }

        else
        {
            variable k_describe = Null_String;

            if(MODIFIER==0)      k_describe = " ";
            else if(MODIFIER==1) k_describe = "Shift ";
            else if(MODIFIER==4) k_describe = "Ctrl ";
            else if(MODIFIER==5) k_describe = "ShiftCtrl ";
            else if(MODIFIER==8) k_describe = "Alt ";

            %   really, have a look at get_key4node
            if (KLUDGE and MODIFIER)
            {
                if(KLUDGE < 32) k_describe += char(KLUDGE+64);
                else k_describe += char(KLUDGE);
                if(KLUDGE==9) k_describe = "Ctrl Tab";
            }

            else if (KLUDGE and MODIFIER==0)
            {
                k_describe += "^";
                k_describe += char(KLUDGE+64);
                if(KLUDGE==127) k_describe = "Backspace";
                else if(KLUDGE==13) k_describe = "Enter";
                else if(KLUDGE==9) k_describe = "Tab";
                else if(KLUDGE==32) k_describe = "Space";
                else if(KLUDGE==42) k_describe = "*";
            }

            else
            {
                variable i, kstring=KEY4NODE;
                variable L=length(greykeys[*,0]);

                for (i=0; i < L; i++)
                {
                     !if (strcmp(greykeys[i,0],KEY4NODE)) break;
                }
                if (i<L) kstring=greykeys[i,1];

                %k_describe += KEY4NODE;
                k_describe += kstring;
            }

            insert(k_describe);

            %   if (strlen(KEY4NODE) < 33) insert(KEY4NODE);
            %   junk
        }

        %   reduce status to hline
        set_status_line("  Press any key to exit  ", 0);
        set_readonly(1);

        %   update and wait
        update(0);
        flush_input(getkey());
        flush(" ");
    }
}


define make_backup_filename (dir, file)
%----------------------------------------------------------
%   returns backup filename
%   DON'T TOUCH, this function is run as a hook from file.c
%----------------------------------------------------------
{
    if (dir == "/tmp/") return "";
    strcat (dir, file, "~");
}


define make_autosave_filename (dir, file)
%----------------------------------------------------------
%   returns autosave filename
%   DON'T TOUCH, this function is run as a hook from file.c
%----------------------------------------------------------
{
     sprintf ("%s#%s#", dir, file);
}


define call_function ()
%----------------------------------------------------------
%   Usage: call_function (String_Type f, [optional args...])
%----------------------------------------------------------
{
    variable args = __pop_args (_NARGS - 1);
    variable func = ();
    if (typeof (func) != Ref_Type) func = __get_reference (func);
    if (func != NULL) (@func) (__push_args (args));
}


define runhooks ()
%----------------------------------------------------------
%   void runhooks (string hook, [optional args...])
%----------------------------------------------------------
{
   variable args = __pop_args (_NARGS - 1);
   variable hook = ();
   hook = __get_reference (hook);
   if (hook != NULL) @hook (__push_args (args));
}


define jed_startup_hook ()
%----------------------------------------------------------
%   Function that gets executed right before JED
%   enters its main editing loop. This is for last minute
%   modifications of data structures that did not exist
%   when startup files were loaded.
%----------------------------------------------------------
{
    variable n, hlp, ok = 0, file;
    variable scratch = "*scratch*";

    % turn on Abort character processing
    IGNORE_USER_ABORT = 0;

    runhooks ("startup_hook");

    ERROR_BLOCK eval (". ()jed_startup_hook");

    !if (strcmp(whatbuf (), scratch) or buffer_modified ())
    {
        ERROR_BLOCK
        {
            setbuf (scratch);
            erase_buffer ();
            set_buffer_modified_flag (0);
        }

        set_buffer_modified_flag (0);
        bob ();
        file = Null_String;
        message (Null_String);

        EXECUTE_ERROR_BLOCK;
        if (strlen (file)) () = find_file (file);
    }
    EXECUTE_ERROR_BLOCK;
}
add_to_hook ("_jed_startup_hooks", &jed_startup_hook);


define command_line_hook ()
%----------------------------------------------------------
%   Function called from main.c to process the command line.
%   The general form of the command line is:
%
%       sue [-a other_rc_file] [-viewer] [file ...]
%
%   Options must come first and in the order stated.
%
%   Arguments:
%
%   [-a other_rc_file]          resource file other than default
%   [-viewer]                   readonly/no shell
%
%   [file ...]                  one or more files to be edited
%                               files will be opened if they exist
%                               or created if not [not in the viewer]
%                               wild cards can be used
%----------------------------------------------------------
{
    %   very essential, very sensitive function
    %   changes need plenty of time for debugging

    %   The global variable __argc gives the number of
    %   command line arguments. The arguments are in the
    %   array __argv numbered from 0 to __argc - 1.

    variable
        n,
        nf,
        i,
        file=Null_String,
        depth,
        init_file;

    init_file = dircat (Jed_Home_Directory, "sue.rc");

    if (1 != file_status (init_file))
        init_file = Default_Jedrc_Startup_File;

    set_color_scheme (_Jed_Default_Color_Scheme);

    if (__argc==1) return;

    %   malformed input: "-a" and nothing follows
    else if (__argc==2 and __argv[1] == "-a") return;

    %   malformed input: "-a -viewer"
    else if (__argc > 2)
    {
        if (__argv[1] == "-a" and __argv[2] == "-viewer")
            return;
    }

    %   avoiding malformed arguments implies that you have
    %   to trick to edit files called "-a" or "-viewer"

    n = __argc;

    %  skip argv[0], i.e. editor name
    --n;
    i = 1;
    file = __argv[i];

    %   alternative init_file must be first argument
    if (n)
    {
        if (file == "-a")
        {
            i++;
            n--;
            init_file = __argv[i];

            i++;
            n--;

            if (1 != file_status (init_file))
            {
                flush (strcat ("Init File does not exist: ", init_file));
                sleep (2);
                return;
                %   editor will start with defaults
                %   will not start as viewer
            }
        }

        else if (file == "-viewer")
        {
            i++;
            n--;
            jed_load_readonly=1;

            if (__argc < 3) return;
            %   here the viewer starts with no file
        }
    }

    if (init_file != NULL)
    {
        if (file_status (init_file) != 1)
            init_file = Default_Jedrc_Startup_File;

        depth = _stkdepth ();
        if (file_status (init_file) == 1) require(init_file);
        depth = _stkdepth () - depth;

        if (depth)
        {
            flush ("Excess junk left on stack by " + init_file);
            sleep (1);
            _pop_n (depth);
        }
    }

    !if (n) return;

    nf = n;
    while (n > 0)
    {
        file = __argv[i];

        --n;
        ++i;

        if (jed_load_readonly and file_status (file) != 1)
            continue;
        else () = find_file(patch_cmdline_file (file));
    }

    if (nf and jed_load_readonly==0)
        flush ("Total files opened or new: " + string(nf));
}


define shape_cursor1 ()
%----------------------------------------------------------
%   Shape cursor for read-only, overwrite, insert
%----------------------------------------------------------
{
    if (is_readonly) tt_send ("\033[?1c");
    %   invisible cursor

    else if (is_overwrite_mode) tt_send ("\033[?6c");
    %   huge block cursor

    else tt_send ("\033[?2c");
    %   blinking underscore
}


define shape_cursor0 ()
%----------------------------------------------------------
%   Shape cursor when changing buffer
%----------------------------------------------------------
{
    !if (MINIBUFFER_ACTIVE)
        set_buffer_hook ("update_hook", &shape_cursor1);
}
add_to_hook ("_jed_switch_active_buffer_hooks", &shape_cursor0);


%*********************************************************
%  Initialize minibuffer, load file manager
%*********************************************************

autoload("mini_init_minibuffer","basics");
mini_init_minibuffer ();
autoload("dls","dls");
autoload("dtr","dtr");

wordset();
require("sue");
