%**********************************************************
%   SUE, Single User Editor
%   derived from sundry JED modes
%   targets only the LINUX CONSOLE on stand-alone x86 PCs
%   targets only one mode: no mode!
%   Latin1 character set, US keyboard only, no mouse
%**********************************************************

  _debug_info = 1;
  _traceback = 1;

%   BASIC VARIABLES  **************************************

static variable rect_reminder =
    "Cancel rectangle first.";

variable qkm = Assoc_Type [Mark_Type];
%   used to ensure just one quickmark per buffer

define block_all();
define block_append();
define block_cancel();
define block_copy();
define block_cut();
define block_cut_append();
define block_fill_rectangle();
define block_kill_rectangle();
define block_paste();
define block_restore_rectangle();
define block_restore_selection();
define block_save();
define block_shift_left();
define block_shift_right();
define block_toggle_lines();
define block_toggle_rectangle();
define case_generic(what);
define case_lower();
define case_toggle();
define case_toggle_initial();
define case_upper();
define disk_dirlist();
define disk_exit();
define disk_open();
define disk_read_and_insert();
define disk_treeview();
define ed_backtab();
define ed_delete_backward();
define ed_delete_here();
define ed_delete_word();
define ed_display(stringA, stringB);
define ed_evalbuffer();
define ed_insert_literal();
define ed_manage_readonly();
define ed_nextbuffer();
define ed_permissions(o_string);
define ed_properties();
define ed_quick_untab();
define ed_shellcmd();
define ed_softtab();
define ed_time(sstt);
define ed_undelete();
define ed_untab();
define ed_wrap_lines();
define ed_wrap_or_skip();
define go_beginbuffer_select();
define go_beginline_select();
define go_begin_buffer();
define go_begin_line();
define go_cursorleft_select();
define go_cursorright_select();
define go_cursor_left();
define go_cursor_right();
define go_down_select();
define go_down_vertically();
define go_endbuffer_select();
define go_endline_select();
define go_end_buffer();
define go_end_line();
define go_pagedn_select();
define go_pageup_select();
define go_page_down();
define go_page_up();
define go_paragraph_down();
define go_paragraph_up();
define go_to();
define go_to_column();
define go_to_line();
define go_to_matching_fence();
define go_to_quickmark();
define go_try_match_again();
define go_up_select();
define go_up_vertically();
define go_wordleft_select();
define go_wordright_select();
define go_word_left();
define go_word_right();
define macro_play();
define macro_repeat();
define macro_start_or_stop();
define replace_do_replace(str, len);
define replace_interactively();
define replace_on_demand(pat, rep);
define replace_rx_interactively();
define search_again_backward();
define search_again_forward();
define search_backward();
define search_forward();
define search_generic(direction);
define search_mark_nchars(n, scroll);
define search_occurrences_0();
define search_occurrences_1();
define search_rx_backward();
define search_rx_forward();
define set_quickmark();
define shiftkey_check();
define ttest();


define shiftkey_check ()
%----------------------------------------------------------
%   Shift not pressed:  cancel an existing character block
%                       moving the cursor
%
%   Shift pressed:      start a character block moving
%                       the cursor
%
%   To start/stop extending a block, movement must be with keys:
%   Left/Right/Up/Down, Home/End/CtrlHome/CtrlEnd, PageUp/PageDn,
%   CtrlLeft/CtrlRight.
%
%   This function is not assigned to any key,
%   it is used with cursor movements to extend a selection,
%   more precisely selection of a character stream
%----------------------------------------------------------
{
    if (what_block () != 0) return;
    %   block type must be char stream

    ERROR_BLOCK _clear_error ();

    variable s = shiftkey_select ();
    %  this intrinsic reports whether the shiftkey
    %  is on or off in the current buffer

    if (SHIFTKEY_PRESSED)
    {
        if (markp)
        {
            %   block selection already started
            !if (s)
            {
                %   block selection not based on shift-marking
                %   cancel and restart
                pop_mark (0);
                call ("set_mark_cmd");
            }
        }

        else
        {
            %   there is no selection, start one
            call ("set_mark_cmd");
        }
    }

    %   shiftkey not pressed, cancel any block selection
    else if (markp) pop_mark (0);
}


%**********************************************************
%   search and replace functions incl. regular expressions
%   normal search always case insensitive
%   regex  search always case   sensitive
%**********************************************************


define search_mark_nchars (n, scroll)
%----------------------------------------------------------
%   used to highlight a string found in normal search
%   or the begin of a string found in regex search
%----------------------------------------------------------
{
    if (what_block==2) return;

    %   scroll if string reaches past right border of screen
    if (scroll)
        if (right(n))
        {
            update(0);
            go_left(n);
        }

    %   string found is highlighted as region
    %   but should have a different colour from normal block
    %   we redefine temporarily the colours for regions

    variable fgc_region=Null_String, bgc_region=Null_String;
    (fgc_region, bgc_region) = get_color ("region");
    set_color ("region", "white", "lightgray");

    push_mark ();
    call ("set_mark_cmd");
    go_right (n);
    call("exchange");
    update (0);
    pop_mark (0);
    %   region colours back to usual
    set_color ("region", fgc_region, bgc_region);
    ungetkey(getkey);
}


define search_rx_forward ()
%----------------------------------------------------------
%    Will ask for a regular expression and search forward for it.
%    Will stop at the first hit between start position and end of
%    buffer; will stay put if nothing found.
%----------------------------------------------------------
{
    if (check_rectangle)
    {
        flush (rect_reminder);
        return;
    }

    variable pat=Null_String, ffound;

    SEARCH_TYPE = 1;
    CASE_SEARCH = 1;

    pat = read_mini2 ("Regex Search Forward:", Null_String, Null_String);
    !if (strlen(pat)) return;

    push_mark ();

    go_right (1);

    ffound = re_fsearch(pat);

    if (ffound)
    {
        %search_mark_nchars (1,1);
        LAST_SEARCH = pat;
    }
    else flush("Not found.");

    pop_mark (not(ffound));
}


define search_rx_backward ()
%----------------------------------------------------------
%    Will ask for a regular expression and search backward for it.
%    Will stop at the first hit between start position and top of
%    buffer; will stay put if nothing found.
%----------------------------------------------------------
{
    if (check_rectangle)
    {
        flush (rect_reminder);
        return;
    }

    variable pat=Null_String, ffound;

    SEARCH_TYPE = 1;
    CASE_SEARCH = 1;

    pat = read_mini2
    ("Regex Search Backward:", Null_String, Null_String);

    !if (strlen(pat)) return;

    push_mark ();

    ffound = re_bsearch(pat);

    if (ffound) LAST_SEARCH = pat;
    else flush("Not found.");

    pop_mark (not(ffound));
}


define search_generic (direction)
%------------------------------------------------
%   when argument positive, search_forward
%   or else search_backward
%------------------------------------------------
{
    %   normal (non regex) search case-insensitive
    %   use regular expressions for case sensitivity

    variable sstr=Null_String, prompt, ffound;

    if (direction > 0) prompt = "Search forward:";
    else prompt = "Search backward:";

    sstr = read_mini2 (prompt, Null_String, Null_String);

    !if (strlen(sstr)) return;

    if(is_substr (sstr, "\n"))
    {
        %   0 if '\n' is not in string
        flush ("Search string must not include a line break.");
        return;
    }
    push_mark ();

    if (direction > 0)  go_right (1);

    SEARCH_TYPE=0;
    %   search is not regex

    if (direction > 0) ffound = fsearch (sstr);
    else ffound = bsearch (sstr);

    if (ffound)
    {
        search_mark_nchars (strlen(sstr),1);
        LAST_SEARCH = sstr;
    }
    else flush ("String not found.");

    %   not(0) is 1
    %   not(1) is 0

    pop_mark (not(ffound));
}


define search_forward ()
%----------------------------------------------------------
%   Will ask for a string and search forward for it disregarding
%   case. Will stop at the first hit between start position and end
%   of buffer; will stay put if nothing found.
%----------------------------------------------------------
{
    if (check_rectangle)
    {
        flush (rect_reminder);
        return;
    }

    search_generic(1);
}


define search_backward ()
%----------------------------------------------------------
%    Will ask for a string and search backward for it disregarding
%    case. Will stop at the first hit between start position and
%    begin of buffer; will stay put if nothing found.
%----------------------------------------------------------
{
    if (check_rectangle)
    {
        flush (rect_reminder);
        return;
    }

    search_generic(-1);
}


define search_again_forward ()
{
    if (LAST_SEARCH=="") return;
    %   empty string, nothing to do

    if (check_rectangle)
    {
        flush (rect_reminder);
        return;
    }

    if (markp and block_type == 0) pop_mark(0);

    variable r;

    go_right(1);
    if (SEARCH_TYPE) r = re_fsearch (LAST_SEARCH);
    else r = fsearch (LAST_SEARCH);

    if (r)
    {
        if (SEARCH_TYPE) search_mark_nchars (1,1);
        else search_mark_nchars (strlen(LAST_SEARCH),1);
        %   in regex search highlight one char only
        %   since length may vary
    }
    else
    {
        go_left (1);
        flush ("String not found.");
    }
}


define search_again_backward ()
{
    if (LAST_SEARCH=="") return;
    %   empty string, nothing to do

    if (check_rectangle)
    {
        flush (rect_reminder);
        return;
    }

    if (markp and block_type == 0) pop_mark(0);

    variable r;

    if (SEARCH_TYPE) r = re_bsearch (LAST_SEARCH);
    else r = bsearch (LAST_SEARCH);

    if (r)
    {
        if (SEARCH_TYPE) search_mark_nchars (1,1);
        else search_mark_nchars (strlen(LAST_SEARCH),1);
        %   in regex search highlight one char only
        %   since length may vary
    }
    else flush ("String not found.");
}


define search_occurrences_0 ()
%----------------------------------------------------------
%   case insensitive occurrences
%----------------------------------------------------------
{
    if (check_rectangle) return;
    if (markp and block_type) return;

    if (markp)
    {
        call ("copy_region");
        %   get selection into the pastebuffer

        restore_selection();
    }

    occur_list (0);
}


define search_occurrences_1 ()
%----------------------------------------------------------
%   regex occurrrences, case insensitive here
%----------------------------------------------------------
{
    if (check_rectangle) return;
    if (markp and block_type) return;

    if (markp)
    {
        call ("copy_region");
        %   get selection into the pastebuffer

        restore_selection();
    }

    CASE_SEARCH = 1;
    occur_list (1);
}


define replace_do_replace (str, len)
%----------------------------------------------------------
%   ancillary function, used in (non regex) replacements
%----------------------------------------------------------
{
    return replace_chars (len, str);

    %   non regex, return is strlen(str)
}


define replace_on_demand(pat, rep)
%----------------------------------------------------------
%   For both regex and non regex replacements
%   pat is the string/pattern to be replaced, never empty!
%   rep is the replacement for it, possibly empty
%----------------------------------------------------------
{
    variable ch, i=0, pat_len, query=1, tmp;
    variable tmpstr=NULL, tmplen=NULL, tmpmrk=NULL;
    variable search_fun, rep_fun;

    variable prompt =
    "R: Replace,  S: Skip,  A: AllReplace,  Q: Quit";

    if (SEARCH_TYPE)
    {
        %   regex
        CASE_SEARCH = 1;
        search_fun  =   &re_fsearch;
        rep_fun     =   &replace_match;
    }

    else
    {
        %   normal (non regex)
        search_fun  =   &fsearch;
        rep_fun     =   &replace_do_replace;
    }

    while (pat_len = @search_fun(pat), pat_len > 0)
    {
        !if (query)
        {
            %   replacing all at once from cursor downwards

            tmp = create_user_mark ();
            () = @rep_fun (rep, pat_len);
            if (tmp == create_user_mark) go_right (1);
            i++;
            continue;
        }

        flush (prompt);

        if (SEARCH_TYPE); %search_mark_nchars(1,0);
        else search_mark_nchars(pat_len,0);

        %   hilight only first character for regex
        %   regular expression can match varying lengths

        %   if instance is at bottom of window, recenter
        if (window_info('r') == what_line)
            recenter (window_info('r') / 2);

        ch = toupper (getkey);

        switch (ch)

            {
            case 'R' or case '\r' :
                push_spot ();

                push_mark ();
                go_right (pat_len);
                tmpstr = bufsubstr ();
                %   mark is popped

                pop_spot ();
                %   back to initial position

                tmpmrk = create_user_mark ();

                tmplen = @rep_fun (rep, pat_len);
                %   cursor goes past replacement

                %   tmplen only relevant for non regex case
                %   any return from rep_fun is discarded for regex
                %   also pat_len is not used for regex replacement
                %   string effectively replaced is in "tmpstr"
                %
                %   do not use strlen(rep) when string
                %   chopped up across lines

                i++;
            }

            {
            case 'S' or case ' ':
                go_right(1);
            }

            {
            case 'A':
                query = 0;
            }

            {
            case 'Q' or case '' or case '':
                %   ESC or ctrl-w or 'Q'
                break;
            }

            {
                flush_input ();
                continue;
            }
        %end switch on ch
    }
    flush ("Total replacements: "+string(i));
}


define replace_rx_interactively ()
%----------------------------------------------------------
%   starts interactive regex replacements
%----------------------------------------------------------
{
    %   this is regex replace in the pulldowns

    if (is_readonly) return;

    if (markp)
    {
        flush ("Not allowed when block highlighted.");
        return;
    }

    variable
        pat = Null_String,
        n,
        rep = Null_String,
        prompt = "String to insert as replacement:",
        ch;

    pat = read_mini2
    ("Regex expression to replace:", Null_String, Null_String);

    !if (strlen (pat)) return;
    if (LASTKEY[0] == 23) return;

    rep = read_mini2 (prompt, Null_String, pat);
    if (LASTKEY[0] == 23) return;

    SEARCH_TYPE=1;
    CASE_SEARCH = 1;
    %   search for pattern to be replaced is regex and case-sensitive

    replace_on_demand(pat, rep);
}


define replace_interactively ()
%----------------------------------------------------------
%   Will start non-regex interactive replacement
%----------------------------------------------------------
{
    if (is_readonly) return;

    if (markp)
    {
        flush ("Not allowed when block highlighted.");
        return;
    }

    variable pat=Null_String, prompt=Null_String, rep=Null_String;

    pat = read_mini2
    ("String to be replaced:", Null_String, Null_String);

    !if (strlen(pat)) return;

    if (LASTKEY[0] == 23) return;
    %   discard minibuffer reply with ^w, ascii 23

    prompt = "String to be inserted as replacement:";

    rep = read_mini2 (prompt, Null_String, pat);

    if (rep == pat) return;

    if (LASTKEY[0] == 23) return;
    %   discard minibuffer reply with ^w, ascii 23

    SEARCH_TYPE=0;
    %   search for pattern to be replaced is not regex

    replace_on_demand(pat, rep);
}



%**********************************************************
%   functions for macro recording and re-playing
%**********************************************************


define macro_start_or_stop ()
%----------------------------------------------------------
%   Toggles recording of a macro on or off
%----------------------------------------------------------
{
    if (DEFINING_MACRO) call("end_macro");
    else call("begin_macro");
}


define macro_repeat ()
%----------------------------------------------------------
%   Runs a macro as many times as entered at the prompt.
%----------------------------------------------------------
{
    variable r=0, prompt="Repeat last macro how many times?";

    r = integer(read_mini2(prompt, "0", "0"));

    loop (r) call("execute_macro");
}


define macro_play ()
%----------------------------------------------------------
%   Execute last macro, i.e. repeat once
%----------------------------------------------------------
{
    call("execute_macro");
}


%**********************************************************
%   blocks copying, pasting, filling etc.
%**********************************************************


define block_cancel ()
{
    if (markp) call("cancell");
}



define block_toggle_rectangle ()
%----------------------------------------------------------
%   Starts a rectangle, if there is none.
%   Cancels a rectangle, if there is one.
%----------------------------------------------------------
{
    if (is_readonly) return;
    toggle_block (1);
}


define block_kill_rectangle ()
%----------------------------------------------------------
%   Will delete a rectangle
%----------------------------------------------------------
{
    %   if no column block, nothing to do
    !if (check_rectangle) return;

    %   global variables provided by check_rectangle ()
    variable
        beg_col   = W_COL,
        end_col   = E_COL,
        beg_line  = N_ROW,
        end_line  = S_ROW;

    %   do not delete rectangle with cut_rect
    %   to avoid polluting the pastebuffer
    %   del_region does not work well here
    %   so delete piecewise while looping

    goto_line(beg_line);
    goto_column(beg_col);
    variable rwidth = end_col - beg_col +1;

    loop (end_line - beg_line + 1)
    {
        goto_column (beg_col);
        loop(rwidth) call ("delete_char_cmd");
        go_down (1);
    }

    goto_line (beg_line);
    goto_column (beg_col);
    %   cursor in NW corner

    toggle_block (0);
}


define block_fill_rectangle ()
%----------------------------------------------------------
%   Fills a rectangle with string entered at the prompt
%----------------------------------------------------------
{
    variable str=Null_String, tmpstr, l, width;
    variable beg_col, beg_line, end_col, end_line;

    %   if no column block, nothing to do
    !if (check_rectangle) return;

    %   global variables provided by check_rectangle ()
    beg_col   = W_COL;
    end_col   = E_COL;
    beg_line  = N_ROW;
    end_line  = S_ROW;

    str = read_mini2 ("Fill rectangle with:", "", "");
    if (str=="") str = " ";
    %   empty reply will blank rectangle

    tmpstr = str;
    l = strlen (str);

    width = end_col - beg_col + 1;
    %   width corresponds to cursor inside rectangle

    if (l > width) str = substr (str, 1, width);
    %   string too long, cut to width

    if (l < width)
    {
        loop (width/l-1) str = str + tmpstr;
        %   string too short
        %   concatenate to itself as far as possible

        l = strlen (str);
        str = str + substr (str, 1, width - l);
        %   add a substr to fill gap to width
    }

    %   do not delete rectangle with cut_rect
    %   to avoid polluting the pastebuffer
    %   del_region does not work well here
    %   so delete piecewise while inserting

    goto_line(beg_line);
    goto_column(beg_col);

    loop (end_line - beg_line + 1)
    {
        goto_column (beg_col);
        insert (str);
        %   insert is an intrinsic
        loop(strlen(str)) call ("delete_char_cmd");
        go_down (1);
    }

    goto_line (beg_line);
    goto_column (beg_col);
    %   cursor in NW corner

    toggle_block (0);

    %   use restore_rectangle but not restore_selection
    %   to restore highlight
}


define block_toggle_lines ()
%----------------------------------------------------------
%   if whole lines are selected, cancel the selection
%   otherwise start whole lines selection
%----------------------------------------------------------
{
    toggle_block (2);
}


define block_save ()
%----------------------------------------------------------
%   Will save a block to disk file.
%----------------------------------------------------------
{
    save_selection();
    flush ("Done");
    %   error message or more precise result possible
}


define block_shift_left ()
%----------------------------------------------------------
%   Will shift block to the left.
%----------------------------------------------------------
{
    variable tmpblk = what_block ();

    if (tmpblk == 0)
    {
        flush ("First select lines or columns.");
        return;
    }

    if (tmpblk == 1)
    {
        shift_rectangle_left ();
        return;
    }

    if (tmpblk == 2) shift_linetext_left ();
}


define block_shift_right ()
%----------------------------------------------------------
%   Will shift block to the right.
%----------------------------------------------------------
{
    variable tmpblk = what_block ();

    if (tmpblk == 0)
    {
        flush ("First select lines or columns.");
        return;
    }

    if (tmpblk == 1)
    {
        shift_rectangle_right ();
        return;
    }

    if (tmpblk == 2) shift_linetext_right ();
}


define block_all ()
%----------------------------------------------------------
%   Selects entire buffer.
%----------------------------------------------------------
{
    if (bobp and eobp) return;

    if (markp) toggle_block (0);

    !if (is_readonly) trim_lines ();
    %   trailing spaces will disappear

    bob ();
    %   to begin of buffer

    toggle_block (2);
    %   start selection here

    eob ();
    %   extend to end of buffer
}


define block_paste ()
%----------------------------------------------------------
%   Will insert content of clipboard.
%----------------------------------------------------------
{
    if (markp) return;
    %   no insertion if block active

    call ("yank");

    %   from pastebuffer (character stream)
    %   or rectangle buffer
    %   whichever was last filled with block_copy

    %   to revive highlight for the entire insertion
    %   use restore_selection or restore_rectangle

    %   when pasting stream or whole lines
    %       cursor positioned immediately after insertion
    %   when pasting rectangles
    %       cursor does not move
    %   whole lines get a newline at end of insertion
}


define block_copy ()
%----------------------------------------------------------
%   Copy selection to clipboard.
%   No selection, no action
%----------------------------------------------------------
{
    !if (markp) return;

    if (check_rectangle)
    {
        copy_rect ();
        toggle_block (0);
        return;
    }

    push_spot ();
    call ("copy_region");
    pop_spot ();

    %   copy_region ->      copy_region_cmd             intrin.c
    %   this is not the function used above

    %   copy_region ->          copy_to_pastebuffer     keymap.c
    %   copy_to_pastebuffer ->  copy_region_to_buffer   paste.c
    %   copy_region_to_buffer                           paste.c
}


define block_append ()
%----------------------------------------------------------
%   Copy to clipboard and append selection to scratch buffer
%   No selection, no action
%----------------------------------------------------------
{
    !if (markp) return;

    if (check_rectangle)
    {
        copy_rect ();
        toggle_block (0);
        return;
    }

    push_spot ();
    call ("copy_region");
    pop_spot ();

    %   copy_region ->      copy_region_cmd             intrin.c
    %   this is not the function used above

    %   copy_region ->          copy_to_pastebuffer     keymap.c
    %   copy_to_pastebuffer ->  copy_region_to_buffer   paste.c
    %   copy_region_to_buffer                           paste.c

    %   append selection to scratch buffer
    variable tmpbf=whatbuf ();
    sw2buf ("*scratch*");
    eob ();
    newline ();
    block_paste ();
    sw2buf (tmpbf);
}


define block_cut ()
%----------------------------------------------------------
%   cut to clipboard
%----------------------------------------------------------
{
    if (check_rectangle)
    {
        cut_rect ();
        return;
    }

    variable thisblock = what_block ();
    call ("kill_region");
    if (thisblock==2) call ("delete_char_cmd");

    %   rectangle_buffer or default pastebuffer used
    %   whichever was filled last
}


define block_cut_append ()
%----------------------------------------------------------
%   cut to clipboard and also append to scratch buffer
%----------------------------------------------------------
{
    if (check_rectangle)
    {
        cut_rect ();
        return;
    }

    variable thisblock = what_block ();
    call ("kill_region");
    if (thisblock==2) call ("delete_char_cmd");

     %   append the region to scratch buffer, if you so want
     variable tmpbf=whatbuf ();
     sw2buf ("*scratch*");
     eob ();
     newline ();
     block_paste ();
     sw2buf (tmpbf);

    %   rectangle_buffer or default pastebuffer used
    %   whichever was filled last
}


define block_restore_rectangle ()
%----------------------------------------------------------
%   Will restore a rectangle immediately after cancelling it.
%----------------------------------------------------------
{
    restore_rectangle();
}


define block_restore_selection ()
%----------------------------------------------------------
%   Will restore a selection immediately after cancelling it.
%----------------------------------------------------------
{
    restore_selection();
}



%**********************************************************
%   functions to get upper/lower/toggle case
%**********************************************************


define case_generic (what)
%----------------------------------------------------------
%   if block selected, operate on block
%   otherwise operate on current character
%   parameter 'what' can be
%   'l'     lower case
%   'u'     upper case
%   't'     toggle case
%----------------------------------------------------------
{
    if (is_readonly) return;

    if (check_rectangle)
    {
        handle_case_in_rectangle (what);
        return;
    }

    push_spot();
    () = check_stream();
    %   required so restoring hilight is possible

    if (markp)
    {
        xform_region (what);
        pop_spot();
        return;
    }

    %   no block, operate on character

    push_mark ();
    go_right (1);
    xform_region (what);
    pop_spot ();
}


define case_lower ()
{
    case_generic ('l');
}


define case_upper ()
{
    case_generic ('u');
}


define case_toggle ()
{
    case_generic ('t');
}


define case_toggle_initial ()
{
    if (is_readonly) return;

    if (markp) return;

    push_spot ();

    begin_of_word_right ();
    begin_of_word_left ();

    push_mark ();
    go_right (1);
    xform_region ('t');
    pop_spot ();
}



%**********************************************************
%   functions to move the cursor
%**********************************************************


define go_to_quickmark ()
%----------------------------------------------------------
%   Jump to quickmark in current buffer, if set
%----------------------------------------------------------
{
    if (check_rectangle) return;
    shiftkey_check();
    %   line block will be extended, but not char stream

    if (assoc_key_exists (qkm, whatbuf))
        goto_user_mark (qkm[whatbuf ()]);
}


define set_quickmark ()
%----------------------------------------------------------
%   Will set a quickmark, exactly one per buffer.
%   To unset a quickmark, set it elsewhere. Or just ignore it.
%----------------------------------------------------------
{
    qkm [whatbuf ()] = create_user_mark ();
}


define go_try_match_again ()
%----------------------------------------------------------
%   If fence matching fails, try again with simple search
%----------------------------------------------------------
{
    if (what_block != 0) return;

    variable
        cs,
        match_cs,
        direction;

    cs = char (what_char);

    switch (cs)
        { case roundL:   direction = 0; }
        { case roundR:   direction = 1; }
        { case squareL:  direction = 0; }
        { case squareR:  direction = 1; }
        { case curlyL:   direction = 0; }
        { case curlyR:   direction = 1; }
        {                return;        }

    if (direction)
    {
        go_right (1);
        fsearch (cs);
    }
    else bsearch (cs);
}


define go_to_matching_fence ()
%----------------------------------------------------------
%   If cursor on any of (){}[] jump to its match, if possible
%----------------------------------------------------------
{
    if (markp and what_block () < 2) return;
    %   do not move if there is a rectangle or a stream
    %   but move if it is a line block

    push_mark ();

    variable r = 0;
    r = find_matching_delimiter(0);
    %   function modified to return either 0 or 1

    pop_mark (1-r);
    %   if match found, jump there; if not, stay put

    return r;
}


define go_page_up ()
%----------------------------------------------------------
%   In the console of 25 lines, JED pages down und up to line
%   21*n + 1, i.e. 1, 22, 43 etc. If the cursor is say at 47,
%   page_up takes it to 22 and the current line is top of window.
%
%   The following function takes first to top of window
%   then to next page at top of window again.
%----------------------------------------------------------
{
    if (check_rectangle) return;
    shiftkey_check();

    variable start_line = window_line ();
    %   cursor is on start_line

    variable current_column = what_column ();
    %   save column position

    goto_top_of_window ();
        % move to top of window, if not already there

    if (window_line () != start_line)
    {
        goto_column (current_column);
        return;
        % moved, current line is not start_line
    }
    else
    {
        %   already at top, now paginate
        call ("page_up");
        goto_top_of_window ();
    }
    goto_column (current_column);
}


define go_page_down ()
%----------------------------------------------------------
%   In the console of 25 lines, JED pages down und up to line
%   21*n + 1, i.e. 1, 22, 43 etc. If the cursor is say at 17,
%   page_down takes it to 22 and the current line is top of window.
%
%   The following function takes first to bottom of window
%   then to next page at bottom of window again.
%----------------------------------------------------------
{
    if (check_rectangle) return;
    shiftkey_check();

    variable start_line = window_line ();
    % cursor is on start_line

    variable current_column = what_column ();
    %   save column position

    goto_bottom_of_window ();
    % move to bottom of window, if not already there

    if (window_line () != start_line)
    {
        goto_column (current_column);
        return;
        % moved, current line is not start_line
    }

    else
    {
        % already at bottom of window, now paginate
        call ("page_down");
        goto_bottom_of_window ();
    }
    goto_column (current_column);
}


define go_pageup_select ()
%----------------------------------------------------------
%   Function extends a selection using PgUp
%----------------------------------------------------------
{
    if (check_rectangle) return;
    shiftkey_check();
    go_page_up ();
}


define go_pagedn_select ()
%----------------------------------------------------------
%   Function extends a selection using PgDn
%----------------------------------------------------------
{
    if (check_rectangle) return;
    shiftkey_check();
    go_page_down ();
}


define go_to ()
%----------------------------------------------------------
%   go to line/column/matching fence/previous position/quickmark
%   will pull down menu to select destination
%----------------------------------------------------------
{
    if (check_rectangle) return;

    menu_select_menu ("Global.&Search.&Goto");

    %   wait 8/10 of a second
    %   if whithin this delay a key arrives, process the key
    %   if no key arrives, display menu

    variable ky=1;
    if (0!=input_pending(8)) ky=getkey ();
    ungetkey(ky);

    %   after getting familiar whith the menu you will know,
    %   for instance, that typing ^gl quickly means go to line
    %   without seeing the menu
    %   if you are slow, the menu will be pulled down
}


define go_to_line ()
{
    if (check_rectangle) return;
    shiftkey_check();

    ERROR_BLOCK _clear_error ();
    %   this allows excessive line numbers
    %   to take cursor to end of buffer

    variable s = 0;
    s = integer (read_mini2 ("Goto line:", "0", "0"));

    !if (s) return;

    goto_line (s);
    %   cursor is at bol
}


define go_to_column ()
{
    if (check_rectangle) return;
    shiftkey_check();

    ERROR_BLOCK _clear_error ();
    %   this will handle excessive numbers

    variable s = 0;
    s = integer (read_mini2 ("Goto column:", "0", "0"));

    !if (s) return;

    if (s < 10000000) goto_column (s);
    else flush ("Be reasonable.");
    %   spaces will be inserted if necessary
    %   monstruous numbers will be accepted
    %   but execution will take forever
    %   so it is better to impose a limit
}


define go_cursor_left ()
{
    shiftkey_check ();
    call ("previous_char_cmd");
}


define go_cursor_right ()
{
    shiftkey_check ();
    call ("next_char_cmd");
}


define go_cursorleft_select ()
{
    shiftkey_check ();
    call ("previous_char_cmd");
}


define go_cursorright_select ()
{
    shiftkey_check ();
    call ("next_char_cmd");
}



define go_word_right ()
{
    shiftkey_check ();
    begin_of_word_right ();
}


define go_word_left ()
{
    shiftkey_check ();
    begin_of_word_left ();
}


define go_wordright_select ()
{
    shiftkey_check ();
    begin_of_word_right ();
}


define go_wordleft_select ()
{
    shiftkey_check ();
    begin_of_word_left ();
}


define go_up_vertically ()
{
    shiftkey_check();
    call("previous_line_cmd");
}


define go_down_vertically ()
{
    shiftkey_check();
    call("next_line_cmd");
}

define go_up_select ()
{
    shiftkey_check();
    call("previous_line_cmd");
}


define go_down_select ()
{
    shiftkey_check();
    call("next_line_cmd");
}


define go_paragraph_up ();
%   needed for recursive call

define go_paragraph_up ()
%----------------------------------------------------------
%   Move back one paragraph
%----------------------------------------------------------
{
    if (what_block () != 0) return;
    %   stay put when lines/rectangles highlighted

    %   cursor is on a whiteline
    %   and will skip whitelines up to prev paragraph
    if (is_whiteline)
    {
        while (is_whiteline () and not bobp) go_up (1);
        if (bobp) return;
    }

    %   cursor is somewhere in a paragraph
    %   possibly on the top line of paragraph?
    else
    {
        go_left (1);
        %   if cursor was at bol, it is now on prev line
        %   otherwise it is still in the paragraph

        if (is_whiteline () and not bobp) go_paragraph_up ();
        %   recursive call
    }

    %   cursor is on a paragraph
    %   jump back past it

    backward_paragraph ();
    if (bobp) return;

    %   cursor is just above paragraph, step back
    go_down (1);
}


define go_paragraph_down ()
%----------------------------------------------------------
%   Move forward one paragraph
%----------------------------------------------------------
{
    if (what_block () != 0) return;
    %   stay put when lines/rectangles highlighted

    %   cursor is on a whiteline
    %   and will glide down to the next non-whiteline
    if (is_whiteline)
    {
        while (is_whiteline () and not eobp) go_down (1);
        return;
    }

    %   cursor is somewhere in a paragraph
    %   jump past the paragraph then skip whitelines
    forward_paragraph ();
    while (is_whiteline () and not eobp)
        !if (down(1)) break;

    %   may reach eobp
}


define go_begin_line ()
{
    shiftkey_check();
    bol();
}


define go_end_line ()
%----------------------------------------------------------
%   go to eol but before trailing whitespace
%   cursor will rest on newline code or eob-code
%----------------------------------------------------------
{
    shiftkey_check();
    eol();
    bskip_chars (" \t");

    %call("eol_cmd");   does trim
}


define go_beginline_select ()
{
    shiftkey_check();
    bol();
}


define go_endline_select ()
%----------------------------------------------------------
%   go to eol but before trailing whitespace
%   cursor will rest on newline code or eob-code
%----------------------------------------------------------
{
    shiftkey_check();
    eol();
    bskip_chars (" \t");

    %call("eol_cmd");   does trim
}


define go_begin_buffer ()
{
    if (check_rectangle)
    {
        flush ("Not allowed when rectangle highlighted.");
        return;
    }

    shiftkey_check();
    bob();
}


define go_end_buffer ()
{
    if (check_rectangle)
    {
        flush ("Not allowed when rectangle highlighted.");
        return;
    }

    shiftkey_check();
    eob();
}


define go_beginbuffer_select ()
{
    if (check_rectangle)
    {
        flush ("Not allowed when rectangle highlighted.");
        return;
    }

    shiftkey_check();
    bob();
}


define go_endbuffer_select ()
{
    if (check_rectangle)
    {
        flush ("Not allowed when rectangle highlighted.");
        return;
    }

    shiftkey_check();
    eob();
}


define ed_delete_word ()
%----------------------------------------------------------
%   If cursor is on word or right-adjacent to it, delete
%   this word reducing trailing delimiters to just one.
%   Otherwise delete delimiters up to next word right.
%   Do nothing at end of line (delimited by hard return)
%   if cursor not adjacent to word. Current line only.
%   Modelled after Wordperfect 5.1 for DOS
%----------------------------------------------------------
{
    if (markp)
    {
        flush ("Not allowed when block highlighted.");
        return;
    }
    %   block is selected, do nothing

    variable p = what_column ();

    bskip_word_chars ();
    %   cursor only moves if right-adjacent to word
    %   or in mid-word position

    if (p != what_column)
    {
        % cursor has moved to begin of word

        push_mark ();
        skip_word_chars ();

        del_region ();
        %   word has been deleted but there might be delimiters right

        push_mark ();
        skip_non_word_chars ();
        % same line, next line untouched

        del_region ();
        % delimiters to next word also deleted
        % next word has now location of deleted word
        % line boundary not trespassed
        return;
    }

    % if control here, cursor is positioned
    % either at begin of word or between words

    push_mark ();
    skip_word_chars ();

    del_region ();
    %   if cursor was at begin of word, word is deleted
    %   otherwise nothing happened

    % cursor is on delimiter

    push_mark ();
    skip_non_word_chars ();

    del_region ();
    %   delimiters up to next word deleted
    %   same line, next line untouched
}


define ed_insert_literal ()
%----------------------------------------------------------
%   accepts character code 0-255,
%   inserts the corresponding character of current font
%----------------------------------------------------------
{
    if (is_readonly) return;

    if (markp)
    {
        flush ("Not allowed when block highlighted.");
        return;
    }

    variable i, j=0, ch, c;

    c = read_mini2
    ("Enter character code (0-255):", Null_String, Null_String);

    !if (Integer_Type == _slang_guess_type (c)) return;

    ch = integer (c);

    if (ch >= 0 and  ch < 256)
    {
        insert_char (ch);
        vmessage ("Character  %d  inserted", ch);
    }
    else flush("Out of range");
}


define ed_delete_here ()
%----------------------------------------------------------
%   Deletion stored in kill_array,
%   char selection on one line can be undeleted anywhere
%   rectangles cannot be undeleted
%----------------------------------------------------------
{
    %   if block selected, delete block
    %   otherwise delete character

    if (check_rectangle)
    {
        block_kill_rectangle ();
        toggle_block (0);
        return;
    }

    if (markp)
    {
        variable thisblock = what_block ();
        del_region ();
        if (thisblock==2) call ("delete_char_cmd");
    }

    else
    {
        if (is_whiteline)
            if (bolp) trim ();

        call ("delete_char_cmd");
    }
}


define ed_delete_backward ()
%----------------------------------------------------------
%   If block selected, do nothing
%   otherwise delete character left
%----------------------------------------------------------
{
    if (markp)
    {
        flush ("Not allowed when block highlighted.");
        return;
    }

    call ("backward_delete_char");
}


define ed_undelete ()
%----------------------------------------------------------
%   Undelete from kill_array, place anywhere
%----------------------------------------------------------
{
    if (is_readonly) return;
    if (markp) return;

    variable rply = "n";

    rply = read_mini2
    ("Undelete last deletion at cursor (y/n)?:", "n", Null_String);
    if (LASTKEY[0] == 23) return;

    !if(strlen(rply)) return;

    if (rply=="y" or rply=="Y")
        insert_from_kill_array(0);
}


define ed_untab ()
%----------------------------------------------------------
%   Replace real tabs with spaces after prompting for it
%
%   It acts from cursor downwards, cursor stays put.
%   Replacement is not a fixed number of spaces for a tab
%   but rather a variable number to reach next tab stop.
%----------------------------------------------------------
{
    if (is_readonly)
    {
        flush ("Buffer is read-only.");
        return;
    }

    if (check_rectangle)
    {
        flush ("Not allowed when rectangle highlighted.");
        return;
    }

    variable
        n,
        q=0,
        ttab=0,
        first_stop=0,
        s=Null_String;

    s = read_mini2
    ("Untab: how many spaces for a tab?", Null_String, Null_String);

    !if (strlen (s)) return;

    ttab = integer (s);

    %   unsuitable entries
    if (ttab < 1) return;

    s = read_mini2
    ("Count tabstops from 0 or 1?", Null_String, "1");

    !if (strlen (s)) return;

    first_stop = integer(s);

    if (first_stop != 1) first_stop = 0;

    push_spot ();

    while (fsearch ("\t"))
    {
        ++q;
        call ("delete_char_cmd");
        n = - first_stop + (what_column() mod ttab);
        loop(ttab - n) insert_char(' ');
    }

    pop_spot();

    flush("Tabs replaced: " + string(q));
}


define ed_quick_untab ()
%----------------------------------------------------------
%   Will replace tabs with spaces, first tab stop in col 1
%
%   It acts only from cursor onwards, cursor will not move.
%   Replacement is not a fixed number of spaces for a tab
%   but rather a variable number to reach next tab stop.
%----------------------------------------------------------
{
    if (is_readonly)
    {
        flush ("Buffer is read-only.");
        return;
    }

    if (check_rectangle)
    {
        flush ("Not allowed when rectangle highlighted.");
        return;
    }

    variable
        n,
        q=0,
        ttab=8,
        first_stop=1,
        s=Null_String;

    push_spot ();

    while (fsearch ("\t"))
    {
        ++q;
        call ("delete_char_cmd");
        n = - first_stop + (what_column() mod ttab);
        loop(ttab - n) insert_char(' ');
    }

    pop_spot();

    flush("Tabs replaced: " + string(q));
}


define ed_softtab ()
%----------------------------------------------------------
%   Line or rectangle selected
%     shift selection right
%
%   No selection
%     insert spaces up to next tab stop
%----------------------------------------------------------
{
    if (markp)
    {
        block_shift_right ();
        return;
    }

    if (TAB_DEFAULT < 1) TAB_DEFAULT=4;

    spaces_for_tabs ();
    %   tab key will insert spaces up to next tab position
}


define ed_backtab ()
%----------------------------------------------------------
%   Line or rectangle selected
%     shift selection right
%
%   No selection
%     jump to previous tab stop
%----------------------------------------------------------
{
    if (markp)
    {
        block_shift_left ();
        return;
    }

    if (TAB_DEFAULT < 1) TAB_DEFAULT=4;

    backtab ();
    %   jump back to tab stops
}


define ed_wrap_lines ()
%----------------------------------------------------------
%   Wraps lines from current line to end of paragraph. Lines are
%   wrapped at the column determined by the variable WRAP which
%   is 70 in the C source but can be set in the resource file.
%   It can also be changed on the fly from the command line.
%----------------------------------------------------------
{
    if (is_readonly) return;

    if (markp)
    {
        flush ("Not allowed when block highlighted.");
        return;
    }

    if(is_whiteline) return;

    %   format from current line downwards
    bol ();
    push_spot ();
    newline ();
    call ("format_paragraph");
    pop_spot ();
    call ("delete_char_cmd");
    %   cursor is at bol of line where it was before
}


define ed_wrap_or_skip ()
%----------------------------------------------------------
%   Shift not pressed:  wrap remaining lines in paragraph
%   Shift pressed:      do not wrap, jump to next paragraph
%----------------------------------------------------------
{
    if (markp)
    {
        flush ("Not allowed when block highlighted.");
        return;
    }

    %   to avoid formatting press shiftkey
    if (SHIFTKEY_PRESSED);
    else ed_wrap_lines ();

    % cursor is at bol of line where it was before
    go_paragraph_down ();

    variable lm = window_info ('r') / 2;
    recenter (lm);
}


define ed_permissions (o_string)
{
    %   o-string is the octal digit string
    %   defining permissions

    variable i, o, perms_letters=Null_String;

    for (i=4;i<7;i++)
    {
        o = substr (o_string,i,1);

        switch (o)
        {
            case "0":
                perms_letters += "---";
        }
        {
            case "1":
                perms_letters += "--x";
        }
        {
            case "2":
                perms_letters += "-w-";
        }
        {
            case "3":
                perms_letters += "-wx";
        }
        {
            case "4":
                perms_letters += "r--";
        }
        {
            case "5":
                perms_letters += "r-x";
        }
        {
            case "6":
                perms_letters += "rw-";
        }
        {
            case "7":
                perms_letters += "rwx";
        }
    }
    return perms_letters;
}


define ed_display (stringA, stringB)
%----------------------------------------------------------
%   displays text built on the fly
%----------------------------------------------------------
{
    sw2buf(stringA);

    set_buffer_modified_flag (0);
    set_readonly(1);
    set_status_line ("  Press any key to exit  ", 0);

    dummy_menu(stringB);

    bob();
    update_sans_update_hook(0);
    flush_input(getkey ());
    % unconditionally
    delbuf(whatbuf);
}


define ed_time (sstt)
{
    variable ttime  = localtime (sstt);
    variable yyear;
    variable mmonth = string(ttime.tm_mon + 1);
    variable dday   = string(ttime.tm_mday);
    variable hhours = string(ttime.tm_hour);
    variable mmins  = string(ttime.tm_min);
    variable msecs  = string(ttime.tm_sec);

    if(ttime.tm_year > 0) yyear=string(ttime.tm_year+1900);
    else yyear="1900";

    if (strlen(dday)<2)   dday   = "0"+dday;
    if (strlen(mmonth)<2) mmonth = "0"+mmonth;
    if (strlen(hhours)<2) hhours = "0"+hhours;
    if (strlen(mmins)<2)  mmins  = "0"+mmins;
    if (strlen(mmins)<2)  msecs  = "0"+msecs;

    %   formatted time needs 20 columns
    variable ftime  =
    yyear + "-" + mmonth + "-" + dday + "  " + hhours + ":" + mmins + ":" + msecs;
    %   4    1         2    1       2     2         2    1        2    1        2

    return ftime;
}


autoload ("viewer_keymap", "suemenu");
autoload ("viewer_menu", "suemenu");
autoload ("suemenu", "suemenu");
define ed_manage_readonly ()
{
    toggle_readonly ();
   return;
    if (is_readonly)
    {
        viewer_keymap ();
        viewer_menu ();
    }

    else
    {
        use_keymap("global");
        suemenu ();
    }
}


define ed_properties()
{
    variable
        tmpbf    = Null_String,
        stringA  = " /on the fly/",
        stringB  = "&Properties",
        full_fn  = Null_String,
        thisfile = Null_String,
        thisdir  = Null_String,
        thisbf   = Null_String,
        fs_flag  = 0,
        fstatus  = 0,
        fmode    = 0,
        bf_flags = 0,
        bf_modif = Null_String,
        bf_ro    = Null_String,
        spaces5  = "     ",
        spaces4  = "    ";

    tmpbf = whatbuf ();
    fs_flag = fs_readonly ();
    fmode = is_readonly();

    (thisfile, thisdir, thisbf, bf_flags) = getbuf_info ();

    full_fn = thisdir + thisfile;
    fstatus = file_status (full_fn);

    %   open temp buffer
    delbuf(stringA);
    sw2buf(stringA);
    newline();
    newline();

    insert(spaces4);
    insert("Buffer:");
    goto_column (21);
    insert(thisbf);
    newline();
    newline();

    insert(spaces4);
    insert("File on disk:");
    goto_column (21);
    if (1 == fstatus) insert(thisfile);

    if (fstatus==1)
    {
        variable st;
        st = stat_file (full_fn);

        newline();
        insert(spaces5);
        insert("permissions:");
        goto_column (21);
        %   convert mode to octal first
        insert(ed_permissions(sprintf("%o",st.st_mode)));

        newline();
        insert(spaces5);
        insert("userID:");
        goto_column (21);
        insert(string(st.st_uid));

        newline();
        insert(spaces5);
        insert("groupID:");
        goto_column (21);
        insert(string(st.st_gid));

        newline();
        insert(spaces5);
        insert("bytes:");
        goto_column (21);
        insert(string(st.st_size));

        newline();
        insert(spaces5);
        insert("modified:");
        goto_column (21);
        insert(ed_time(st.st_mtime));

        if (fs_flag > 0)
        {
            newline();
            insert(spaces5);
            insert("warning!");
            goto_column (21);
            insert("READ-ONLY file system");
        }

        else if (fmode)
        {
            newline();
            insert(spaces5);
            insert("mode:");
            goto_column (21);
            insert("read-only");
        }
    }
    newline();
    newline();

    insert(spaces4);
    insert("Directory:");
    goto_column (21);
    insert(thisdir);
    newline();

    newline();

    insert(spaces4);
    insert("Stack depth:");
    goto_column (21);
    insert(string(_stkdepth));
    newline();

    ed_display (stringA, stringB);
    sw2buf(tmpbf);
}


define ed_evalbuffer ()
{
    ERROR_BLOCK  _clear_error ();

    variable
        tmpex = Null_String,
        tmpbf = whatbuf ();

    tmpex = substr (tmpbf, -2+strlen(tmpbf), 3);

    if (tmpex == ".sl") evalbuffer();
    else
        flush ("Is this an SLang script? Do evaluation by hand.");
}


define ed_shellcmd ()
%----------------------------------------------------------
%   Execute a non-interactive shell command
%----------------------------------------------------------
{
    ERROR_BLOCK
    {
        _clear_error ();
        reset_screen();
    }

    variable
        stringA = "Execute Shell Commands",
        stringB = "non-interactive",
        stringC = "&Shell Commands",
        sstr    = Null_String,
        thisbf  = whatbuf (),
        result  = -1,
        rfile   = "/shellcmd_results/";

    %   display help and warning in temp buffer to be deleted on exit
    display_help (stringA, stringB, stringC);
    %   display_hlp2 bad idea

    sw2buf(rfile);
    erase_buffer ();

    sstr = read_mini2 ("*Non-interactive* Shell Command:", Null_String, Null_String);
    if (LASTKEY[0] == 23 or sstr==Null_String)
    {
        delbuf (rfile);
        sw2buf(thisbf);
        return;
    }

    result = run_shell_cmd (sstr);

    %   as long as the shell command is running,
    %   control will be in the minibuffer

    !if (MINIBUFFER_ACTIVE)
    {
        if (result > 0) reset_screen();
        bob();
        newline();
        insert("Command issued:  ");
        insert(sstr);
        newline();
        newline();
        if(result > 0) insert ("Errors!");
        else insert ("Results:");
        newline();
        newline();
    }
}


define ed_nextbuffer ()
{
    %   only defined to avoid overlapping keybindings
    %   nextbuffer should not have two keybindings
    nextbuffer();
}



%**********************************************************
%   file operations on disk
%**********************************************************


define disk_read_and_insert ()
%----------------------------------------------------------
%   reads a file from disk and inserts it into current buffer
%----------------------------------------------------------
{
    %   avoid interference with blocks
    if (markp)
    {
        flush ("First cancel selection.");
        return;
    }

    if (whatbuf ()=="*scratch*")
    {
        flush ("No insertion from disk to scratch buffer!");
        return;
    }

    if (is_readonly)
    {
        flush ("Buffer is read-only.");
        return;
    }

    variable dir = Null_String, file = Null_String;

    %   ask user for file
    file = read_path_from_mini ("File to insert:");

    !if (strlen(file)) return;

    %   clean up entry
    (dir, file) = parse_filename (file);
    variable fullpath = dircat (dir, file);

    %   if entry ok, insert file
    if (file_status (fullpath) == 1)
    {
        %   move cursor to begin of next line before inserting
        %   so the inserted file will not mix with existing lines
        !if (down(1))
        {
            eol();
            insert_char ('\n');
        }
        insert_file (fullpath);
        return;
    }

    %   if control here, entry failed
    %   use entered directory to display files to pick up

    %   make sure pick up from listing is for insertion
    %InsertFromListing=1;

    dls (dir);
    flush ("To insert text file, highlight it and press ShiftEnter.");
}


define disk_open ()
%----------------------------------------------------------
%   Will prompt for file to open,
%   listing files if name not correct
%----------------------------------------------------------
{
    variable dir = Null_String, file = Null_String;

    %   ask user for file
    file = read_path_from_mini ("Open:");

    if (LASTKEY[0] == 23) return;
    %   discard reply ^w, ascii 23

    !if (strlen(file)) return;

    %   was entry a directory?
    if (file_status (file) == 2)
    {
        %   directory exists, open it
        dls (file);
        return;
    }

    %   clean up entry
    (dir, file) = parse_filename (file);

    variable fullpath = dircat (dir, file);

    %   if entry ok, load file
    if (strlen (fullpath) and file_status (fullpath) == 1)
    {
        () = find_file (fullpath);
        sue_defaults ();
        return;
    }

    %   if control here, entry failed
    %   use entered directory to display files to pick up

    if (strlen (file))
        flush ("Not found! Pick it up from listing.");

    dls (dir);
    %   go to listing
}


define disk_dirlist ()
%----------------------------------------------------------
%   Will display listing of current directory
%----------------------------------------------------------
{
    variable thisdir = Null_String;
    (,thisdir,,) = getbuf_info ();
    dls(thisdir);
}

define disk_treeview ()
%----------------------------------------------------------
%   Will display directory tree
%----------------------------------------------------------
{
        dtr ();
}


%**********************************************************
%   Exit editor prompting to save modified buffers.
%   If no buffers modified, still ask for confirmation.
%**********************************************************
define disk_exit ()
{
    if (jed_load_readonly)
    {
        quit_jed();
        return;
    }

    variable ch, caution=0;
    variable tmpbf = whatbuf ();

    forever
    {
        caution = buffer_modified ();
        if (caution) break;
        nextbuffer();
        if (tmpbf == whatbuf) break;
    }

    sw2buf(tmpbf);

    if (caution)
    {
        exit_jed ();
        %   this will prompt for saving
        return;
    }

    %   no modified buffer to be saved
    flush ("Exit Editor? (y/n)");
    ch = toupper (getkey);
    if (ch=='Y') exit_jed;
    else flush (" ");
}


%**********************************************************
%   MENUS & BASICS
%**********************************************************

%   for viewer, other menu and other keyboard

%   create the menus loading, activating and zapping suemenu

require("suemenu");
require("dls");
require("basics");

if (jed_load_readonly) viewer_load ();
else
{
    suemenu ();
    set_status_line (" %O %N %p  %F  |  F1=Help ", 1);
}

%**********************************************************
%   SETKEYS
%**********************************************************

%   Each buffer has a keymap attached, normally a custom
%   copy of the global keymap in editing buffers or a custom
%   copy of slim_kmap in special buffers. Keymaps dedicated
%   to the modifiers (a_kmap, c_kmap, s_c_kmap, s_kmap)
%   are not attached to a buffer and should not be
%   customized on the fly. Avoid trouble, keep them as you
%   define them here.
%
%   shifkey_pressed is the only modifier available
%   in sl-files. It is used sparsely and it is not
%   reflected in the keymaps. Its usage can be tracked
%   down from the help menu.

definekey ("select_pulldown_edit",      "e",        a_kmap);
definekey ("select_pulldown_file",      "f",        a_kmap);
definekey ("select_pulldown_help",      "h",        a_kmap);
definekey ("select_pulldown_search",    "s",        a_kmap);
definekey ("select_pulldown_xtras",     "x",        a_kmap);
definekey ("select_pulldown_system",    "y",        a_kmap);

definekey ("scroll_up",                 Key_Down,   a_kmap);
definekey ("end_of_trail",              Key_End,    a_kmap);
definekey ("display_doclist",           Key_F1,     a_kmap);
definekey ("back_to_prev_position",     Key_F2,     a_kmap);
definekey ("occur_functions",           Key_F3,     a_kmap);
definekey ("disk_exit",                 Key_F4,     a_kmap);
definekey ("toggle_wrap",               Key_F7,     a_kmap);
definekey ("ed_evalbuffer",             Key_F8,     a_kmap);
definekey ("go_to_line",                Key_F9,     a_kmap);
definekey ("insert_code_for_key",       Key_Ins,    a_kmap);
definekey ("prevbuffer",                Key_Left,   a_kmap);
definekey ("nextbuffer",                Key_Right,  a_kmap);
definekey ("scroll_down",               Key_Up,     a_kmap);

definekey ("ed_delete_word",            Key_BS,     c_kmap);
definekey ("del_to_eol",                Key_Del,    c_kmap);
definekey ("go_paragraph_down",         Key_Down,   c_kmap);
definekey ("go_end_buffer",             Key_End,    c_kmap);
definekey ("prghelp_keyword",           Key_F1,     c_kmap);
definekey ("block_restore_rectangle",   Key_F11,    c_kmap);
definekey ("block_restore_selection",   Key_F12,    c_kmap);
definekey ("go_try_match_again",        Key_F2,     c_kmap);
definekey ("search_occurrences_0",      Key_F3,     c_kmap);
definekey ("ed_wrap_or_skip",           Key_F7,     c_kmap);
definekey ("macro_play",                Key_F8,     c_kmap);

definekey ("duplicate_line",            Key_F9,     c_kmap);
definekey ("go_begin_buffer",           Key_Home,   c_kmap);
definekey ("block_copy",                Key_Ins,    c_kmap);
definekey ("go_word_left",              Key_Left,   c_kmap);
definekey ("go_word_right",             Key_Right,  c_kmap);
definekey ("smart_tabs",                Key_Tab,    c_kmap);
definekey ("go_paragraph_up",           Key_Up,     c_kmap);
definekey ("search_rx_forward",         "0",        c_kmap);
definekey ("search_rx_backward",        "1",        c_kmap);

definekey ("ed_untab",                  "^B",       s_c_kmap);
definekey ("block_append",              "^C",       s_c_kmap);
definekey ("search_backward",           "^F",       s_c_kmap);
definekey ("replace_rx_interactively",  "^H",       s_c_kmap);
definekey ("sort1(1)",                  "^KO",      s_c_kmap);
definekey ("set_quickmark",             "^Q",       s_c_kmap);
definekey ("block_cut_append",          "^X",       s_c_kmap);
definekey ("search_occurrences_1",      Key_F3,     s_c_kmap);
definekey ("ed_wrap_or_skip",           Key_F7,     s_c_kmap);
definekey ("macro_repeat",              Key_F8,     s_c_kmap);
definekey ("go_endbuffer_select",       Key_End,    s_c_kmap);
definekey ("go_beginbuffer_select",     Key_Home,   s_c_kmap);
definekey ("go_wordleft_select",        Key_Left,   s_c_kmap);
definekey ("go_wordright_select",       Key_Right,  s_c_kmap);

definekey ("block_cut",                 Key_Del,    s_kmap);
definekey ("go_down_select",            Key_Down,   s_kmap);
definekey ("go_endline_select",         Key_End,    s_kmap);
definekey ("newline",                   Key_Enter,  s_kmap);
definekey ("menu_last_entry",           Key_F10,    s_kmap);
definekey ("search_again_backward",     Key_F3,     s_kmap);
definekey ("disk_treeview",             Key_F4,     s_kmap);
definekey ("prevbuffer",                Key_F6,     s_kmap);
definekey ("duplicate_floor_char",      Key_F9,     s_kmap);
definekey ("go_beginline_select",       Key_Home,   s_kmap);
definekey ("block_paste ",              Key_Ins,    s_kmap);
definekey ("go_cursorleft_select",      Key_Left,   s_kmap);
definekey ("go_pagedn_select",          Key_PgDn,   s_kmap);
definekey ("go_pageup_select",          Key_PgUp,   s_kmap);
definekey ("go_cursorright_select",     Key_Right,  s_kmap);
definekey ("ed_backtab",                Key_Tab,    s_kmap);
definekey ("go_up_select",              Key_Up,     s_kmap);


setkey ("block_all",                    "^A");
setkey ("ed_quick_untab",               "^B");
setkey ("block_copy",                   "^C");
setkey ("evaluate_cmd",                 "^E");
setkey ("search_forward",               "^F");
setkey ("go_to",                        "^G");
setkey ("replace_interactively",        "^H");

setkey ("block_fill_rectangle",         "^KF");
setkey ("case_toggle_initial",          "^KI");
setkey ("case_lower",                   "^KL");
setkey ("trim",                         "^KM");
setkey ("sort1(0)",                     "^KO");
setkey ("block_save",                   "^KS");
setkey ("case_toggle",                  "^KT");
setkey ("case_upper",                   "^KU");

setkey ("repeat_command",               "^L");
setkey ("disk_new_file",                "^N");
setkey ("disk_open",                    "^O");
setkey ("go_to_quickmark",              "^Q");
setkey ("redraw",                       "^R");
setkey ("disk_save_buffer",             "^S");
setkey ("ed_undelete",                  "^U");
setkey ("block_paste",                  "^V");
setkey ("close_unconditionally",        "^W");
setkey ("block_cut",                    "^X");
setkey ("del_line",                     "^Y");
setkey ("undo",                         "^Z");

setkey ("ed_delete_backward",           Key_BS);
setkey ("insert_code_for_key",          Key_Ctrl_K_Ins);
setkey ("ed_delete_here",               Key_Del);
setkey ("go_down_vertically",           Key_Down);
setkey ("go_end_line",                  Key_End);
setkey ("select_pulldown_help",         Key_F1);
setkey ("block_toggle_rectangle",       Key_F11);
setkey ("block_toggle_lines",           Key_F12);
setkey ("go_to_matching_fence",         Key_F2);
setkey ("search_again_forward",         Key_F3);
setkey ("disk_dirlist",                 Key_F4);
setkey ("reset_screen",                 Key_F5);
setkey ("ed_nextbuffer",                Key_F6);
setkey ("ed_wrap_lines",                Key_F7);
setkey ("macro_start_or_stop",          Key_F8);
setkey ("duplicate_ceiling_char",       Key_F9);
setkey ("go_begin_line",                Key_Home);
setkey ("toggle_overwrite",             Key_Ins);
setkey ("go_cursor_left",               Key_Left);
setkey ("go_page_down",                 Key_PgDn);
setkey ("go_page_up",                   Key_PgUp);
setkey ("go_cursor_right",              Key_Right);
setkey ("ed_softtab",                   Key_Tab);
setkey ("go_up_vertically",             Key_Up);

setkey ("select_menubar", Key_F10);
%   F10 works as a toggle: it is defined in menu.c
%   to exit the menus and return to editing buffer



%%%%%%%%%%  TEST ZONE  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


define ttest()
%----------------------------------------------------------
%   Does nothing specific, used to test user code.
%   Not same as test() which is an intrinsic from intrin.c
%   used to test C code
%----------------------------------------------------------
{


}
setkey ("ttest", "^t");
