%**********************************************************
%
%   suemenu.sl defines and creates menubar and pulldowns
%   upon call from sue.sl.
%   It includes some functions from the Help pulldown
%   which are only used here.
%
%**********************************************************

%    _debug_info = 1;
%    _traceback = 1;


define display_help(stringA, stringB, stringC)
%----------------------------------------------------------
%   displays selected text of help file
%   relying on format explained in the help file
%   locates section stringA in first step
%   locates keyword stringB in final step
%----------------------------------------------------------
{
    variable found, ffile = "sue.hlp", hfile;

    hfile = expand_jedlib_file (ffile);

    !if (strlen (hfile))
    {
        flush ("File missing in Jed library directory");
        return;
    }

    %   open temp buffer for help
    variable suehelp = " /Help Temp Buffer/";
    %   string begins with space
    delbuf(suehelp);

    if (MINIBUFFER_ACTIVE)
    {
        pop2buf(suehelp);
        onewindow();
    }
    else sw2buf(suehelp);

    %   insert into it copy of help file
    insert_file (hfile);
    bob ();

    CASE_SEARCH=1;
    found = bol_fsearch(stringA+" [#]");

    !if (found)
    {
        if (bufferp(suehelp)) close_unconditionally ();
        flush ("Not found: " + stringA);
        return;
    }

    eol();
    found = fsearch(stringB);

    !if (found)
    {
        if (bufferp(suehelp)) close_unconditionally ();
        stringB = substr (stringB, 1, strlen(stringB)-4);
        flush ("Not found: " + stringB);
        return;
    }

    % both stringA and stringB found

    () = ffind("[#]");
    del_to_eol ();

    go_up(1);
    push_mark();
    bob();
    del_region();

    %   text above keyword deleted
    %   search for delimiter downstream

    %   some keywords are grouped together, skip through
    forever
    {
        go_down(1);
        !if (ffind(" [#]")) break;
    }

    %   find delimiting tag [#]
    !if (re_fsearch("[\[]+[#]+[\]]+$"))
    {
        flush ("Delimiter not found.");
        %   unlikely to happen
        return;
    }

    %   delimiter found, delete text below it
    bol();
    push_mark();
    eob();
    del_region();

    %   display help text at line 2
    if(bol_bsearch(stringB));
    else bob();
    recenter(2);

    set_buffer_modified_flag (0);
    set_readonly(1);
    set_status_line ("  Press any key to exit  ", 0);

    dummy_menu(stringC);

    update_sans_update_hook(0);
    flush_input(getkey ());
    close_unconditionally ();
}


define display_ro_keymap ()
%----------------------------------------------------------
%   slim_kmap is defined in keymap.c
%   and used repeatedly for different purposes
%   each time, it must be reset
%----------------------------------------------------------
{
    reset_slim_keymap ();
    $1 = "slim_k_readonly";
    !if (keymap_p($1)) copy_keymap ($1, "slim_kmap");
    use_keymap ($1);

    variable i;

    for (i=32; i<256; i++) undefinekey (char(i), $1);
    undefinekey (char(13), $1);

    definekey ("previous_line_cmd",         Key_Up,    $1);
    definekey ("next_line_cmd",             Key_Down,  $1);
    definekey ("bol",                       Key_Home,  $1);
    definekey ("eol",                       Key_End,   $1);
}


define display_hlp2 (stringA, stringB, stringC)
%----------------------------------------------------------
%   displays selected text of help file
%   relying on format explained in the help file
%   locates section stringA in first step
%   locates keyword stringB in final step
%----------------------------------------------------------
{
    variable found, ffile = "sue.hlp", hfile;

    hfile = expand_jedlib_file (ffile);

    !if (strlen (hfile))
    {
        flush ("File missing in Jed library directory");
        return;
    }

    %   open temp buffer for help
    variable suehelp = " /Help Snippets/";
    delbuf(suehelp);

    if (MINIBUFFER_ACTIVE)
    {
        pop2buf(suehelp);
        onewindow();
    }
    else sw2buf(suehelp);

    %   insert into it copy of help file
    insert_file (hfile);
    bob ();

    CASE_SEARCH=1;
    found = bol_fsearch(stringA+" [#]");

    !if (found)
    {
        if (bufferp(suehelp)) close_unconditionally ();
        flush ("Not found: " + stringA);
        return;
    }

    eol();
    found = fsearch(stringB);

    !if (found)
    {
        if (bufferp(suehelp)) close_unconditionally ();
        stringB = substr (stringB, 1, strlen(stringB)-4);
        flush ("Not found: " + stringB);
        return;
    }

    % both stringA and stringB found

    () = ffind("[#]");
    del_to_eol ();

    go_up(1);
    push_mark();
    bob();
    del_region();

    %   text above keyword deleted
    %   search for delimiter downstream

    %   some keywords are grouped together, skip through
    forever
    {
        go_down(1);
        !if (ffind(" [#]")) break;
    }

    %   find delimiting tag [#]
    !if (re_fsearch("[\[]+[#]+[\]]+$"))
    {
        flush ("Delimiter not found.");
        %   unlikely to happen
        return;
    }

    %   delimiter found, delete text below it
    bol();
    push_mark();
    eob();
    del_region();

    %   display help text at line 2
    if(bol_bsearch(stringB));
    else bob();
    recenter(2);

    %%  use slim_k_readonly
    display_ro_keymap ();

    set_buffer_modified_flag (0);
    set_readonly(1);
    set_status_line ("  Escape to exit  ", 0);

    dummy_menu(stringC);
}


define display_newcomers ()
%----------------------------------------------------------
%   One screen orientation for newcomers
%----------------------------------------------------------
{
    variable stringA = "PULLDOWN HELP";
    variable stringB = "    cursor  ";
    variable stringC = "&Quick Orientation for Newcomers";

    display_hlp2 (stringA,stringB,stringC);
}


define display_guide ()
%----------------------------------------------------------
%   Brief guide for newcomers
%----------------------------------------------------------
{
    variable ffile = "briefguide.hlp", hfile;

    hfile = expand_jedlib_file (ffile);

    !if (strlen (hfile))
    {
        flush ("File missing in Jed library directory");
        return;
    }

    %   open temp buffer for help
    variable sueguide = " /Brief Guide/";
    delbuf(sueguide);

    sw2buf(sueguide);

    %   insert into it copy of help file
    insert_file (hfile);
    bob ();
    set_buffer_modified_flag (0);
    set_readonly (1);
}


define display_cmd ()
%----------------------------------------------------------
%   Pressing a key will reveal the corresponding function
%----------------------------------------------------------
{
    variable f, type, tmp, modif;

    flush ("Press key, command will show...");

    (type, f) = get_key_binding ();

    variable ks = LASTKEY;

    if (f == NULL)
    {
        vmessage ("Key \"%s\" is undefined.",ks);
        return;
    }

    %   type -1 same as f NULL
    %   type is now 10*modifier plus traditional values
    %   as used in the switch below
    %   separate modifier and actual type

    tmp = type;

    type = type mod 10;
    %   actual type

    modif = (tmp-type)/10;
    %   full modifier

    ks = keyname (ks,modif);

    if      (modif==8) ks = "Alt "+ ks;
    else if (modif==5) ks = "ShiftCtrl " + ks;
    else if (modif==4) ks = "Ctrl "+ ks;
    else if (modif==1) ks = "Shift "+ ks;
    else if (modif)
    {
        %   should not happen
        flush("Not defined.");
        return;
    }

    switch (type)
    {
        case 0:
            if (1 == is_defined (f))
                vmessage
                ("Key \"%s\" runs the intrinsic function \"%s\".", ks, f);
            else
                vmessage
                ("Key \"%s\" runs the SLang function \"%s\".", ks, f);
    }

    {
        case 1:
            vmessage
            ("Key \"%s\" runs the internal function \"%s\".", ks, f);
    }

    {
        case 2:
            vmessage
            ("Key \"%s\" runs the keyboard macro \"%s\".", ks, f);
    }

    {
        case 3:
            vmessage ("Key \"%s\" inserts \"%s\".", ks, f);
    }
}


define display_doclist ()
%----------------------------------------------------------
%   Will list the jed docs directory.
%----------------------------------------------------------
{
    %   list the jed docs directory
    variable dir = dircat (JED_ROOT, "doc/txt/");
    dls (dir);
}


define display_aboutsue ()
%----------------------------------------------------------
%   Will display details of jed, sue, slang as used here
%----------------------------------------------------------
{
    variable found, ffile = "sue.hlp", hfile;
    variable stringA = "AboutSue";
    variable stringB = "Copyright";

    hfile = expand_jedlib_file (ffile);

    !if (strlen (hfile))
    {
        flush ("File missing in Jed library directory");
        return;
    }

    %   open temp buffer for help
    variable suehelp = "/Help Temp Buffer/";
    sw2buf(suehelp);
    erase_buffer ();

    %   insert into it copy of help file
    insert_file (hfile);
    bob ();

    CASE_SEARCH=1;
    found = bol_fsearch(stringA+" [#]");

    !if (found)
    {
        if (bufferp(ffile)) close_unconditionally ();
        flush (stringA + "  not found.");
        return;
    }

    eol();
    found = fsearch(stringB);

    !if (found)
    {
        if (bufferp(ffile)) close_unconditionally ();
        flush (stringB + " not found.");
        return;
    }

    recenter(2);
    bol ();
    newline ();
    newline ();
    go_up(1);
    goto_column(5);

    vinsert ("Version:        %s", _jed_version_string);
    newline ();
    goto_column(5);

    vinsert ("SLang Version:  %s", _slang_version_string);

    newline ();

    set_buffer_modified_flag (0);
    set_status_line ("  Press any key to exit  ", 0);
    update_sans_update_hook(0);

    flush_input(getkey ());

    close_unconditionally ();
}


define display_clean_manpage ()
%----------------------------------------------------------
%   Simplifies the man page format so the man page can
%   be shown in the editor.
%----------------------------------------------------------
{
    bob ();

    flush ("Processing man page...");

    %   remove _^H underscores
    replace ("_\010", Null_String);

    %   remove ovrstrike
    while ( fsearch ("\010") )
    {
        del ();
        del ();
    }
    flush (" ");
}


define display_manpage ()
%----------------------------------------------------------
%   Retrieves a man page and displays it in the editor
%   simplifying the format.
%----------------------------------------------------------
{
    variable subj=Null_String, buf = "manpage";
    variable frombuf = whatbuf ();

    subj = read_mini2 ("man", Null_String, Null_String);
    !if (strlen (subj)) return;

    buf = buf + " " + subj;

    sw2buf (buf);
    erase_buffer ();

    set_status_line (" %b %p %O F1=Help ", 0);

    variable rsc =
    () = run_shell_cmd (sprintf ("man %s 2> /dev/null", subj));

    if (rsc==0)
    {
        display_clean_manpage ();
        bob ();
        set_buffer_modified_flag (0);
        flush (" ");
    }

    else
    {
        reset_screen ();
        delbuf (buf);
        sw2buf (frombuf);
        flush ("Not found.");
    }
}


define display_info_doc ()
%----------------------------------------------------------
%   Will display an 'info' document in a buffer
%   with normal cursor movements, no node jumping
%----------------------------------------------------------
{
    variable
        subj    = Null_String,
        thisbf  = Null_String,
        tmpbf   = Null_String;

    subj = read_mini2 ("info", Null_String, Null_String);
    !if (strlen (subj)) return;

    thisbf = whatbuf ();
    tmpbf = "info " + subj;

    delbuf(tmpbf);
    sw2buf(tmpbf);
    set_status_line (" %b %p %O F1=Help ", 1);

    () = run_shell_cmd(tmpbf);

    !if (MINIBUFFER_ACTIVE) reset_screen();
    %   essential since 'info' is very renitent

    bob();
    if(is_whiteline)
    {
        sw2buf(thisbf);
        delbuf(tmpbf);
        flush("Not found.");
        return;
    }
    newline();

    bol_fsearch("* Index::");
    newline();
    push_mark();
    () = fsearch("*******");
    go_up(2);
    del_region ();

    while(bol_fsearch("* Menu:"))
    {
        push_mark();
        go_down(2);
        bol_fsearch("\n");
        del_region();
    }

    bob();
    while(bol_fsearch("File: ")) del_line();

    eob();
    if(bol_bsearch("*****"))
    {
        if (bsearch("Index"))
        {
            bol();
            push_mark();
            eob();
            del_region();
        }
    }

    bob();
    if (bol_fsearch("* Menu:"))
    {
        del_to_eol();
        insert("Chapters:");
        bob();
    }
}


define display_calendar ()
%----------------------------------------------------------
%   Will display a calendar at the bottom of scratch buffer,
%   year is prompted for.
%----------------------------------------------------------
{
    variable n, s=Null_String, yy;
    variable tmpbf = whatbuf();

    s = strtrim (read_mini2 ("Calendar for year: ","",""));

    %   if minibuffer entry not empty, use it
    %   otherwise use the current year

    if (strlen (s)) yy = s;
    else yy = strtok (time()," ")[4];

    %   year is in yy as string
    %   convert to integer for checking
    %   if yy returned from minibuffer is junk,
    %   conversion to integer is 0

    n = integer (yy);
    if (n < 1 or n > 9999) return;

    %   switch to scratch buffer to display calendar
    sw2buf("*scratch*");
    eob();
    newline();
    push_spot();

    () = run_shell_cmd ("cal " + yy);
    %   will insert output into jed's current buffer

    pop_spot();
    recenter(2);

    sw2buf(tmpbf);
    flush("Calendar appended to scratch buffer.");
}


define display_ascii_table ()
%----------------------------------------------------------
%   Will display an ascii table at the bottom of scratch buffer
%----------------------------------------------------------
{
    variable found, ffile = "sue.hlp", hfile;

    hfile = expand_jedlib_file (ffile);

    !if (strlen (hfile))
    {
        flush ("File missing in Jed library directory");
        return;
    }

    variable tmpbf = whatbuf();

    %   open temp buffer for help
    variable suehelp = "/Help Temp Buffer/";
    sw2buf(suehelp);
    erase_buffer ();

    %   insert into it copy of help file
    insert_file (hfile);
    bob ();

    found = fsearch("Oct   Dec   Hex   Char");

    !if (found)
    {
        if (bufferp(ffile)) close_unconditionally ();
        flush ("Not found.");
        return;
    }

    bol ();
    toggle_block(2);

    found = fsearch("077   63    3F");

    !if (found)
    {
        if (bufferp(suehelp)) close_unconditionally ();
        flush ("Not found.");
        return;
    }

    variable txt = bufsubstr ();

    close_unconditionally ();

    %   insert table in scratch buffer
    sw2buf("*scratch*");
    eob ();
    newline ();
    push_spot ();
    insert(txt);
    pop_spot ();

    sw2buf(tmpbf);
    flush("ASCII table appended to scratch buffer.");
}


define display_conversion ()
%----------------------------------------------------------
%   Converts between binary/octal/decimal/hexadecinal
%   using utility bc
%----------------------------------------------------------
{
    variable
        binarY  = "[0-1]+",
        octaL   = "[0-7]+",
        decimaL = "[0-9]+",
        heX     = "[0-9A-F]+",
        smatch  =  Null_String,
        sprompt =  "binary/octal/decimal/hexadecimal (b/o/d/x)? ",
        from =  Null_String,
        to   =  Null_String,
        what =  Null_String,
        txtF =  Null_String,
        txtT =  Null_String;


    from = read_mini2 ("Convert from " + sprompt, Null_String, Null_String);

    !if(strlen(from)) return;

    switch (strlow(from))

        {
            case "b":
            from = "2";
            smatch = binarY;
            txtF = "binary ";
        }

        {
            case "o":
            from = "8";
            smatch = octaL;
            txtF = "octal ";
        }

        {
            case "d":
            from = "A";
            smatch = decimaL;
            txtF = "decimal ";
        }

        {
            case "x":
            from = "16";
            smatch = heX;
            txtF = "hex ";
        }

        {
            flush("Wrong conversion origin.");
            return;
        }

    flush_input();
    to = read_mini2 ("Convert to " + sprompt, Null_String, Null_String);

    !if(strlen(from)) return;

    switch (strlow(to))

        {
            case "b":
            to = "2";
            txtT = "binary ";
        }

        {
            case "o":
            to = "8";
            txtT = "octal ";
        }

        {
            case "d":
            to = "A";
            txtT = "decimal ";
        }

        {
            case "x":
            to = "16";
            txtT = "hex ";
        }

        {
            flush("Wrong conversion target.");
            return;
        }

    flush_input();
    what = strup (read_mini2
    ("Digits only please: convert what? ", Null_String, Null_String));
    %   bc chokes on lower case letters

    !if(strlen(from)) return;

    if (string_match (what,smatch,1)==0)
    {
        flush("Wrong input.");
        return;
    }

    variable thisbf = whatbuf();

    %   input is ok
    %   generate pseudo-random name for buffer/file

    variable fn = string(_time);

    fn = strcat (fn, ".tmp");
    sw2buf (fn);
    set_readonly (0);
    erase_buffer();

    %   the repeated ibase/obase line is due to
    %   crappy undocumented bc behaviour

    insert ("ibase=A");
    newline();
    insert ("ibase=" + from);
    newline();
    insert ("obase=A");
    newline();
    insert ("obase=" + to);
    newline();
    insert (what);
    newline();
    insert ("quit");
    newline();

    () = write_buffer (fn);
    () = run_shell_cmd("bc -q "+ fn);
    %   bc's output is in buffer

    go_up(1);
    bol();
    push_mark();
    eol();

    variable cnvrtd = bufsubstr();
    set_buffer_modified_flag (0);
    delbuf(fn);
    sw2buf("*scratch*");
    eob();
    newline();
    insert(txtF + what);
    insert("    ");
    insert(txtT + cnvrtd);
    newline();

    sw2buf(thisbf);
    variable msgResult =
    txtF + what + "  " + txtT + cnvrtd;

    flush(msgResult + "    (details also in scratch buffer)");
}


define keyhelp_hlp ()
%----------------------------------------------------------
%   Quick help when navigating the Global, Ctrl, Alt keymap
%   currently disabled, status line can do
%----------------------------------------------------------
{
    variable
        stringA="KEYMAPS Global, Ctrl_Map, Alt_Map",
        stringB="    [#]",
        stringC="&Quick Help: Keybindings";

    display_hlp2 (stringA,stringB,stringC);
}


define keyhelp_cursor ()
%----------------------------------------------------------
%   Ancillary function: determines position of cursor
%   in groups plain, ctrl, alt of Global keymap
%----------------------------------------------------------
{
    variable r = 0;

    r = bol_fsearch ("Ctrl Keys");

    if (r) return 1;

    r = bol_fsearch ("Alt Keys");

    if (r) return 2;

    return 3;
}


define keyhelp_enter ()
%----------------------------------------------------------
%   Will display details about current key in the
%   Global, Ctrl, Alt keymap
%----------------------------------------------------------
{
    bol ();
    if (eolp) return;
    if (what_char () == ' ') return;

    variable
        where, stringA="", stringB="",
        stringC="&Keybindings Details";

    push_spot ();
    where = keyhelp_cursor ();
    pop_spot ();

    if      (where == 1) stringA="Global Keys";
    else if (where == 2) stringA="Ctrl Keys";
    else if (where == 3) stringA="Alt Keys";
    else return;

    goto_column (15);
    !if(eolp) begin_of_word_right();
    %   just in case the margin should be chnaged

    push_mark ();
    eol ();
    variable strb = strtrim(bufsubstr ());
    stringB = strb + " [#]";
    bol();

    if (strlen(strb))
        display_hlp2 (stringA,stringB,stringC);
}


define keyhelp_init ()
%----------------------------------------------------------
%   Initialize list for the Global, Ctrl, Alt keymap
%----------------------------------------------------------
{
    reset_slim_keymap ();
    $1 = "keyhelp_slimkmap";
    !if (keymap_p($1)) copy_keymap ($1, "slim_kmap");
    use_keymap ($1);

    variable i;
    for (i=32; i<256; i++) undefinekey (char(i), $1);

    undefinekey (Key_Right, $1);
    undefinekey (Key_Left, $1);

    definekey ("bob",                   Key_Home, $1);
    definekey ("eob",                   Key_End, $1);
    definekey ("keyhelp_enter",         Key_Enter, $1);
    definekey ("keyhelp_hlp",           Key_F1,$1);
    definekey ("page_down",             Key_PgDn, $1);
    definekey ("page_up",               Key_PgUp, $1);
    definekey ("prghelp_find_again",    Key_F3, $1);
    definekey ("prghelp_find",          "^F", $1);
}


define keyhelp_bindings ()
%----------------------------------------------------------
%   Some formatting for the keymap listings
%----------------------------------------------------------
{
    sw2buf ("/KeyBindings/");
    erase_buffer ();

    newline();
    goto_column(17);
    insert ("Press Enter on highlighted command to get details.");
    newline();
    newline();
    newline();

    insert ("Plain Keys");
    newline();
    newline();
    dump_bindings ("global");
    dump_bindings ("esc_kmap");
    eob ();

    newline();
    newline();
    insert ("Ctrl Keys");
    newline();
    newline();
    dump_bindings ("ctrl_kmap");
    eob ();

    newline();
    newline();
    insert ("Alt Keys");
    newline();
    newline();
    dump_bindings ("alt_kmap");
    eob ();

    newline();
    newline();
    insert ("Shift Keys");
    newline();
    newline();
    dump_bindings ("shift_kmap");
    eob ();

    newline();
    newline();
    insert ("ShiftCtrl Keys");
    newline();
    newline();
    dump_bindings ("shiftctrl_kmap");

    %   suppress lines containing self-insert characters
    bob ();
    while (fsearch ("self_insert_cmd")) delete_line ();

    readable_keys();

    %   sort plain keys, case-insensitive
    bob ();
    bol_fsearch ("^");
    %   the first keybindings is always a char 1-31
    %   represented by ^ with a letter
    push_mark ();
    while (0==is_whiteline) go_down(1);
    go_up(1);
    sort2 ();

    %   sort Ctrl keys, case-insensitive
    bob ();
    () = fsearch("Ctrl Keys");
    go_down(1);
    go_down(1);
    push_mark ();
    while (0==is_whiteline) go_down(1);
    go_up(1);
    sort2 ();

    %   sort Alt keys, case-insensitive
    bob();
    () = fsearch("Alt Keys");
    go_down(1);
    go_down(1);
    push_mark ();
    while (0==is_whiteline) go_down(1);
    go_up(1);
    sort2 ();

    %   sort Shift keys, case-insensitive
    bob();
    () = fsearch("Shift Keys");
    go_down(1);
    go_down(1);
    push_mark ();
    while (0==is_whiteline) go_down(1);
    go_up(1);
    sort2 ();

    %   sort ShiftCtrl keys, case-insensitive
    bob();
    () = fsearch("ShiftCtrl Keys");
    go_down(1);
    go_down(1);
    push_mark ();
    while (0==is_whiteline) go_down(1);
    go_up(1);
    sort2 ();

    bob ();
    () = fsearch("Press");

    keyhelp_init ();
    set_readonly(1);
    set_status_line
    ("  Escape: exit  |  F1: help  ",0);
    variable m = "&Keybindings Global";
    dummy_menu(m);
}


define keyhelp_quick ()
%----------------------------------------------------------
%   Quick help on Global keys
%----------------------------------------------------------
{
    variable stringA="Global";
    variable stringB="FILE/BUFFER";
    variable stringC="&Quick Help: Global";

    display_hlp2 (stringA,stringB,stringC);
}


define keyhelp_functions ()
%----------------------------------------------------------
%   Quick tabular help on function keys
%----------------------------------------------------------
{
    variable stringA="F-Keys";
    variable stringB="FUNCTION KEYS";
    variable stringC="&Function Keys";

    display_hlp2 (stringA,stringB,stringC);
}


define keyhelp_minibuffer ()
%----------------------------------------------------------
%   List of Minibuffer keymap
%----------------------------------------------------------
{
    variable
        stringA = "MiniBuffer",
        stringB = "F1",
        stringC = "&MiniBuffer Keybindings";

    display_help (stringA,stringB,stringC);
    %   display_hlp2 bad idea
}


define prghelp_clip ()
%----------------------------------------------------------
%   Ancillary function to Programming help
%----------------------------------------------------------
{
    push_spot ();
    bob ();
    toggle_block(2);
    eob ();
    block_copy ();
    pop_spot ();
}


define prghelp_find ()
%----------------------------------------------------------
%   Ancillary function to Programming help
%----------------------------------------------------------
{
    variable sstr=Null_String;
    sstr = read_mini2 ("Search for:", Null_String, Null_String);
    !if (strlen (sstr)) return;
    LAST_SEARCH=sstr;
    %() = fsearch (sstr);
    !if(fsearch(sstr)) flush("Not found.");
}


define prghelp_find_again ()
%----------------------------------------------------------
%   Ancillary function to Programming help
%----------------------------------------------------------
{
    !if (right(1)) return;
    !if(fsearch(LAST_SEARCH)) flush("Not found.");
}


define prghelp_init_display ()
%----------------------------------------------------------
%   Initialize listings in Programming help
%----------------------------------------------------------
{
    reset_slim_keymap ();
    $1 = "prghelp_d_slimkmap";
    !if (keymap_p($1)) copy_keymap ($1, "slim_kmap");
    use_keymap ($1);

    variable i;
    for (i=32; i<256; i++) undefinekey (char(i), $1);

    definekey ("previous_line_cmd",     Key_Up,     $1);
    definekey ("next_line_cmd",         Key_Down,   $1);
    definekey ("prghelp_clip",          "^C",       $1);
}


define prghelp_grep (strngK)
%----------------------------------------------------------
%   Ancillary function to Programming help
%   used when no specific documentation found for function
%   in this case source code for function will be displayed
%----------------------------------------------------------
{
    %   approach: grep for argument strngK
    %   in lib directory and discard prototypes
    %   normally function definition is found
    %   insert that file in current buffer
    %   display first 20 lines of function

    variable fmask = get_jed_library_path();
    fmask = substr(fmask,1, -2+strlen(fmask));
    fmask = fmask + "/*.sl";
    %   mask for files ending with "sl" in lib directory

    variable cmdstr = "grep -His " + strngK + fmask;

    if (0 != run_shell_cmd (cmdstr)) return -1;

    %   grep results are in current buffer

    %   we need to consider cases like 'compile' and 'compile_save'
    %   i.e. a short function name that may start other names
    %   we extract it from argument strngK
    %   strngK looks like 'define ....' with single quotes included
    %                     123456789

    variable shortn = substr(strngK, 9, -10 + strlen(strngK));
    variable lshortn = strlen(shortn);
    variable tmpcn;

    while (bsearch(shortn))
    {
        right(lshortn);
        tmpcn = what_column();
        skip_chars("( ");
        if (what_column() == tmpcn)
        {
            %   function name is longer than warranted
            bol();
            del_to_eol():
        }
        bol();
        if(bobp) break;
    }

    %   wipe out prototypes

    eob();
    while (re_bsearch(";$"))
    {
        bol();
        del_to_eol();
        if(bobp) break;
    }
    bob();

    %   get full path for file found, normally unique,
    %   and also the function name

    fmask = substr(fmask,1,-3+strlen(fmask));

    !if (fsearch(fmask)) return -1;

    push_mark();
    ffind(":");
    variable pathfl = bufsubstr();
    go_right(1);
    push_mark();
    eol();
    variable fnhead = bufsubstr();

    %   insert file found into current buffer

    erase_buffer();
    insert_file (pathfl);
    bob();
    () = re_fsearch("^"+fnhead+"$");

    %   delete text above location just found
    push_spot();
    if (up(1))
    {
        push_mark();
        bob();
        del_region();
    }
    pop_spot();

    %   keep max. 20 lines of function requested test

    push_spot();
    go_down (21);
    eol();

    push_mark();
    eob();
    del_region();
    pop_spot();
    recenter(2);
    hide_status_indicators(1);
    set_status_line ("  " + pathfl + "  ", 0);
    update_sans_update_hook(1);
    flush("Press any key to exit");
    flush_input(getkey);
    flush(" ");
}


define prghelp_display (stringK)
%----------------------------------------------------------
%   Ancillary function to Programming help
%   Will find documentation in file functions.hlp
%----------------------------------------------------------
{
    !if(strlen(stringK)) return;

    variable flag, found, ffile = "functions.hlp", hfile;

    hfile = expand_jedlib_file (ffile);

    !if (strlen (hfile))
    {
        flush ("File missing in Jed library directory");
        return;
    }

    %   open temp buffer for help
    variable prghelp = slash + string(_time) + slash;
    sw2buf(prghelp);
    erase_buffer ();

    variable m = "&Details of Functions/Variables";
    dummy_menu(m);


    %   insert into it copy of hfile
    insert_file (hfile);
    bob ();

    found = re_fsearch("^"+stringK+"$");

    !if (found)
    {
        erase_buffer();
        if (bufferp(prghelp)) close_unconditionally ();
        flush ("Not found: " + stringK);
        return;
    }

    %   delete text above location just found
    push_spot();
    if (up(1))
    {
        push_mark();
        bob();
        del_region();
        %   delete trailing info not needed here test
        () = re_fsearch("^----------");
        push_mark();
        eob();
        del_region():
    }
    pop_spot();

    recenter(2);

    set_readonly(1);

    update_sans_update_hook(0);

%   if (keymap_p("prghelp_d_slimkmap"))
%       use_keymap("prghelp_d_slimkmap");
%   else prghelp_init_display ();
    prghelp_init_display ();

    set_status_line
    ("   Escape: exit   |   Up/Down: move   |   ctrl-c: copy   |   ctrl-s: save  ",0);
}


define prghelp_enter ()
%----------------------------------------------------------
%   Will reqyest details for functions listed in Programming
%   help
%----------------------------------------------------------
{
    variable stringK="";

    if ("/8/"==whatbuf ()) return;
    %   no further details for sl-variables

    if (4 > what_line) return;

    bol();
    if (what_char () != ' ') return;
    begin_of_word_right ();
    push_mark ();
    eol ();
    stringK = strtrim(bufsubstr ());

    if (strlen(stringK)>2) prghelp_display (stringK);
    else flush("Keyword for search too short.");
}


define prghelp_list_hlp ()
%----------------------------------------------------------
%   Quick help when navigating lists from submenu Programming
%----------------------------------------------------------
{
    variable
        stringA="KEYMAPS Global, Ctrl_Map, Alt_Map",
        stringB="    [#]",
        stringC="&Quick Help: Intrinsics/Variables";

    display_hlp2 (stringA,stringB,stringC);
}


define prghelp_init ()
%----------------------------------------------------------
%   Initialize lists for submenu Programming help
%----------------------------------------------------------
{
    reset_slim_keymap ();
    $1 = "prghelp_slimkmap";
    !if (keymap_p($1)) copy_keymap ($1, "slim_kmap");
    use_keymap ($1);

    variable i;
    for (i=32; i<256; i++) undefinekey (char(i), $1);

    definekey ("bob",                   Key_Home, $1);
    definekey ("eob",                   Key_End, $1);
    definekey ("page_down",             Key_PgDn, $1);
    definekey ("page_up",               Key_PgUp, $1);
    definekey ("prghelp_enter",         Key_Enter, $1);
    definekey ("prghelp_find",          "^F", $1);
    definekey ("prghelp_find_again",    Key_F3, $1);
    definekey ("prghelp_list_hlp",      Key_F1,$1);
}


define prghelp_listings (s)
%------------------------------------------------
%   Will list active functions/variables,
%   based on the global keymap and customization.
%   Returns nothing.
%------------------------------------------------
{
    variable a = _apropos ("Global", "", s);

    a = a[array_sort(a, "strcmpi")];

    variable strbf = slash + string(s) + slash;
    sw2buf (strbf);
    erase_buffer ();

    variable lin2="", m;

    lin2 = "Press Enter on highlighted item to get details.";

    if (s==1)
        m = " &C-coded Intrinsics";

    else if (s==2)
        m = " &SL-coded Functions";

    else if (s==4)
        m = " &C-coded Variables";

    else if (s==8)
    {
        m = " &SL-coded Variables";
        lin2 = "For further details, try and grep the editor's library.";
    }

    else
    {
        close_unconditionally ();
        return;
    }

    dummy_menu(m);

    variable tmpstr = Null_String;
    foreach (__tmp(a))
    {
        insert ("    ");
        insert(());
        newline ();
    }

    if (s==2)
    {
        bob();
        while (fsearch("    dls_")) delete_line();
        bob();
        while (fsearch("    tree_")) delete_line();
    }

    bob ();
    newline();
    insert ("    ");
    insert (lin2);
    newline();
    newline();
    bob();
    () = fsearch("    ");
    set_readonly(1);

    prghelp_init ();

    set_status_line ("  %p  |  Escape to exit  |  F1=Help  ", 0);
}


define prghelp_intrinsics ()
%----------------------------------------------------------
%   Lists C coded intrinsics in submenu Programming help
%----------------------------------------------------------
{
    variable s=0x1;
    prghelp_listings (s);
}


define prghelp_slfunctions ()
%----------------------------------------------------------
%   Lists SL intrinsics in submenu Programming help
%----------------------------------------------------------
{
    variable s=0x2;
    prghelp_listings (s);
}


define prghelp_c_vars ()
%----------------------------------------------------------
%   Lists C coded variables in submenu Programming help
%----------------------------------------------------------
{
    variable s=0x4;
    prghelp_listings (s);
}


define prghelp_sl_vars ()
%----------------------------------------------------------
%   Lists SL coded variables in submenu Programming help
%----------------------------------------------------------
{
    variable s=0x8;
    prghelp_listings (s);
}


define prghelp_keyword ()
%----------------------------------------------------------
%   can be called from normal editing buffer
%   to pick up details about keyword in documentation
%   keyword is determined by selection or by current word
%----------------------------------------------------------
{
    if (what_block () !=0)
    {
        flush ("Selection unsuitable for keyword search.");
        return;
    }

    variable fflag=0, kword="";
    variable cstr=what_char ();

    if (markp)
    {
        dupmark ();
        kword = bufsubstr ();

        %   any whitespace in selection?
        if (is_substr(kword," ")) fflag=1;
        else if (is_substr(kword,"\t")) fflag=1;
        else if (is_substr(kword,"\n")) fflag=1;
        if (fflag)
        {
            flush ("Selection unsuitable for keyword search.");
            return;
        }
    }

    %   there is no mark,
    %   is cursor on a word?

    else !if (is_substr(WordSet,char(what_char ()))) return;

    %   cursor is on a word
    else
    {
        push_spot();
        select_word ();
        kword = strtrim(bufsubstr());
        pop_spot();
    }

    %   we got the keyword
    if (strlen(kword)>2) prghelp_display (kword);
    else flush("Keyword for search too short.");
}


%**********************************************************
%   Functions to list and activate color schemes
%**********************************************************


define color_enter ()
%----------------------------------------------------------
%   Used in the color selection to activate a color scheme
%----------------------------------------------------------
{
    if(3 > what_line()) return;
    bol ();
    if (what_char () != ' ') return;

    begin_of_word_right ();

    push_mark ();
    eol ();
    go_left(3);
    %   this to let the extension out

    variable cscheme = bufsubstr ();

    set_color_scheme(cscheme);

    close_unconditionally ();
}


define color_create_keymap ()
%----------------------------------------------------------
%   Defines the keys for the color selection screen.
%----------------------------------------------------------
{
    reset_slim_keymap ();
    $1 = "color_slimkmap";
    !if (keymap_p($1)) copy_keymap ($1, "slim_kmap");
    use_keymap ($1);

    undefinekey (Key_Right, $1);
    undefinekey (Key_Left,  $1);
    undefinekey (Key_Del,   $1);
    undefinekey (Key_BS,    $1);

    variable i;
    for (i=32; i<256; i++) undefinekey (char(i), $1);

    definekey ("bob",           Key_Home,  $1);
    definekey ("eob",           Key_End,   $1);
    definekey ("color_enter",   Key_Enter, $1);
}


define color_start ()
%----------------------------------------------------------
%   Starts the color selection screen.
%----------------------------------------------------------
{
    variable colorbf = string(_time);
    sw2buf (slash+colorbf+slash);
    erase_buffer ();

    variable lstD = listdir (Color_Scheme_Path);
    if (lstD == NULL) return;

    lstD = lstD[array_sort(lstD)];

    variable lenL = length(lstD);
    !if (lenL) return;

    variable i=0;

    for (i=0;i<lenL;i++)
        if (is_substr (lstD[i],".sl"))
        {
            insert ("    ");
            insert(lstD[i]);
            newline();
        }

    if (bobp)
    {
        delbuf(whatbuf());
        flush ("No color schemes found.");
        return;
    }

    bob ();
    newline ();
    insert ("    ");
    insert("Press Enter on highlighted color scheme to activate it.");
    newline();
    newline();
    bsearch("Press");

    color_create_keymap ();
    set_readonly(1);
    set_status_line("  Escape to exit  |  Up/Down to move  ",0);
    variable m = "&Color Schemes";
    dummy_menu(m);
}


%**********************************************************
%   Manage functions from modules 'required' via the menu
%   They have their own namespace 'AddOn'
%**********************************************************


define addon_list_hlp ()
%----------------------------------------------------------
%   Quick help when navigating list of added functions
%----------------------------------------------------------
{
    variable
        stringA="Added Functions",
        stringB="    [#]",
        stringC="&Quick Help: Added Functions";

    display_hlp2 (stringA,stringB,stringC);
}


define addon_enter ()
%----------------------------------------------------------
%   Will execute the highlighted function in the list of
%   Added Functions.
%----------------------------------------------------------
{
    variable stringK="";

    if (5 > what_line) return;

    bol();
    if (what_char () != ' ') return;
    begin_of_word_right ();
    push_mark ();
    eol ();
    stringK = strtrim(bufsubstr ());

    variable tmpbf = whatbuf();
    variable sstr = Null_String;
    sstr =
    read_mini2 ("Check command, press Enter to execute:", stringK, stringK);
    !if (strlen(sstr)) return;

    sstr = "AddOn->" + sstr;
    eval(sstr);
    delbuf(tmpbf);
}


define addon_init ()
%----------------------------------------------------------
%   Initialize the list of added functions
%----------------------------------------------------------
{
    reset_slim_keymap ();
    $1 = "addon_slimkmap";
    !if (keymap_p($1)) copy_keymap ($1, "slim_kmap");
    use_keymap ($1);

    variable i;
    for (i=32; i<256; i++) undefinekey (char(i), $1);

    definekey ("addon_enter",           Key_Enter, $1);
    definekey ("addon_list_hlp",        Key_F1,$1);
    definekey ("bob",                   Key_Home, $1);
    definekey ("eob",                   Key_End, $1);
    definekey ("page_down",             Key_PgDn, $1);
    definekey ("page_up",               Key_PgUp, $1);
    definekey ("prghelp_find",          "^F", $1);
    definekey ("prghelp_find_again",    Key_F3, $1);
}


define addon_functions ()
%------------------------------------------------
%   Will list for pick up functions added
%   when sl-modules are evaluated.
%------------------------------------------------
{
    variable a = _apropos (AddSpace, "", 2);
    if (a==NULL or length(a)==0)
    {
        flush("No functions have been added.");
        return;
    }

    a = a[array_sort(a, "strcmpi")];

    variable strbf = "/Added Functions/";
    delbuf (strbf);
    sw2buf (strbf);

    variable lin2="Press Enter to execute highlighted function.";
    variable lin3="Function must be suitable for execution by user.";
    variable m = " &Added Functions ";

    dummy_menu(m);

    foreach (__tmp(a))
    {
        insert ("    ");
        insert(());
        newline ();
    }

    bob ();
    newline();
    insert ("    ");
    insert (lin2);
    newline();
    insert ("    ");
    insert (lin3);
    newline();
    newline();
    bob();
    () = fsearch("Enter");
    set_readonly(1);

    addon_init ();
    set_status_line ("  %p  |  Escape to exit  |  F1=Help  ", 0);
}


%**********************************************************

define select_pulldown_file ()
{
    menu_select_menu ("Global.&File");
}


define select_pulldown_edit ()
{
    menu_select_menu ("Global.&Edit");
}


define select_pulldown_search ()
{
    menu_select_menu ("Global.&Search");
}


define select_pulldown_system ()
{
    menu_select_menu ("Global.S&ystem");
}


define select_pulldown_xtras ()
{
    menu_select_menu ("Global.&Xtras");
}


define select_pulldown_help ()
{
    menu_select_menu ("Global.&Help");
}


define suemenu ()
{
    variable m;

    m = "Global";
    menu_create_menu_bar (m);
    menu_use_menu_bar (m);

    menu_delete_items (m);
    menu_append_popup (m, "&File");
    menu_append_popup (m, "&Edit");
    menu_append_popup (m, "&Search");
    menu_append_popup (m, "S&ystem");
    menu_append_popup (m, "&Xtras");
    menu_append_popup (m, "&Help");

    m = "Global.&File";
    menu_delete_items (m);
    menu_append_item (m, "&New",    "disk_new_file");
    menu_append_item (m, "&Open",   "disk_open");
    menu_append_item (m, "&Close",  "close_unconditionally");
    menu_append_item (m, "&Save",   "disk_save_buffer");
    menu_append_item (m, "Save &As","disk_save_as");
    menu_append_item (m, "&Insert", "disk_read_and_insert");
    menu_append_separator (m);
    menu_append_item (m, "&Properties", "ed_properties");
    menu_append_item (m, "Recent &Files", "recent_files");
    menu_append_item (m, "Recent &Dirs", "recent_dirs");
    menu_append_separator (m);
    menu_append_item (m, "E&xit",   "disk_exit");
    menu_append_item (m, "&Quit",   "quit_jed");

    m = "Global.&Edit";
    menu_delete_items (m);
    menu_append_item  (m, "Select &All", "block_all");
    menu_append_item  (m, "&Cancel Selection", "block_cancel");
    menu_append_item  (m, "Toggle &Rectangle", "block_toggle_rectangle");
    menu_append_item  (m, "Toggle &Lines", "block_toggle_lines");
    menu_append_popup (m, "&Block Operations");
    menu_append_separator (m);
    menu_append_item  (m, "&Undo", "undo");
    menu_append_popup (m, "Ca&se");
    menu_append_popup (m, "&Macros");
    menu_append_popup (m, "M&ore");

    m = "Global.&Edit.&Block Operations";
    menu_append_item (m, "Cop&y", "block_copy");
    menu_append_item (m, "&Cut", "block_cut");
    menu_append_item (m, "&Paste", "block_paste");
    menu_append_item (m, "&Write to File", "block_save");
    menu_append_separator (m);
    menu_append_item (m, "&Fill Rectangle", "block_fill_rectangle");
    menu_append_item (m, "S&ort", "sort1(0)");

    m = "Global.&Edit.Ca&se";
    menu_delete_items (m);
    menu_append_item (m, "&Upper Case", "case_upper");
    menu_append_item (m, "&Lower Case", "case_lower");
    menu_append_item (m, "&Toggle Case","case_toggle");
    menu_append_item (m, "Toggle &Initial", "case_toggle_initial");

    m = "Global.&Edit.&Macros";
    menu_delete_items (m);
    menu_append_item (m, "Macro St&art", "macro_start_or_stop");
    menu_append_item (m, "Macro St&op", "macro_start_or_stop");
    menu_append_item (m, "Macro &Play", "macro_play");
    menu_append_item (m, "Macro &Repeat", "macro_repeat");

    m = "Global.&Edit.M&ore";
    menu_delete_items (m);
    menu_append_item (m, "&Toggle Wrap Mode", "toggle_wrap");
    menu_append_item (m, "&Wrap Some Lines", "ed_wrap_lines");
    menu_append_item (m, "&Character for Code", "ed_insert_literal");
    menu_append_item (m, "Code for &Key", "insert_code_for_key");
    menu_append_item (m, "&Untab", "ed_untab");
    menu_append_item (m, "&Read-only on/off", "ed_manage_readonly");

    m = "Global.&Search";
    menu_delete_items (m);
    menu_append_item (m,  "Search &Forward", "search_forward");
    menu_append_item (m,  "Search &Backward", "search_backward");
    menu_append_item (m,  "&Replace", "replace_interactively");
    menu_append_item (m,  "Repeat &Last Search", "search_again_forward");
    menu_append_separator (m);
    %menu_append_item (m,  "&Occurrences", "occur_list(0)");
    menu_append_item (m,  "&Occurrences", "search_occurrences_0");
    menu_append_popup (m, "Regular E&xpressions");
    menu_append_separator (m);
    menu_append_item (m,  "Set &Quickmark", "set_quickmark");
    menu_append_popup (m, "&Goto");

    m = "Global.&Search.Regular E&xpressions";
    menu_delete_items (m);
    menu_append_item (m, "Regex Search &Forward", "search_rx_forward");
    menu_append_item (m, "Regex Search &Backward", "search_rx_backward");
    menu_append_item (m, "Regex &Replace", "replace_rx_interactively");
    menu_append_item (m, "Regex &Occurrences", "occur_list(1)");

    m = "Global.&Search.&Goto";
    menu_delete_items (m);
    menu_append_item (m,  "&Line", "go_to_line");
    menu_append_item (m,  "&Column", "go_to_column");
    menu_append_item (m,  "&Matching Fence", "go_to_matching_fence");
    menu_append_item (m,  "&Previous Position", "back_to_prev_position");
    menu_append_item (m,  "&Quickmark", "go_to_quickmark");
    menu_append_item (m,  "mid-&Window", "goto_middle_of_window");

    m = "Global.S&ystem";
    menu_delete_items (m);
    menu_append_item (m, "&List Buffers", "bufed_start");
    menu_append_item (m, "&Previous Buffer", "prevbuffer");
    menu_append_item (m, "&Next Buffer", "nextbuffer");
    menu_append_separator (m);
    menu_append_item (m, "&Split on/off", "togglewindows");
    %menu_append_item (m, "&Other Window", "otherwindow");
    menu_append_item (m, "&Other Window", "smart_tabs");
    menu_append_separator (m);
    menu_append_item (m, "&Redraw", "redraw");
    menu_append_item (m, "SUE &Defaults", "sue_defaults");
    menu_append_item (m, "&Color Schemes", "color_start");
    menu_append_separator (m);
    menu_append_item (m, "&Internal Command", "evaluate_cmd");

    m = "Global.&Xtras";
    menu_delete_items (m);
    menu_append_item (m, "&ASCII Table", "display_ascii_table");
    menu_append_item (m, "Calen&dar", "display_calendar");
    menu_append_item (m, "Convert &Numbers", "display_conversion");
    menu_append_item (m, "&Toggle CR", "toggle_crmode");
    menu_append_item (m, "&Extensions", "addon_functions");
    menu_append_separator (m);
    menu_append_item (m, "&C Compile", "compile");
    menu_append_item (m, "&Highlight C Syntax", "c_mode");
    menu_append_item (m, "List &Functions", "occur_functions");
    menu_append_separator (m);
    menu_append_item (m, "&Shell Command",  "ed_shellcmd");

    m = "Global.&Help";
    menu_delete_items (m);
    menu_append_item (m, "&Here's the Beef", "display_newcomers");
    menu_append_item (m, "Brief &Guide", "display_guide");
    menu_append_popup (m, "&Keys");
    menu_append_popup (m, "&Programming");
    menu_append_item (m, "Which &Command", "display_cmd");
    menu_append_separator (m);
    menu_append_item (m, "Sundry &Docs", "display_doclist");
    menu_append_item (m, "&Man Pages", "display_manpage");
    menu_append_item (m, "&Info Docs", "display_info_doc");
    menu_append_separator (m);
    menu_append_item (m, "&About ...", "display_aboutsue");

    m = "Global.&Help.&Keys";
    menu_append_item (m, "&Quick", "keyhelp_quick");
    menu_append_item (m, "&General", "keyhelp_bindings");
    menu_append_item (m, "&Function Keys", "keyhelp_functions");
    menu_append_item (m, "&Shiftkey Specials", "occur_shiftkeys");
    menu_append_item (m, "&Minibuffer ", "keyhelp_minibuffer");

    m = "Global.&Help.&Programming";
    menu_append_item (m, "&Intrinsics",   "prghelp_intrinsics");
    menu_append_item (m, "SL &Functions", "prghelp_slfunctions");
    menu_append_item (m, "&C Variables",  "prghelp_c_vars");
    menu_append_item (m, "SL &Variables", "prghelp_sl_vars");
}



%**********************************************************
%   viewer functions
%**********************************************************

define viewer_pulldown_file ()
{
    menu_select_menu ("viewer.&File");
}


define viewer_pulldown_edit ()
{
    menu_select_menu ("viewer.&Edit");
}


define viewer_pulldown_search ()
{
    menu_select_menu ("viewer.&Search");
}


define viewer_pulldown_windows ()
{
    menu_select_menu ("viewer.&Windows");
}


define viewer_pulldown_help ()
{
    menu_select_menu ("viewer.&Help");
    %   it works in other keymaps so no need to bother
}


define viewer_enter ()
{
    bol();
    if (what_char () == ' ') return;
    %   enter pushed on an empty line

    variable
        stringA = "Viewer Keys",
        stringB = Null_String,
        stringC = "&Quick Help: Viewer Keys";

    goto_column (15);
    begin_of_word_right();
    %   just in case the margin should change

    push_mark ();
    eol ();
    stringB = strtrim(bufsubstr);

    if (strlen(stringB))
        display_hlp2 (stringA,stringB,stringC);
}


define viewer_close ()
{
    variable tmpbf = whatbuf ();

    nextbuffer ();

    if (whatbuf () == tmpbf)
    {
        variable rply=0;
        rply = get_y_or_n("Exit viewer");
        if (rply>0) exit_jed();
    }
    else delbuf(tmpbf);
}


define viewer_k_help ()
%----------------------------------------------------------
%   Initialize keymap for the list of viewer keybindings
%----------------------------------------------------------
{
    reset_slim_keymap ();
    $1 = "viewerhelp_slimkmap";
    !if (keymap_p($1)) copy_keymap ($1, "slim_kmap");
    use_keymap ($1);

    variable i;
    for (i=32; i<256; i++) undefinekey (char(i), $1);

    undefinekey (Key_Right, $1);
    undefinekey (Key_Left, $1);

    definekey ("bob",                   Key_Home, $1);
    definekey ("eob",                   Key_End, $1);
    definekey ("viewer_enter",          Key_Enter, $1);
    definekey ("keyhelp_hlp",           Key_F1,$1);
    definekey ("page_down",             Key_PgDn, $1);
    definekey ("page_up",               Key_PgUp, $1);
}


define viewer_help ()
{
    variable thiskmap = what_keymap ();

    sw2buf ("/ViewerKeymap/");
    erase_buffer ();

    viewer_k_help ();
    %   keymap for the keybindings list

    newline();
    goto_column(17);
    insert ("Press Enter on highlighted command to get details");
    newline();
    newline();

    dump_bindings (thiskmap);

    readable_keys();

    %   sort keys, case-insensitive
    bob ();
    () = fsearch("Press");
    go_down(1);
    begin_of_word_right ();

    push_mark ();
    eob();
    while (bolp) call ("backward_delete_char");
    %   cursor at end of last non-empty line

    sort2 ();

    bob ();
    () = fsearch("Press");

    set_readonly(1);

    set_status_line ("  %p  |  Escape to exit  |  F1=vHelp  ", 0);
    variable m = "&Keybindings Viewer";
    dummy_menu(m);
}


define viewer_keymap ()
{
    reset_slim_keymap ();
    $1 = "viewer_slimkmap";
    !if (keymap_p($1)) copy_keymap ($1, "slim_kmap");
    use_keymap ($1);

    variable i;
    for (i=1; i<256; i++) undefinekey (char(i), $1);

    definekey ("viewer_close",              "^W",       $1);
    definekey ("viewer_pulldown_file",      "f",        $1);
    definekey ("viewer_pulldown_edit",      "e",        $1);
    definekey ("viewer_pulldown_search",    "s",        $1);
    definekey ("viewer_pulldown_windows",   "w",        $1);
    definekey ("viewer_pulldown_help",      "h",        $1);
    definekey ("viewer_pulldown_help",      Key_F1,     $1);

    definekey ("select_menubar",            Key_F10,    $1);
    definekey ("previous_line_cmd",         Key_Up,     $1);
    definekey ("next_line_cmd",             Key_Down,   $1);
    definekey ("scroll_up",                 Key_Left,   $1);
    definekey ("scroll_down",               Key_Right,  $1);
    definekey ("bob",                       Key_Home,   $1);
    definekey ("eob",                       Key_End,    $1);
    definekey ("page_down",                 Key_PgDn,   $1);
    definekey ("page_up",                   Key_PgUp,   $1);
    definekey ("block_toggle_lines",        Key_F12,    $1);
    definekey ("search_again",              Key_F3,     $1);
    definekey ("reset_screen",              Key_F5,     $1);
    definekey ("prevbuffer",                Key_F6,     $1);

    definekey ("block_all",                 "^A",       $1);
    definekey ("disk_open",                 "^O",       $1);
    definekey ("exit_jed",                  "^KE",      $1);
    definekey ("go_quickmark",              "^Q",       $1);
    definekey ("go_to",                     "^G",       $1);
    definekey ("redraw",                    "^R",       $1);
    definekey ("repeat_command",            "^L",       $1);
    definekey ("search_anywhere",           "^F",       $1);
    definekey ("write_buffer",              "^S",       $1);
}


define viewer_init ()
{
    if (keymap_p("viewer_slimkmap"))
    {
        use_keymap("viewer_slimkmap");
        return;
    }

    %   if keymap does not exist, create it
    viewer_keymap ();
}


define viewer_menu ()
{
    variable m;

    m = "viewer";
    menu_create_menu_bar (m);
    menu_use_menu_bar (m);

    menu_delete_items (m);
    menu_append_popup (m, "&File");
    menu_append_popup (m, "&Edit");
    menu_append_popup (m, "&Search");
    menu_append_popup (m, "&Windows");
    menu_append_popup (m, "&Help");

    m = "viewer.&File";
    menu_delete_items (m);
    menu_append_item (m, "&Open",   "disk_open");
    menu_append_item (m, "&Close",  "viewer_close");
    menu_append_item (m, "Save &As","disk_save_as");
    menu_append_separator (m);
    menu_append_item (m, "&Previous Buffer", "prevbuffer");
    menu_append_item (m, "&Next Buffer", "nextbuffer");
    menu_append_separator (m);
    menu_append_item (m, "&Quit", "quit_jed");

    m = "viewer.&Edit";
    menu_delete_items (m);
    menu_append_item  (m, "Select &All", "block_all");
    menu_append_item  (m, "Toggle &Selection", "block_toggle_lines");
    menu_append_item (m, "&Write to File", "block_save");

    m = "viewer.&Search";
    menu_delete_items (m);
    menu_append_item (m,  "Search &Forward", "search_forward");
    menu_append_item (m,  "Search &Backward", "search_backward");
    menu_append_item (m,  "Repeat &Last Search", "search_again");
    menu_append_separator (m);
    menu_append_item (m,  "&Occurrences", "occur_list(0)");
    menu_append_popup (m, "Regular E&xpressions");
    menu_append_separator (m);
    menu_append_item (m,  "Set &Quickmark", "go_set_quickmark");
    menu_append_popup (m, "&Goto");

    m = "viewer.&Search.Regular E&xpressions";
    menu_delete_items (m);
    menu_append_item (m, "Regex Search &Forward", "search_rx_forward");
    menu_append_item (m, "Regex Search &Backward", "search_rx_backward");
    menu_append_item (m, "Regex &Occurrences", "occur_list(1)");

    m = "viewer.&Search.&Goto";
    menu_delete_items (m);
    menu_append_item (m,  "&Line", "go_to_line");
    menu_append_item (m,  "&Column", "go_to_column");
    menu_append_item (m,  "&Matching Fence", "go_to_matching_fence");
    menu_append_item (m,  "&Previous Position", "back_to_prev_position");
    menu_append_item (m,  "&Quickmark", "go_to_quickmark");
    menu_append_item (m,  "mid-&Window", "goto_middle_of_window");

    m = "viewer.&Windows";
    menu_delete_items (m);
    menu_append_item (m, "&Redraw", "redraw");
    menu_append_item (m, "&Color Schemes", "color_start");

    m = "viewer.&Help";
    menu_delete_items (m);
    menu_append_item (m, "&Keys", "viewer_help");
    menu_append_item (m, "Sundry &Docs", "display_doclist");
    menu_append_item (m, "Unix &Man Pages", "display_manpage");
}


define viewer_load()
{
    if (is_substr(whatbuf(),"/")) return;

    viewer_menu ();
    viewer_keymap ();
    viewer_init ();
    set_status_line ("  %F  %p  |  F1=Help  ", 1);
}
if (jed_load_readonly)
    add_to_hook ("_jed_switch_active_buffer_hooks", &viewer_load);
