/* Copyright (c) 1992, 1998, 2000, 2002 John E. Davis
 * This file is part of JED editor library source.
 *
 * You may distribute this file under the terms the GNU General Public
 * License.  See the file COPYING for more information.
 */
#include "config.h"
#include "jed-feat.h"
/*{{{ Include Files */

#ifdef __WIN32__
/* This needs to go first before stdio is included. */
# include <windows.h>
# if !defined(__MINGW32__) && !defined(__CYGWIN32__)
#  define sleep Sleep
# endif
#endif


#include <stdio.h>
#include <slang.h>

#include "jdmacros.h"

#ifdef HAVE_STDLIB_H
# include <stdlib.h>
#endif

#ifdef REAL_UNIX_SYSTEM
# include <sys/types.h>
# include <sys/stat.h>
#endif

#include <errno.h>

#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif

#include <string.h>
#include <setjmp.h>
#ifdef __unix__
# ifdef SYSV
#   include <sys/types.h>
#   include <fcntl.h>
# endif
# include <sys/file.h>
#endif

#ifdef __MSDOS__
# include <io.h>
# if !defined(__WATCOMC__)
#  include <dir.h>
#  ifdef __BORLANDC__
extern unsigned _stklen = 10000U;
#  endif
# else
#  include <direct.h>
# endif
#endif

#if (defined(__WIN32__) && !defined(__CYGWIN32__)) || defined(__IBMC__)
# include <io.h>
# include <direct.h>
#endif

#if defined(__DECC) && defined(VMS)
#include <unixio.h>
#include <unixlib.h>
#endif

#ifndef O_RDWR
#ifndef VMS
# include <fcntl.h>
#endif
#endif

#include "file.h"
#include "buffer.h"
#include "display.h"
#include "sysdep.h"
#include "sig.h"
#include "misc.h"
#include "keymap.h"
#include "screen.h"
#include "ledit.h"
#include "search.h"
#include "text.h"
#include "hooks.h"
#include "paste.h"
#include "version.h"
#include "indent.h"
#include "colors.h"

#ifdef __WIN32__
# include "win32.h"
#endif


int Jed_Secure_Mode;
int Jed_Load_Quietly = 0;
#if JED_HAS_UTF8_SUPPORT
int Jed_UTF8_Mode;
#endif

int Jed_Load_Readonly = 0;
// if 1, files will open readonly

char Jed_Root_Dir [JED_MAX_PATH_LEN];

typedef struct
{
   jmp_buf b;
}

jmp_buf_struct;
extern jmp_buf_struct Jump_Buffer, *Jump_Buffer_Ptr;

int Batch = 0;

#include "j2s.h"
void revise_keys();
#include "main_k.c"
//

char *extract_path(char *pfile);


///////////////////////////////////////////////////////////

char *get_cwd (void)
{
   static char cwd[JED_MAX_PATH_LEN];
   char *c;

#ifdef HAVE_GETCWD
# if defined (__EMX__)
   c = _getcwd2(cwd, sizeof(cwd));
   /* includes drive specifier */
# else
   c = getcwd(cwd, sizeof(cwd));
   /* djggp includes drive specifier */
# endif
#else
   c = (char *) getwd(cwd);
#endif

   if (c == NULL)
     {
#ifndef REAL_UNIX_SYSTEM
        exit_error ("Unable to get cwd.", 0);
#else
        struct stat st1, st2;

        if ((NULL == (c = getenv ("PWD")))
            || (-1 == stat (c, &st1))
            || (-1 == stat (".", &st2))
            || (st1.st_dev != st2.st_dev)
            || (st1.st_ino != st2.st_ino))
          {
             jed_vmessage (1, "Unable to get CWD.  Assuming /");
             if (-1 == chdir ("/"))
               exit_error ("Unable to change directory to /", 0);
             c = "/";
          }

        strncpy (cwd, c, sizeof (cwd));
        cwd [sizeof (cwd) - 1] = 0;
#endif
     }

#ifndef VMS
   slash2slash(cwd);
   fixup_dir(cwd);
#endif
   return(cwd);
}


char *extract_path(char *pfile)
{
    //  argument is path and filename of executable

    char *f;

    //  pointer to end of full path, filename included
    f = pfile + strlen(pfile);

    //  skip backwards to closest slash
    while (f > pfile)
    {
        f--;
        if (*f == '/') break;
    }

    //  f is pointing to slash before filename
    //  truncate array in order to get path only
    pfile[f-pfile] = '\0';

    return pfile;
}


static void set_jed_root (void)
{
    char *jr;
    char jed_lib [JED_MAX_PATH_LEN + 1];

    *jed_lib = 0;

    readlink ("/proc/self/exe", Jed_Root_Dir, sizeof (Jed_Root_Dir));

    jr = extract_path(Jed_Root_Dir);

    if (jr != NULL)
    {
        strcpy (Jed_Root_Dir, jr);
        strcpy (jed_lib, jr);
        fixup_dir (jed_lib);
        strcat (jed_lib, "lib");
    }

    Jed_Library = SLang_create_slstring (jed_lib);
    if (Jed_Library == NULL) exit_error ("Out of memory", 0);
}


void (*X_Init_Global_Structures_Hook)(void);

int Stdin_Is_TTY;

int _Jed_Startup_Argc;
char **_Jed_Startup_Argv;

static int main_initialize (int argc, char **argv)
{
    char *cmd_hook = ".()command_line_hook";
    int i, fd, s=6;
    char *script_file;

    _Jed_Startup_Argv = argv;
    _Jed_Startup_Argc = argc;

#ifdef __EMX__
    _response(&argc, &argv);
    _wildcard(&argc, &argv);
#endif


#if JED_HAS_DISPLAY_TABLE
    // there is no display table, meaning what?
    for (i = 0; i < 256; i++) Output_Display_Table[i] = i;
#endif

    if ((argc > 1) && strcmp(argv[1], "-viewer")==0)
    {

    //  Jed_Secure_Mode = 1;
    //  cannot do with it because it stops shell commands
    //  find remedy for cmds.c, jproces.c, ledit.c, intrin.c
        Jed_Load_Readonly = 1;
    }

    // if shiftstate not available, run as a viewer
    if (ioctl (0, 0x541C, &s) != 0)
    {
        flush_message ("Sue is running in a subprocess as a viewer!");
        Jed_Load_Readonly = 1;
    }
    else revise_keys();
    // from main_k.c to revise annoyances from keymap

   /* If this hook is defined, let it peel off what ever arguments
      it wants.  It should return the number of remaining arguments */

    if (X_Argc_Argv_Hook != NULL)
    {
        i = (*X_Argc_Argv_Hook)(argc, argv) - 1;
        argv[i] = argv[0];
        argc -= i;
        argv += i;
    }

   /* If jed-script is a symbolic link to jed, then treat this as equivalent
    * to jed -script args...
    */
    script_file = NULL;

    if (0 == strcmp (extract_file (argv[0]), "jed-script"))
    {
        Batch = 2;
        SLang_Traceback = 1;
        Jed_Load_Quietly = 1;
        if (argc > 1) script_file = argv[1];
        else
        {
            fprintf (stderr, "Usage: jed-script SCRIPT_FILE [args...]\n");
            return -1;
        }
        argv++;
        /* make argv[0] the name of the script */

        argc--;
    }

    else if (argc > 1)
    {
        char *argv1 = argv[1];

        /* Allow, e.g., --batch */
        if ((argv1[0] == '-') && (argv1[1] == '-'))
            argv1++;

        // batch/script code expunged
    }

#if JED_HAS_UTF8_SUPPORT
    Jed_UTF8_Mode = SLutf8_enable (-1);
    if (Jed_UTF8_Mode == 0)
    {
        char *u = getenv ("JED_UTF8");
        if (u != NULL)
        {
            if ((0 == strcmp (u, "1")) || (0 == strcmp (u, "FORCE")))
            {
                Jed_UTF8_Mode = 1;                   /* force jed to use UTF-8 internally */
                SLsmg_utf8_enable (1);
                SLinterp_utf8_enable (1);
            }
            if (0 == strcmp (u, "FORCE")) SLtt_utf8_enable (1);
        }
    }
#endif

// patch insetion ??????
#if JED_HAS_UTF8_SUPPORT
   if (Jed_UTF8_Mode)
     define_word ("\\w");
   else
     {
#endif
#ifdef __MSDOS__
       define_word("0-9a-zA-Z\200-\232\240-\245\341-\353");
#else
       define_word("0-9a-zA-Z\300-\326\330-\366\370-\377");
#endif
#if JED_HAS_UTF8_SUPPORT
     }
#endif

    //if (Batch == 0) (*tt_get_terminfo) ();
    (*tt_get_terminfo) ();

    fd = 2;
    /* 2 is stderr, assume it is ok */

#if defined(REAL_UNIX_SYSTEM) && !defined(__APPLE__)
    if (Batch == 0)
    {
        int dev_tty_fd;

        if ((dev_tty_fd = open("/dev/tty", O_RDWR)) >= 0)
        {
            fd = dev_tty_fd;
            while (-1 == dup2(fd, 2))
            /* tty uses 2 as the descriptor */
            {
# ifdef EINTR
                if (errno == EINTR) continue;
# endif
                fd = 2;
                break;
            }
            close (dev_tty_fd);
        }
    }
#endif /* REAL_UNIX_SYSTEM */

    if (-1 == init_tty ()) exit_error ("Unable to initialize tty.", 0);

    jed_init_display ();
    /* sets up virtual screen */

    (*tt_set_mono) (JMESSAGE_COLOR, NULL, 0);
    (*tt_set_mono) (JERROR_COLOR, NULL, SLTT_BOLD_MASK);
    (*tt_set_mono) (JDOLLAR_COLOR, NULL, SLTT_BOLD_MASK);
    (*tt_set_mono) (JDOTS_COLOR, NULL, SLTT_BOLD_MASK);

    (*tt_set_mono) (JMENU_CHAR_COLOR, NULL, SLTT_BOLD_MASK);
    (*tt_set_mono) (JMENU_SELECTION_COLOR, NULL, SLTT_REV_MASK|SLTT_BOLD_MASK);
    (*tt_set_mono) (JMENU_POPUP_COLOR, NULL, 0);
    (*tt_set_mono) (JMENU_SELECTED_CHAR_COLOR, NULL, SLTT_BOLD_MASK);

    (void) jed_set_color (JMENU_SHADOW_COLOR, "black", "black");

    (void) jed_set_color (JMENU_POPUP_COLOR, "lightgray", "blue");

    (*tt_set_mono) (JMENU_SHADOW_COLOR, NULL, 0);
    (void) jed_set_color (JMENU_SHADOW_COLOR, "black", "black");

    init_syntax_tables ();

    /* This order here is crucial! */
    init_keymaps();

    if (NULL == (CBuf = make_buffer(NULL, NULL, NULL)))
    {
        exit_error("main: Allocation Failure", 0);
    }
    CLine = NULL;

    init_minibuffer();

    set_file_modes();

    /* what if someone pipes something to jed, allow it unless if is a
    * jed script.
    */

    if (Stdin_Is_TTY == 0) Stdin_Is_TTY = isatty (0);
    if ((Stdin_Is_TTY == 0) && (Batch != 2))
    /* 1 if stdin is a terminal, 0 otherwise */
    {
        set_buffer("tmp_stdin");
        read_file_pointer(fileno(stdin));
        bob();
        fclose(stdin);
        dup2(fd, 0);
    }

    if (CLine == NULL) make_line(25);
    Point = 0;
    window_buffer(CBuf);

#if !defined(IBMPC_SYSTEM) && !defined(VMS)
    init_signals();
#endif

    (void) SLang_set_argc_argv (argc, argv);

    // no batch condition here
    if ((0 != jed_ns_load_file("site", NULL)))
        return -1;

    // batch code expunged

    {
        if (SLang_is_defined(cmd_hook + 3))
        {
            SLang_run_hooks(cmd_hook + 3, 0);

            //if (SLang_Error && Batch) return -1;

             /* This has the effect of removing the memory that cmd_hook
              * is using.
              */
            SLang_load_string (cmd_hook);
        }
        else if (!SLang_get_error() && (argc > 2))
        {
            find_file_in_window(argv[2]);
        }
    }

    //if (SLang_Error && Batch) return -1;

    /* after we have possibly loaded key definitions, we can fix up
    * the minibuffer map. This way user definitions are used.
    */
    jed_setup_minibuffer_keymap ();

    return 0;
}


int main(int argc, char **argv)
{
    if (SLang_Version < SLANG_VERSION)
    {
        fprintf(stderr,
        "***Warning: Executable compiled against S-Lang %d but linked to %d\n",
        SLANG_VERSION, SLang_Version);
        fflush (stderr);
        sleep (2);
    }

    SLang_set_error (0);
    //

    //set_jed_root (argv[0]);
    set_jed_root();

    Jump_Buffer_Ptr = &Jump_Buffer;
    /* incase something goes wrong before we even get started... */
    if (setjmp(Jump_Buffer_Ptr->b) != 0)
    {
        /* hmm.... just exit I guess */
        exit_error("main: Fatal Error", 0);
        exit (1);
    }

    if (0 != main_initialize (argc, argv))
    {
        jed_reset_display();
        reset_tty();
        return SLang_get_error();
    }

    /* edit_loop */
    //if (!Batch) jed();
    jed();
    /* never returns */

    /* exit_error ("Batch End with no exit", 0); */
    jed_reset_display();
    reset_tty();
    // unix.c, not working, sometimes cursor will go missing

    SLtt_flush_output ();
    system("reset");
    //
    return SLang_get_error();
}
