//---------------------------------------------------------
//  this file redefines several keys in the linux keymap
//  some are usually reserved for the operating system
//  some are unix annoyances
//  some are remedial, should defaults have been changed
//  read kmap_worries.txt for more details
//---------------------------------------------------------

#include <linux/kd.h>
#include <linux/keyboard.h>

#define SHIFT_PRESSED       1
#define CTRL_PRESSED        4
#define SHIFTCTRL_PRESSED   5
#define ALT_PRESSED         8

void revise_keys()
{
    struct kbentry ke;

    // manpage console_ioctl
    // struct kbentry
    // {
    //     unsigned char kb_table;
    //     unsigned char kb_index;
    //     unsigned short kb_value;
    // };

    //struct kbsentry ks;
    //     struct kbsentry
    //     {
    //         u_char kb_func;
    //         u_char kb_string[512];
    //     };

    int i, fd=fileno(stdin);

    int kmap[110];
    // array index is scancode, array value is action code
    // array start at 0 but it is only used from 1
    // or indeed from 2 because 1 is escape and is
    // not to be touched.
    // we stop at 110 because PageDown (109) is
    // the highest key we want to change.

    ke.kb_table=0;
    // we are reading only plain keys, no modifier

    for (i=2;i<110;i++)
    {
        ke.kb_index=i;

        ioctl(fd,KDGKBENT,(unsigned int)&ke);
        // disregard errors since nothing can be done
        // we already tested TIOCLINUX successfully

        // the first two entries in structure ke are now set
        // third entry is filled in by kernel
        // we save it in our kmap array

        kmap[i]=ke.kb_value;

        // this loop does not change anything, we are reading
    }

    // we now have the plain action values for most keys.
    // for selected keys (essentially function keys
    // and digits), we set some modifier action values
    // equal to the plain action values.
    // this way, we get rid of fossilized unix crap
    // in international variety (e.g. french and german
    // unix crap is not identical to the us counterpart)
    // and enable all the selected keys to be recognised
    // through TIOCLINUX.
    // all despite the dire warnings in manpage
    // console_ioctl which has been threatening
    // fire and punishement for over 10 years.

    for (i=2;i<12;i++)
    {
        // these keys are the digits in PC keyboards

        ke.kb_index=i;
        ke.kb_value=kmap[i];

        ke.kb_table=CTRL_PRESSED;
        ioctl(fd,KDSKBENT,(unsigned int)&ke);

        ke.kb_table=SHIFTCTRL_PRESSED;
        ioctl(fd,KDSKBENT,(unsigned int)&ke);

        ke.kb_table=ALT_PRESSED;
        ioctl(fd,KDSKBENT,(unsigned int)&ke);
    }

    for (i=40;i<42;i++)
    {
        //  apostrophe and grave in the us keyboard
        //  this is a serious change in respect to i18n
        //  but we had to get rid of US Unix crap

        ke.kb_index=i;
        ke.kb_value=kmap[i];

        ke.kb_table=CTRL_PRESSED;
        ioctl(fd,KDSKBENT,(unsigned int)&ke);

        ke.kb_table=SHIFTCTRL_PRESSED;
        ioctl(fd,KDSKBENT,(unsigned int)&ke);

        ke.kb_table=ALT_PRESSED;
        ioctl(fd,KDSKBENT,(unsigned int)&ke);
    }

    for (i=51;i<54;i++)
    {
        //  comma, period, slash in the us keyboard
        //  this is serious in respect to i18n
        //  but we had to get rid of US Unix crap

        ke.kb_index=i;
        ke.kb_value=kmap[i];

        ke.kb_table=CTRL_PRESSED;
        ioctl(fd,KDSKBENT,(unsigned int)&ke);

        ke.kb_table=SHIFTCTRL_PRESSED;
        ioctl(fd,KDSKBENT,(unsigned int)&ke);

        ke.kb_table=ALT_PRESSED;
        ioctl(fd,KDSKBENT,(unsigned int)&ke);
    }

    for (i=59;i<69;i++)
    {
        // F1, F2,.., F10

        ke.kb_index=i;
        ke.kb_value=kmap[i];

        ke.kb_table=SHIFT_PRESSED;
        ioctl(fd,KDSKBENT,(unsigned int)&ke);

        ke.kb_table=CTRL_PRESSED;
        ioctl(fd,KDSKBENT,(unsigned int)&ke);

        ke.kb_table=SHIFTCTRL_PRESSED;
        ioctl(fd,KDSKBENT,(unsigned int)&ke);

        ke.kb_table=ALT_PRESSED;
        ioctl(fd,KDSKBENT,(unsigned int)&ke);
    }

    for (i=87;i<89;i++)
    {
        // F11, F12

        ke.kb_index=i;
        ke.kb_value=kmap[i];

        ke.kb_table=SHIFT_PRESSED;
        ioctl(fd,KDSKBENT,(unsigned int)&ke);

        ke.kb_table=CTRL_PRESSED;
        ioctl(fd,KDSKBENT,(unsigned int)&ke);

        ke.kb_table=SHIFTCTRL_PRESSED;
        ioctl(fd,KDSKBENT,(unsigned int)&ke);

        ke.kb_table=ALT_PRESSED;
        ioctl(fd,KDSKBENT,(unsigned int)&ke);
    }

    for (i=103;i<110;i++)
    {
        // all the arrows, PageUp, PageDown
        // most settings here below are suplerfluous
        // but so the code is easier to type and read

        ke.kb_index=i;
        ke.kb_value=kmap[i];

        ke.kb_table=SHIFT_PRESSED;
        ioctl(fd,KDSKBENT,(unsigned int)&ke);

        ke.kb_table=CTRL_PRESSED;
        ioctl(fd,KDSKBENT,(unsigned int)&ke);

        ke.kb_table=SHIFTCTRL_PRESSED;
        ioctl(fd,KDSKBENT,(unsigned int)&ke);

        ke.kb_table=ALT_PRESSED;
        ioctl(fd,KDSKBENT,(unsigned int)&ke);
    }

    //  the following revision is unnecessary for the us keymap
    //  but we need it e.g. for the french and german keymap

    {
        // backspace key

        i = 14;
        ke.kb_index=i;
        ke.kb_value=kmap[i];

        ke.kb_table=CTRL_PRESSED;
        ioctl(fd,KDSKBENT,(unsigned int)&ke);

        ke.kb_table=SHIFTCTRL_PRESSED;
        ioctl(fd,KDSKBENT,(unsigned int)&ke);

        ke.kb_table=ALT_PRESSED;
        ioctl(fd,KDSKBENT,(unsigned int)&ke);
    }

    {
        // Tab key

        i = 15;
        ke.kb_index=i;
        ke.kb_value=kmap[i];

        ke.kb_table=CTRL_PRESSED;
        ioctl(fd,KDSKBENT,(unsigned int)&ke);

        ke.kb_table=SHIFTCTRL_PRESSED;
        ioctl(fd,KDSKBENT,(unsigned int)&ke);

        ke.kb_table=ALT_PRESSED;
        ioctl(fd,KDSKBENT,(unsigned int)&ke);
    }

    // final touch, be sure alt-f pulls down the File menu
    ioctl(fd,KDSKBMETA,K_ESCPREFIX);


    //  and what is the total effect of above?
    //  read the file kmap_worries.txt
}


//     KDGKBMETA
//            Gets meta key handling mode.  argp points to a long which is set
//            to one of these:
//
//                K_METABIT     0x03   set high order bit
//                K_ESCPREFIX   0x04   escape prefix
//
//
//     KDSKBMETA
//            Sets meta key handling mode.  argp is a long equal to one of the
//            above values.
